#! /usr/bin/python

"""
SYNOPSIS
    Asset Register Raise Depreciation.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Balances, CCD, GetCtl, ProgressBar, Sql
from TartanClasses import TartanDialog
from tartanFunctions import dateDiff, mthendDate, showError

class ar2030:
    def __init__(self, mf, coy, nam, period):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = period
        no = self.setVariables()
        if not no:
            self.dataHeader()
            self.mf.startLoop()

    def setVariables(self):
        gc = GetCtl(self.mf)
        assctl = gc.getCtl("assctl", self.conum)
        if not assctl:
            return "no"
        self.glint = assctl["cta_glint"]
        self.rordp = assctl["cta_rordp"]
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.curdt = self.sysdtw / 100
        self.sper = self.period[1][0] / 100
        self.eper = self.period[2][0] / 100
        tabs = ["assgrp", "assmst", "assdep", "asstrn", "gentrn"]
        self.sql = Sql(self.mf.dbm, tabs, prog=__name__)
        if self.sql.error:
            return "no"
        chk = self.sql.getRec(tables="asstrn", where=[("ast_cono", "=",
            self.conum), ("ast_mtyp", "=", 4), ("ast_curdt", ">", self.eper)])
        if chk:
            showError(self.mf.body, "Period Error", "Depreciation Has "\
                "Already Been Raised in the Next Financial Period")
            return "no"

    def dataHeader(self):
        tit = ("%03i %s" % (self.conum, self.conam),
                "Asset Register Raise Depreciation (%s)" % __name__)
        fld = (
            (("T",0,0,0),"ID2",7,"Cut-off Period","Cut-off Period",
                self.curdt,"N",self.doTrdate,None,None,None),)
        tnd = ((self.endPage0, "y"),)
        txt = (self.exitPage0,)
        self.df = TartanDialog(self.mf, title=tit, eflds=fld, tend=tnd,
            txit=txt)

    def doTrdate(self, frt, pag, r, c, p, i, w):
        self.curdt = w
        if w < self.sper or w > self.eper:
            return "Invalid Date, Not in Financial Period"

    def endPage0(self):
        self.df.closeProcess()
        all = self.sql.getRec(tables="assmst", where=[("asm_cono", "=",
            self.conum)])
        if all:
            p = ProgressBar(self.mf.body, max=len(all),
                typ="Raising Depreciation ... Please Wait")
            for num, dat in enumerate(all):
                can = p.displayProgress(num)
                if can == "yes":
                    break
                self.group = CCD(dat[self.sql.assmst_col.index("asm_group")],
                    "UA", 3)
                grp = self.sql.getRec(tables="assgrp", where=[("asg_cono", "=",
                    self.conum), ("asg_group", "=", self.group.work)], limit=1)
                if not grp:
                    showError(self.mf.body, "Group Error",
                        "Missing Group (%s) Record" % self.group.disp)
                    can = "yes"
                    break
                self.depacc = grp[self.sql.assgrp_col.index("asg_depacc")]
                self.expacc = grp[self.sql.assgrp_col.index("asg_expacc")]
                self.code = dat[self.sql.assmst_col.index("asm_code")]
                self.depcod = dat[self.sql.assmst_col.index("asm_depcod")]
                bals = Balances(self.mf, "ASS", self.conum, self.sper,
                    keys=(self.group.work, self.code))
                asset = bals.doAssBals()
                if not asset:
                    continue
                self.cap, cdp, rdp, cbl, rbl, mov = asset
                # Get Last Depreciation Run Period
                lyer = self.sper / 100
                lmth = self.sper % 100
                self.lurdt = ((lyer * 100) + lmth)
                sales = False
                while not sales and self.lurdt <= self.curdt:
                    trans = self.sql.getRec(tables="asstrn", cols=["ast_mtyp",
                        "round(sum(ast_amt1), 2)"], where=[("ast_cono", "=",
                        self.conum), ("ast_group", "=", self.group.work),
                        ("ast_code", "=", self.code), ("ast_curdt", "=",
                        self.lurdt), ("ast_mtyp", "<>", 4)], group="ast_mtyp")
                    if trans:
                        for mov, amt in trans:
                            if mov == 5:
                                sales = True
                                self.cap = 0
                                break
                            self.cap = float(ASD(self.cap) + ASD(amt))
                    if self.cap:
                        self.doRaiseDep()
                    self.lurdt = self.doIncrCurdt()
            p.closeProgress()
            if can == "yes":
                self.mf.dbm.rollbackDbase()
            else:
                self.mf.dbm.commitDbase()
        self.mf.closeLoop()

    def doIncrCurdt(self):
        yr = self.lurdt / 100
        mt = self.lurdt % 100
        if mt == 12:
            yr += 1
            mt = 1
        else:
            mt += 1
        return (yr * 100) + mt

    def doRaiseDep(self):
        # Calculate Depreciations and check if Raised
        # else create transaction and accumulate variables
        #
        # Extract Closing Balance at lurdt
        where=[("ast_cono", "=", self.conum), ("ast_group", "=",
            self.group.work), ("ast_code", "=", self.code)]
        whr = where[:]
        whr.append(("ast_curdt", "<=", self.lurdt))
        bal = self.sql.getRec(tables="asstrn", cols=["round(sum(ast_amt1), 2)",
            "round(sum(ast_amt2), 2)"], where=whr, limit=1)
        if bal and bal[0] <= 0:
            return
        # Get Date Purchased and Months to trdt
        self.trdt = mthendDate(self.lurdt * 100)
        acc = self.sql.getRec(tables="asstrn", cols=["min(ast_date)"],
            where=[("ast_cono", "=", self.conum), ("ast_group", "=",
            self.group.work), ("ast_code", "=", self.code), ("ast_mtyp",
            "=", 1)], limit=1)
        years = dateDiff(acc[0], self.trdt)
        if years > 6:
            # Maximum of 7 variable rates in assdep record
            years = 6
        # Get Depreciation Record
        dep = self.sql.getRec(tables="assdep", where=[("asd_cono", "=",
            self.conum), ("asd_code", "=", self.depcod)], limit=1)
        #
        # Extract Depreciation up to previous year end
        whr = where[:]
        whr.extend([("ast_mtyp", "=", 4), ("ast_curdt", "<", self.sper)])
        pmd = self.sql.getRec(tables="asstrn", cols=["round(sum(ast_amt1), 2)",
            "round(sum(ast_amt2), 2)"], where=whr, limit=1)
        if not pmd[0]:
            pmd[0] = 0
        if not pmd[1]:
            pmd[1] = 0
        #
        # Extract Depreciation for lurdt
        whr = where[:]
        whr.extend([("ast_mtyp", "=", 4), ("ast_curdt", "=", self.lurdt)])
        cmd = self.sql.getRec(tables="asstrn", cols=["max(ast_refno)",
            "round(sum(ast_amt1), 2)", "round(sum(ast_amt2), 2)"], where=whr,
            limit=1)
        if not cmd[0] or cmd[0] == "Take-On":
            self.refno = CCD(1, "Na", 9).work
        else:
            self.refno = CCD(cmd[0], "Na", 9).work
        if not cmd[1]:
            cmd[1] = 0
        if not cmd[2]:
            cmd[2] = 0
        if dep[self.sql.assdep_col.index("asd_typec")] == "S":
            # Straight Line
            cval = self.cap
            crat = dep[self.sql.assdep_col.index("asd_rate1c")]
            for x in range(2, years + 2):
                exec "n = dep[self.sql.assdep_col.index('asd_rate%sc')]" % x
                if n:
                    crat = n
        else:
            # Depreciating Balance
            cval = float(ASD(self.cap) + ASD(pmd[0]))
            crat = dep[self.sql.assdep_col.index("asd_rate1c")]
        if cval:
            if cval < 0:
                cdep = 0
            else:
                cdep = round((cval * crat / 100.0 / 12), 2)
                cdep = float(ASD(cdep) + ASD(cmd[1]))
                cbal = float(ASD(bal[0]) - ASD(cdep))
                if cbal < 0:
                    cdep = float(ASD(cdep) + ASD(cbal))
                cdep = float(ASD(0) - ASD(cdep))
        typer = dep[self.sql.assdep_col.index("asd_typer")]
        if self.rordp == "N" or typer == "N":
            rdep = 0
        else:
            if typer == "S":
                # Straight Line
                rval = self.cap
                rrat = dep[self.sql.assdep_col.index("asd_rate1r")]
                for x in range(2, years + 2):
                    exec "n = dep[self.sql.assdep_col.index('asd_rate%sr')]" % x
                    if n:
                        rrat = n
            else:
                # Depreciating Balance
                rval = float(ASD(self.cap) + ASD(pmd[1]))
                rrat = dep[self.sql.assdep_col.index("asd_rate1r")]
            if rval:
                if rval < 0:
                    rdep = 0
                else:
                    rdep = round((rval * rrat / 100.0 / 12), 2)
                    rdep = float(ASD(rdep) + ASD(cmd[2]))
                    rbal = float(ASD(bal[1]) - ASD(rdep))
                    if rbal < 0:
                        rdep = float(ASD(rdep) + ASD(rbal))
                    rdep = float(ASD(0) - ASD(rdep))
        if cdep or rdep:
            data = [self.conum, self.group.work, self.code, 4, self.refno,
                "AutoDep", self.trdt, 4, cdep, rdep, 0, self.lurdt,
                "Auto Raising of Depreciation", "", "", self.sysdtw, 0]
            self.sql.insRec("asstrn", data=data)
            if self.glint == "Y" and cdep:
                self.doWriteGen(self.depacc, cdep)
                cdep = float(ASD(0) - ASD(cdep))
                self.doWriteGen(self.expacc, cdep)

    def doWriteGen(self, accno, amnt):
        whr = [("glt_cono", "=", self.conum), ("glt_acno", "=", accno),
            ("glt_batch", "=", "AutoDep"), ("glt_curdt", "=", self.lurdt),
            ("glt_trdt", "=", self.trdt), ("glt_type", "=", 4)]
        acc = self.sql.getRec(tables="gentrn", where=whr)
        if acc and len(acc) == 1:
            amt = acc[0][self.sql.gentrn_col.index("glt_tramt")]
            amt = float(ASD(amt) + ASD(amnt))
            acc[0][self.sql.gentrn_col.index("glt_tramt")] = amt
            self.sql.updRec("gentrn", where=whr, data=acc[0])
        else:
            data = [self.conum, accno, self.lurdt, self.trdt, 4, self.refno,
                "AutoDep", amnt, 0, "Auto Raising of Depreciation", "", "",
                0, self.sysdtw, 0]
            self.sql.insRec("gentrn", data=data)

    def exitPage0(self):
        self.df.closeProcess()
        self.mf.closeLoop()
