#! /usr/bin/env python

"""
SYNOPSIS
    Asset Register Register.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Balances, CCD, GetCtl, ProgressBar, Sql
from TartanClasses import TartanDialog
from tartanFunctions import getModName, doPrinter, doReportlab, doWriteExport
from tartanFunctions import showError

class ar3040:
    def __init__(self, mf, coy, nam, period):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = period
        no = self.setVariables()
        if not no:
            self.mainProcess()
            self.mf.startLoop()

    def setVariables(self):
        self.gc = GetCtl(self.mf)
        assctl = self.gc.getCtl("assctl", self.conum)
        if not assctl:
            return "no"
        self.rordp = assctl["cta_rordp"]
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sysdtd = "%i/%02i/%02i" % (t[0], t[1], t[2])
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i) %6s" % (t[0],
            t[1], t[2], t[3], t[4], __name__)
        self.sql = Sql(self.mf.dbm, ["assgrp", "assmst", "asstrn"],
            prog=__name__)
        self.sper = self.period[1][0] / 100
        self.eper = self.period[2][0] / 100
        self.head = ("%03u %-30s %s" % (self.conum, self.conam, "%s"))
        self.colsh = ["Code", "Description", "Cost-Price", "Accum-Dep",
            "Opening-Bal", "Purchases", "Improvements", "Write-Offs",
            "Depreciation", "Sales", "Profit/Loss", "Closing-Bal"]
        self.forms = [("Na", 7), ("NA", 30)] + [("SD", 13.2)] * 10
        self.stot = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.gtot = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.oldgrp = ""
        self.t = None

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "Asset Register Report (%s)" % __name__)
        eper = self.sysdtw / 100
        if eper > self.eper:
            eper = self.eper
        grp = {
            "stype": "R",
            "tables": ("assgrp",),
            "cols": (
                ("asg_group", "", 0, "Grp"),
                ("asg_desc", "", 0, "Description", "Y")),
            "where": [("asg_cono", "=", self.conum)]}
        r1s = (("Company","C"), ("Receiver","R"))
        r2s = (("Yes","Y"), ("No","N"))
        fld = [
            (("T",0,0,0),"ID2",7,"Cut-Off Period","Cut-Off Period",
                eper,"Y",self.doPeriod,None,None,("efld",))]
        if self.rordp == "Y":
            fld.append((("T",0,1,0),("IRB",r1s),0,"Report Type","Report Type",
                "C","Y",self.doType,None,None,None))
        else:
            self.rtype = "C"
        fld.extend([
            (("T",0,2,0),"IUA",3,"Asset Group","Asset Group",
                "","Y",self.doGroup,grp,None,None),
            (("T",0,3,0),("IRB",r2s),0,"Ignore Zero Items","Ignore Zero Items",
                "N","Y",self.doZero,None,None,None)])
        tnd = ((self.doEnd,"Y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("Y","V"), mail=("Y","N"))

    def doPeriod(self, frt, pag, r, c, p, i, w):
        if w < self.sper or w > self.eper:
            return "Invalid Period"
        self.endper = w

    def doType(self, frt, pag, r, c, p, i, w):
        self.rtype = w

    def doGroup(self, frt, pag, r, c, p, i, w):
        if w:
            acc = self.sql.getRec(tables="assgrp", cols=["asg_desc"],
                where=[("asg_cono", "=", self.conum), ("asg_group", "=", w)],
                limit=1)
            if not acc:
                return "Invalid Group"
        self.group = w

    def doZero(self, frt, pag, r, c, p, i, w):
        self.zero = w

    def doEnd(self):
        self.df.closeProcess()
        if self.rtype == "C":
            self.typ = "Company"
        else:
            self.typ = "Receiver"
        if not self.group:
            self.sgrp = ""; self.egrp = "zzz"
        else:
            self.sgrp = self.group; self.egrp = self.group
        cols = ["asm_group", "asm_code", "asm_desc"]
        all = self.sql.getRec(tables="assmst", cols=cols, where=[("asm_cono",
            "=", self.conum), ("asm_group", ">=", self.sgrp), ("asm_group",
            "<=", self.egrp)], order="asm_group, asm_code")
        if not all:
            showError(self.mf.body, "Processing Error", "No Records Selected")
        elif self.df.repprt[0][2] == "export":
            self.exportReport(all)
        else:
            self.printReport(all)
        self.closeProcess()

    def exportReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        expnam = getModName(self.mf.wrkdir, __name__, self.conum)
        self.expheads = [self.head % self.sysdttm]
        self.expheads.append("Asset Register Report as at %s" % self.sysdtd)
        self.expheads.append("Options: Report Period: %s Report Type: %s" % \
            (self.df.t_disp[0][0][0], self.typ))
        self.expcolsh = [self.colsh]
        self.expforms = self.forms
        self.expdatas = []
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                return
            if not num:
                desc = self.groupDesc(dat[0])
                self.expdatas.append(["HEAD", ["", "Group %s %s" % (dat[0],
                    desc)]])
                self.expdatas.append(["BLANK"])
            vals = self.getValues(dat)
            if not vals:
                continue
            if self.oldgrp and self.oldgrp != dat[0]:
                self.groupTotal()
                self.expdatas.append(["BLANK"])
                desc = self.groupDesc(dat[0])
                self.expdatas.append(["HEAD", ["", "Group %s %s" % (dat[0],
                    desc)]])
                self.expdatas.append(["BLANK"])
            line = ["BODY", [dat[1], dat[2]]]
            for num, val in enumerate(vals):
                line[1].append(val)
                self.stot[num] = float(ASD(self.stot[num]) + ASD(val))
                self.gtot[num] = float(ASD(self.gtot[num]) + ASD(val))
            self.expdatas.append(line)
            self.oldgrp = dat[0]
        p.closeProgress()
        self.groupTotal()
        self.grandTotal()
        doWriteExport(xtype=self.df.repprt[0][1], name=expnam, sheet=__name__,
            heads=self.expheads, colsh=self.expcolsh, forms=self.expforms,
            datas=self.expdatas, rcdic=self.mf.rcdic)

    def printReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        chrs = len(self.colsh)
        for f in self.forms:
            chrs += int(f[1])
        if f[0][0].lower() == "s":
            chrs -= 2
        else:
            chrs -= 1
        pad = chrs - 35 - len(self.sysdttm)
        self.head1 = self.head % (" " * pad + self.sysdttm)
        self.head2 = "Asset Register Report as at %s%s" % (self.sysdtd, "%s%s")
        pad = chrs - len(self.head2) + 4 - 11 # %s%s and ' Page     1'
        self.head2 = self.head2 % (" " * pad, " Page %5s")
        pdfnam = getModName(self.mf.wrkdir, __name__, self.conum, ext="pdf")
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, pdfnam, self.head1)
        self.pgnum = 0
        self.pglin = 99
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                if self.t:
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                break
            grp = CCD(dat[0], "UA", 3)
            code = CCD(dat[1], "Na", 7)
            desc = CCD(dat[2], "NA", 30)
            vals = self.getValues(dat)
            if not vals:
                continue
            if self.oldgrp and self.oldgrp != grp.work:
                self.groupTotal()
                self.pglin = 99
            if self.pglin > self.lpp:
                if self.pgnum != 0:
                    self.c.drawText(self.t)
                    self.c.showPage()
                self.pageHeading(grp)
            for num, val in enumerate(vals):
                exec "v%s = CCD(val, 'SD', 13.2)" % num
                self.stot[num] = float(ASD(self.stot[num]) + ASD(val))
                self.gtot[num] = float(ASD(self.gtot[num]) + ASD(val))
            self.t.textLine("%s %s %s %s %s %s %s %s %s %s %s %s" % (code.disp,
                desc.disp, v0.disp, v1.disp, v2.disp, v3.disp, v4.disp,
                v5.disp, v6.disp, v7.disp, v8.disp, v9.disp))
            self.pglin = self.pglin + 1
            self.oldgrp = grp.work
        p.closeProgress()
        if self.t:
            if can != "yes":
                self.groupTotal()
                self.c.drawText(self.t)
                if self.pglin > (self.lpp - 2):
                    self.c.showPage()
                    self.pageHeading(grp)
                self.grandTotal()
                self.c.drawText(self.t)
        self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=pdfnam, head=self.tit,
            view=self.df.repprt[0][1], prtnam=self.df.repprt[0][2],
            mail=self.df.repprt[1][2])

    def getValues(self, data):
        line = [data[0]]
        bals = Balances(self.mf, "ASS", self.conum, self.sper, keys=(data[0],
            data[1]))
        asset = bals.doAssBals(start=self.sper, end=self.endper)
        if not asset:
            return
        cap, cdp, rdp, cbl, rbl, mov = asset
        vals = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        vals[0] = cap
        if self.rtype == "C":
            vals[1] = cdp
            vals[2] = cbl
            vals[9] = cbl
        else:
            vals[1] = rdp
            vals[2] = rbl
            vals[9] = rbl
        woff = False
        sold = False
        if mov:
            for m in mov:
                if m[0] == 3:
                    woff = True
                if m[0] == 5:
                    sold = True
                if self.rtype == "C":
                    idx = 1
                else:
                    idx = 2
                vals[m[0]+2] = float(ASD(vals[m[0]+2]) + ASD(m[idx]))
                vals[9] = float(ASD(vals[9]) + ASD(m[idx]))
        if woff or sold:
            vals[8] = vals[9]
            vals[9] = 0
        if self.zero == "Y" and not vals[2] and not vals[3] \
            and not vals[4] and not vals[5] and not vals[6] \
            and not vals[7] and not vals[8] and not vals[9]:
            return
        return vals
            
    def pageHeading(self, grp):
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.pgnum = self.pgnum + 1
        self.t.textLine(self.head1)
        self.t.textLine()
        self.t.textLine(self.head2 % self.pgnum)
        self.t.textLine()
        self.t.textLine("(%s %s %s %s)" % ("Options: Report Period:",
            self.df.t_disp[0][0][0], "Report Type:", self.typ))
        self.t.textLine()
        grpdesc = self.groupDesc(grp.work)
        self.t.textLine("%-5s %3s %-30s" % ("Group", grp.disp, grpdesc))
        self.t.textLine()
        self.t.textLine("%-7s %-30s %13s %13s %13s %13s %13s %13s %13s %13s "\
            "%13s %13s" % ("Code", "Description", "Cost-Price ", "Accum-Dep ",
            "Opening-Bal ", "Purchases ", "Improvements ", "  Write-Offs ",
            "Depreciation ", "Sales ", " Profit/Loss ", " Closing-Bal "))
        self.t.textLine("%s" % (self.suc * len(self.head1)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        self.pglin = 10

    def groupDesc(self, group):
        acc = self.sql.getRec(tables="assgrp", cols=["asg_desc"],
            where=[("asg_cono", "=", self.conum), ("asg_group", "=",
            group)], limit=1)
        if not acc:
            return ""
        else:
            return acc[0]

    def groupTotal(self):
        if self.df.repprt[0][2] == "export":
            self.expdatas.append(["ULINES"])
            self.expdatas.append(["TOTAL",
                ["", "Group %s Totals" % self.oldgrp] + self.stot])
            self.stot = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
            return
        for num, val in enumerate(self.stot):
            exec "v%s = CCD(val, 'SD', 13.2)" % num
        self.t.textLine()
        self.t.textLine("%-7s %-30s %13s %13s %13s %13s %13s %13s %13s %13s "\
            "%13s %13s" % ("", "Group %s Totals" % self.oldgrp , v0.disp,
            v1.disp, v2.disp, v3.disp, v4.disp, v5.disp, v6.disp, v7.disp,
            v8.disp, v9.disp))
        self.t.textLine()
        self.stot = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]

    def grandTotal(self):
        if self.df.repprt[0][2] == "export":
            self.expdatas.append(["ULINES"])
            self.expdatas.append(["TOTAL", ["", "Grand Totals"] + self.gtot])
            self.expdatas.append(["ULINED"])
            return
        for num, val in enumerate(self.gtot):
            exec "v%s = CCD(val, 'SD', 13.2)" % num
        self.t.textLine("%-7s %-30s %13s %13s %13s %13s %13s %13s %13s %13s "\
            "%13s %13s" % (" ", "Grand Totals", v0.disp, v1.disp, v2.disp,
            v3.disp, v4.disp, v5.disp, v6.disp, v7.disp, v8.disp, v9.disp))

    def doExit(self):
        self.df.closeProcess()
        self.closeProcess()

    def closeProcess(self):
        self.mf.closeLoop()
