#! /usr/bin/python

"""
SYNOPSIS
    General Ledger Data Capture for:

        1 Sales
        2 Payments
        3 Petty Cash
        5 Purchases
        6 Receipts
        7 Bank Statements

    Flags used:
        self.incoac - Intercompany accounts posting
        self.dorec  - Bank Statement capture
        self.recon  - Transaction found during Statement Capture
        self.rctimp - RCT (Bank codes) Transactions import
        self.rctupd - Update genrct table

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import AgeAll, ASD, Batches, CCD, GetCtl, ProgressBar
from TartanClasses import PwdConfirm, BankImport, SChoice, Sql, TartanDialog
from tartanFunctions import askQuestion, callModule, getSingleRecords
from tartanFunctions import mthendDate, showError
from tartanWork import armvtp, gltrtp

class gl2030:
    def __init__(self, mf, conum, conam, period, rtn, usr):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        self.rtn = rtn
        self.capnm = usr
        no = self.checkControls()
        if not no:
            self.batchHeader()
            if self.bh.batno:
                no = self.setVariables()
                if not no:
                    self.drawDialog()
                    self.mf.startLoop()

    def checkControls(self):
        # Check for Valid Posting Routine
        if self.rtn not in (1, 2, 3, 5, 6, 7):
            showError(self.mf.body, "Error", "Invalid Routine " + str(self.rtn))
            return "no"
        # Check for Company Record
        self.gc = GetCtl(self.mf)
        ctlmst = self.gc.getCtl("ctlmst", self.conum)
        if not ctlmst:
            return "no"
        self.allcoy = self.conum
        # Check for Intercompany Facility
        sql = Sql(self.mf.dbm, ["ctlmst", "genint"], prog=__name__)
        coy = sql.getRec(tables="ctlmst", cols=["count(*)"],
            where=[("ctm_cono", "<>", self.conum)], limit=1)
        if not coy[0]:
            self.incoac = False
        else:
            int = sql.getRec(tables="genint", cols=["cti_inco"],
                where=[("cti_cono", "=", self.conum)])
            if int:
                self.incoac = [self.conum]
                [self.incoac.append(coy[0]) for coy in int]
            else:
                self.incoac = False
        #
        tab = ["ctlctl", "ctlmst", "ctlvmf", "ctlvtf", "genint", "genmst",
            "gentrn", "genrcc", "genrct"]
        mod = []
        for x in range(0, 21, 2):
            mod.append(ctlmst["ctm_modules"][x:x+2])
        if mod.count("AR"):
            ctl = self.gc.getCtl("assctl", self.conum)
            if not ctl:
                return "no"
            if ctl["cta_glint"] == "Y":
                tab.extend(["assgrp", "assmst", "asstrn"])
        if mod.count("CR"):
            ctl = self.gc.getCtl("crsctl", self.conum)
            if not ctl:
                return "no"
            if ctl["ctc_glint"] == "Y":
                if self.rtn == 5:
                    showError(self.mf.body, "Invalid Routine",
                        "Because Creditors are Integrated, All %ss Must be "\
                        "Entered Via the Creditors System else the Control "\
                        "Account Will Not Balance!" % gltrtp[self.rtn-1][1])
                    return "no"
                tab.extend(["crsmst", "crstrn"])
        if mod.count("DR"):
            ctl = self.gc.getCtl("drsctl", self.conum)
            if not ctl:
                return "no"
            if ctl["ctd_glint"] == "Y":
                if self.rtn == 1:
                    showError(self.mf.body, "Invalid Routine",
                        "Because Debtors are Integrated, All %ss Must be "\
                        "Entered Via the Debtors System else the Control "\
                        "Account Will Not Balance!" % gltrtp[self.rtn-1][1])
                    return "no"
                tab.extend(["drsmst", "drstrn"])
        if mod.count("ML"):
            ctl = self.gc.getCtl("memctl", self.conum)
            if not ctl:
                return "no"
            if ctl["mcm_glint"] == "Y":
                tab.extend(["memmst", "memtrn"])
        # Check if Control Accounts Exist
        self.ctlctl = self.gc.getCtl("ctlctl", self.conum)
        if not self.ctlctl:
            return "no"
        if self.rtn == 1:
            if self.gc.chkRec(self.conum, self.ctlctl, ["drs_ctl"]):
                return "no"
        elif self.rtn == 3:
            if self.gc.chkRec(self.conum, self.ctlctl, ["p_cash"]):
                return "no"
        elif self.rtn == 5:
            if self.gc.chkRec(self.conum, self.ctlctl, ["crs_ctl"]):
                return "no"
        # Create SQL Object
        self.sql = Sql(self.mf.dbm, tab, prog=__name__)
        if not self.incoac:
            return self.doChkLoadCtls()

    def batchHeader(self):
        self.bh = Batches(self.mf, self.conum, self.conam, self.period,
            "GEN", self.rtn)
        self.bh.doBatchHeader()
        if self.bh.batno and self.rtn not in (1, 6):
            self.bh.batval = float(ASD(0) - ASD(self.bh.batval))

    def setVariables(self):
        t = time.localtime()
        self.sysdtw = ((t[0] * 10000) + (t[1] * 100) + t[2])
        # Load Control Accounts
        if self.rtn == 1:
            self.conacc = self.ctlctl["drs_ctl"]
        elif self.rtn == 2:
            self.conacc = self.bh.acc
        elif self.rtn == 3:
            self.conacc = self.ctlctl["p_cash"]
        elif self.rtn == 5:
            self.conacc = self.ctlctl["crs_ctl"]
        elif self.rtn == 6:
            self.conacc = self.bh.acc
        # Initialise Pages
        self.page3 = False
        self.page4 = False
        self.page5 = False
        # Bank Import Codes
        self.rctimp = False
        if self.rtn == 7:
            imp = self.sql.getRec(tables="ctlctl", where=[("ctl_cono", "=",
                self.conum), ("ctl_code", "=", self.bh.ctl)], limit=1)
            self.bankac = imp[self.sql.ctlctl_col.index("ctl_bankac")]
            self.impfmt = imp[self.sql.ctlctl_col.index("ctl_impfmt")]
            self.dtefmt = imp[self.sql.ctlctl_col.index("ctl_dtefmt")]
            self.dorec = True
        else:
            self.dorec = False
        self.cancel = False
        self.agecan = False
        self.batupd = False

    def drawDialog(self):
        self.crspag = 3
        self.drspag = 4
        self.mempag = 5
        self.asspag = 6
        tit = ("%03i %s" % (self.conum, self.conam), "G/Ledger " + \
            gltrtp[self.rtn - 1][1] + " Capture (%s)" % __name__)
        coy = {
            "stype": "R",
            "tables": ("ctlmst",),
            "cols": (
                ("ctm_cono", "", 0, "Coy"),
                ("ctm_name", "", 0, "Name", "Y"))}
        glm = {
            "stype": "R",
            "tables": ("genmst",),
            "cols": (
                ("glm_acno", "", 0, "Acc-Num"),
                ("glm_desc", "", 0, "Description", "Y")),
            "whera": [["C", "glm_cono", 0, 2]],
            "ctype": "int"}
        crm = {
            "stype": "R",
            "tables": ("crsmst",),
            "cols": (
                ("crm_acno", "", 0, "Acc-Num"),
                ("crm_name", "", 0, "Name", "Y")),
            "whera": [["C", "crm_cono", 0, 2]]}
        chn = {
            "stype": "R",
            "tables": ("drschn",),
            "cols": (
                ("chm_chain", "", 0, "Num"),
                ("chm_name", "", 0, "Name", "Y")),
            "whera": [["C", "chm_cono", 0, 2]]}
        drm = {
            "stype": "R",
            "tables": ("drsmst",),
            "cols": (
                ("drm_acno", "", 0, "Acc-Num"),
                ("drm_name", "", 0, "Name", "Y"),
                ("drm_add1", "", 0, "Address Line 1")),
            "whera": [
                ["C", "drm_cono", 0, 2],
                ["C", "drm_chain", 0, self.drspag]]}
        mlm = {
            "stype": "R",
            "tables": ("memmst",),
            "cols": (
                ("mlm_memno", "", 0, "Mem-No"),
                ("mlm_surname", "", 0, "Surname", "Y"),
                ("mlm_names", "", 0, "Names")),
            "whera": [
                ["C", "mlm_cono", 0, 2]]}
        asg = {
            "stype": "R",
            "tables": ("assgrp",),
            "cols": (
                ("asg_group", "", 0, "Grp"),
                ("asg_desc", "", 0, "Description", "Y")),
            "whera": [["C", "asg_cono", 0, 2]]}
        asm = {
            "stype": "R",
            "tables": ("assmst",),
            "cols": (
                ("asm_code", "", 0, "Cod-Num"),
                ("asm_desc", "", 0, "Description", "Y")),
            "whera": [
                ["C", "asm_cono", 0, 2],
                ["C", "asm_group", 0, self.asspag]]}
        data = []
        for x in range(1, len(armvtp)+1):
            if x == 4:
                continue
            data.append([x, armvtp[x-1][1]])
        mov = {
            "stype": "C",
            "titl": "Valid Types",
            "data": data}
        vat = {
            "stype": "R",
            "tables": ("ctlvmf",),
            "cols": (
                ("vtm_code", "", 0, "C"),
                ("vtm_desc", "", 0, "Description", "Y")),
            "whera": [["C", "vtm_cono", 0, 2]]}
        tag = [
            ("Transaction",None,None,None),
            ("Allocation",None,None,None),
            ("Creditors",None,None,None),
            ("Debtors",None,None,None),
            ("Members",None,None,None),
            ("Assets",None,None,None)]
        fld = [
            (("T",0,0,0),"OUI",9,"Batch %s Quantity" % self.bh.batno),
            (("T",0,0,0),"OSD",13.2,"Value"),
            [["C",1,0,0],"INa",9,"Reference","Reference Number",
                "i","Y",self.doTrnRef,None,None,("notblank",)],
            [["C",1,0,1],"ID1",10,"Date","Transaction Date",
                "r","N",self.doTrnDat,None,None,("efld",)],
            [["C",1,0,2],"ISD",13.2,"Amount","Transaction Amount",
                "","N",self.doTrnAmt,None,None,("efld",)],
            [["C",1,0,3],"INA",30,"Details","Transaction Details",
                "","N",self.doTrnDet,None,None,("notblank",)],
            [["T",2,0,0],"OSD",13.2,"Unallocated Balance"],
            [["C",2,0,0],"IUI",3,"Coy","Company Number",
                self.conum,"N",self.doCoyNum,coy,None,("notzero",)],
            [["C",2,0,1],"IUI",7,"Acc-Num","Account Number",
                "","N",self.doAccNum,glm,None,None],
            [["C",2,0,2],"ONA",30,"Description"],
            [["C",2,0,3],"ISD",13.2,"Alloc-Amt","Allocation Amount",
                "","N",self.doAllAmt,None,None,("efld",)],
            [["C",2,0,4],"IUA",1,"V","V.A.T. Code",
                "","N",self.doVatCod,vat,None,("notblank",)],
            [["C",2,0,5],"ISD",13.2,"VAT-Amt","V.A.T. Amount",
                "","N",self.doVatAmt,None,None,("efld",)],
            [["C",2,0,6],"INA",(28,30),"Details","Allocation Details",
                "","N",self.doAllDet,None,None,("notblank",)],
            [["T",3,0,0],"OSD",13.2,"Unallocated Balance"],
            [["C",3,0,0],"INA",7,"Acc-Num","Account Number",
                "","N",self.doCrsAcc,crm,None,("notblank",)],
            [["C",3,0,1],"ONA",30,"Name"],
            [["C",3,0,2],"INa",9,"Ref-No2","Reference Number",
                "i","N",self.doCrsRef,None,None,None],
            [["C",3,0,3],"ISD",13.2,"Discount","Discount",
                "","N",self.doCrsDis,None,None,("efld",)],
            [["C",3,0,4],"ISD",13.2,"Amount","Amount",
                "","N",self.doCrsAmt,None,None,("efld",)],
            [["T",4,0,0],"OSD",13.2,"Unallocated Balance"],
            [["C",4,0,0],"IUI",3,"Chn","Chain Store Code",
                "","N",self.doDrsChn,chn,None,("efld",)],
            [["C",4,0,1],"INA",7,"Acc-Num","Account Number",
                "","N",self.doDrsAcc,drm,None,("notblank",)],
            [["C",4,0,2],"ONA",30,"Name"],
            [["C",4,0,3],"INa",9,"Ref-No2","Reference Number",
                "i","N",self.doDrsRef,None,None,None],
            [["C",4,0,4],"ISD",13.2,"Discount","Discount",
                "","N",self.doDrsDis,None,None,("efld",)],
            [["C",4,0,5],"ISD",13.2,"Amount","Amount",
                "","N",self.doDrsAmt,None,None,("efld",)],
            [["T",5,0,0],"OSD",13.2,"Unallocated Balance"],
            [["C",5,0,1],"INA",7,"Acc-Num","Member Number",
                "","N",self.doMemAcc,mlm,None,("notblank",)],
            [["C",5,0,2],"ONA",30,"Name"],
            [["C",5,0,4],"ISD",13.2,"Discount","Discount",
                "","N",self.doMemDis,None,None,("efld",)],
            [["C",5,0,5],"ISD",13.2,"Amount","Amount",
                "","N",self.doMemAmt,None,None,("efld",)],
            [["T",6,0,0],"OSD",13.2,"Unallocated Balance"],
            [["C",6,0,0],"IUA",3,"Grp","Asset Group",
                "r","N",self.doAssGrp,asg,None,("notblank",)],
            [["C",6,0,1],"INa",7,"Cod-Num","Asset Code",
                "r","N",self.doAssCod,asm,None,("notblank",)],
            [["C",6,0,2],"ONA",30,"Description"],
            [["C",6,0,3],"IUI",1,"M","Movement Type",
                "r","N",self.doAssMov,mov,None,("in",(1,2,5))],
            [["C",6,0,4],"ISD",13.2,"Amount","Amount",
                "","N",self.doAssAmt,None,None,("efld",)],
            [["C",6,0,5],"INA",30,"Details","Transaction Details",
                "","N",self.doTrnDet,None,None,("efld",)]]
        if not self.incoac:
            fld[7][1] = "OUI"
        cnd = [
            (None, "n"),
            (self.endPage1, "y"),
            (self.endPage2, "n"),
            (self.endCrsDrsMem, "n"),
            (self.endCrsDrsMem, "n"),
            (self.endCrsDrsMem, "n"),
            (self.endAssets, "n")]
        cxt = [
            None,
            self.exitPage1,
            self.exitPage2,
            self.exitPage3,
            self.exitPage4,
            self.exitPage5,
            self.exitPage6]
        if (self.rtn in (2, 6) or self.dorec):
            self.but = [
                ["Age _Normal",None,self.doAgeNormal,0,None,None,
                    "Only Show Unallocated Transactions"],
                ["Age _History",None,self.doAgeHistory,0,None,None,
                    "Show All Transactions, Including Already Allocated"],
                ["Age _Automatic",None,self.doAgeAuto,0,None,None,
                    "Automatically Allocate the Amount Starting With the "\
                    "Oldest Unallocated One"],
                ["Age _Current",None,self.doAgeCurrent,0,None,None,
                    "Leave the Transaction Unallocated"]]
        else:
            if self.rtn == 3:
                fld[2][10] = None
            self.but = []
        if self.dorec:
            sel = {
                "stype": "M",
                "func": self.doSelections}
            self.glt = {
                "stype": "R",
                "tables": ("gentrn",),
                "cols": (
                    ("glt_trdt", "", 0, "Date"),
                    ("glt_type", ("XX", gltrtp), 3, "Typ"),
                    ("glt_refno", "", 0, "Reference", "Y"),
                    ("glt_seq", "", 0, "Sequence"),
                    ("glt_tramt", "", 0, "Amount"),
                    ("glt_desc", "", 0, "Details")),
                "where": [
                    ("glt_cono", "=", self.conum),
                    ("glt_acno", "=", self.bh.acc),
                    ("glt_recon", "=", 0)],
                "order": "glt_trdt, glt_type, glt_refno",
                "index": 2,
                "zero": "0"}
            fld.insert(2, [("C",1,0,0),"IUA",1,"T","Reference Type (P/D)",
                "P","N",self.doTrnTyp,sel,None,("in", ("P","D"))])
            fld[3][0][3] = 1
            fld[3][8] = self.glt
            fld[4][0][3] = 2
            fld[5][0][3] = 3
            fld[6][0][3] = 4
            for b in range(0, len(self.but)):
                self.but[b].extend([None,1,b])
            self.but.insert(0,
                ("Exit",None,self.exitPage1,0,("C",1,1),("C",1,2)))
            self.but.insert(1,
                ("Import Bank File",None,self.doImpBank,0,("C",1,1),("C",1,2)))
            self.but.insert(2,
                ("Process Bank Data",None,self.doImpRct,0,("C",1,1),("C",1,2)))
            self.but.insert(3,
                ("Cancel",None,self.doCancel,1,("C",1,0),(("C",1,1),("C",1,5))))
        else:
            self.but.append(
                ("Cancel",None,self.doCancel,1,("C",1,0),("C",1,1)))
        self.df = TartanDialog(self.mf, title=tit, tags=tag, eflds=fld,
            butt=self.but, cend=cnd, cxit=cxt)
        self.df.loadEntry("T", 0, 0, data=self.bh.batqty)
        self.df.loadEntry("T", 0, 1, data=self.bh.batval)
        self.df.focusField("C", 1, 1)

    def doImpBank(self, widget=None):
        if not self.impfmt in ("O", "Q", "S"):
            showError(self.mf.body, "Recon Error",
                "Invalid Import Format Type for Control %s" % self.bh.ctl)
            self.df.focusField("C", 1, self.df.col)
            return
        if self.impfmt == "S" and not self.bankac:
            showError(self.mf.body, "Recon Error",
                "Invalid Bank Account for Control %s" % self.bh.ctl)
            self.df.focusField("C", 1, self.df.col)
            return
        if self.impfmt in ("O", "Q") and not self.dtefmt:
            showError(self.mf.body, "Recon Error",
                "Invalid Date Format for Control %s" % self.bh.ctl)
            self.df.focusField("C", 1, self.df.col)
            return
        self.df.setWidget(self.df.B0, "disabled")
        self.df.setWidget(self.df.B1, "disabled")
        self.df.setWidget(self.df.B2, "disabled")
        self.df.setWidget(self.df.colEntry[1][self.df.pos], "disabled")
        state = self.df.disableButtons()
        tit = ("Bank %s Import" % self.bankac,)
        fld = (
            (("T",0,0,0),"ID1",10,"From Date","Import From Date",
                "","N",self.doFmDate,None,None,("efld",)),
            (("T",0,1,0),"ID1",10,"To  Date","Import To Date",
                "","N",self.doToDate,None,None,("efld",)))
        tnd = ((self.doImpEnd,"n"), )
        txt = (self.doImpExit, )
        self.ip = TartanDialog(self.mf, newscn=(0, 0), tops=True, title=tit,
            eflds=fld, tend=tnd, txit=txt)
        self.mf.startLoop()
        self.df.enableButtons(state=state)
        self.df.focusField("C", 1, self.df.col)

    def doFmDate(self, frt, pag, r, c, p, i, w):
        self.ifrom = w

    def doToDate(self, frt, pag, r, c, p, i, w):
        self.ito = w

    def doImpEnd(self):
        self.ip.closeProcess()
        self.mf.closeLoop()
        data = BankImport(self.mf, self.impfmt, self.bankac, self.dtefmt)
        if data.trans:
            for date, memo, refno, payee, amount in data.trans:
                date = CCD(date, "D1", 10)
                memo = CCD(memo, "NA", 100)
                refno = CCD(refno, "Na", 9)
                payee = CCD(payee, "NA", 50)
                amount = CCD(amount, "SD", 13.2)
                if date.work < self.ifrom or date.work > self.ito:
                    continue
                if refno.work in ("", "        0", 0):
                    autoref = True
                    ref = memo.work.split()
                    if amount.work < 0 and ref[0] == "CHEQUE" and \
                                           ref[1] != "CASHED":
                        # Nedbank cheque number
                        refno = ref[1][1:8]
                    else:
                        # Next automatic reference
                        acc = self.sql.getRec(tables="genrct",
                            cols=["max(grt_refno)"], where=[("grt_cono", "=",
                            self.conum), ("grt_acno", "=", self.bh.acc),
                            ("grt_refno", "like", "BS_______")], limit=1)
                        if acc:
                            try:
                                auto = int(acc[0][2:]) + 1
                            except:
                                auto = 1
                        else:
                            auto = 1
                        refno = "BS%07d" % auto
                    refno = CCD(refno, "Na", 9)
                else:
                    autoref = False
                whr = [
                    ("grt_cono", "=", self.conum),
                    ("grt_acno", "=", self.bh.acc),
                    ("grt_date", "=", date.work),
                    ("grt_memo", "=", memo.work),
                    ("grt_payee", "=", payee.work),
                    ("grt_amount", "=", amount.work)]
                if not autoref:
                    whr.append(("grt_refno", "=", refno.work))
                acc = self.sql.getRec(tables="genrct", where=whr)
                if acc and autoref:
                    ok = askQuestion(self.mf.body, head="Import Error",
                        mess="This record seems to be duplicated,\n"\
                            "  Date:      %s\n"\
                            "  Payee:     %s\n"\
                            "  Amount:    %s\n\n"\
                            "  Import Anyway?" % \
                            (date.disp, payee.disp, amount.disp),
                        default="no")
                    if ok == "no":
                        continue
                elif acc:
                    continue
                data = [self.conum, self.bh.acc, date.work, memo.work,
                    refno.work, payee.work, amount.work, "N", 0]
                self.sql.insRec("genrct", data=data)
            self.mf.dbm.commitDbase()

    def doImpExit(self):
        self.ip.closeProcess()
        self.mf.closeLoop()

    def doImpRct(self, widget=None):
        self.df.setWidget(self.df.B0, "disabled")
        self.df.setWidget(self.df.B1, "disabled")
        self.df.setWidget(self.df.B2, "disabled")
        self.rctimp = True
        self.df.setWidget(self.df.colEntry[1][self.df.pos], "disabled")
        # Try and automatically allocate genrct records
        colf = self.df.colf[1][4]
        self.df.colf[1][4][4] += "(noesc)"
        data = self.doReadRct()
        if not data:
            data = []
        else:
            self.df.mstAlw.hide()
            p = ProgressBar(self.mf.body, typ="Automatically "\
                "Allocating, Please Wait....", max=len(data))
            for num, line in enumerate(data):
                p.displayProgress(num)
                dte = line[self.sql.genrct_col.index("grt_date")]
                mem = line[self.sql.genrct_col.index("grt_memo")][:50]
                ref = line[self.sql.genrct_col.index("grt_refno")]
                amt = line[self.sql.genrct_col.index("grt_amount")]
                seq = line[self.sql.genrct_col.index("grt_seq")]
                self.doAutoRct(dte, mem, ref, amt, seq)
            p.closeProgress()
            self.df.mstAlw.show()
        # Show all unallocated genrct and select one to allocate
        while "doall":
            data = self.doReadRct()
            if not data:
                break
            # Manual Postings and Allocations of unallocated entries
            titl = "Unallocated Statement Entries"
            head = ("   Date", "Memo-Details", "Reference",
                "Payee-Details", "      Amount ")
            types = [["D1", 10],["NA", 50],["Na", 9],["NA", 20],["SD", 13.2]]
            lines = []
            for line in data:
                dte = line[self.sql.genrct_col.index("grt_date")]
                mem = line[self.sql.genrct_col.index("grt_memo")][:50]
                ref = line[self.sql.genrct_col.index("grt_refno")]
                pay = line[self.sql.genrct_col.index("grt_payee")][:20]
                amt = line[self.sql.genrct_col.index("grt_amount")]
                lines.append([dte, mem, ref, pay, amt])
            state = self.df.disableButtons()
            but = [("Create Import Record", self.doImpRec, True)]
            sc = SChoice(self.mf, scrn=self.mf.body, titl=titl, head=head,
                data=lines, typs=types, retn="I", mode="S", butt=but)
            self.df.enableButtons(state=state)
            if not sc.selection:
                break
            index = sc.selection[0]
            self.recon = False
            self.rctupd = False
            a,b,self.trndat,memo,self.trnref,payee,amount,c,seq = data[index]
            i = self.df.idx
            p = self.df.pos
            r = self.df.row
            if amount < 0:
                self.typ = "P"
                self.rtn = 2
                self.trnamt = float(ASD(0) - ASD(amount))
            else:
                self.typ = "D"
                self.rtn = 6
                self.trnamt = amount
            self.df.loadEntry("C", 1, p, data=self.typ)
            self.df.loadEntry("C", 1, p+1, data=self.trnref)
            self.df.loadEntry("C", 1, p+2, data=self.trndat)
            self.df.loadEntry("C", 1, p+3, data=self.trnamt)
            self.df.loadEntry("C", 1, p+4, data=memo[:30])
            self.df.focusField("C", 1, p+5)
            self.mf.startLoop()
            if self.rctupd and not self.cancel:
                self.doUpdateRct(seq)
        self.rctimp = False
        self.df.colf[1][4][4] = colf
        self.df.clearLine(1, self.df.row)
        self.df.focusField("C", 1, self.df.col)

    def doImpRec(self, *args):
        coy = (self.conum, self.conam)
        args = (self.bh.acc, args[0])
        self.df.mstAlw.hide()
        callModule(self.mf, None, "gl1060", coy=coy, args=args)
        self.df.mstAlw.show()

    def doAutoRct(self, date, memo, refno, amount, seq):
        # Automatic Postings and Allocations of unallocated entries
        # Search for rct transactions in gentrn and flag if found
        col = ["glt_seq"]
        sve = [("glt_cono", "=", self.conum), ("glt_acno", "=", self.bh.acc),
            ("glt_recon", "=", 0)]
        # Test with Refno and Amount
        whr = sve[:]
        whr.extend([("glt_refno", "=", refno), ("glt_tramt", "=", amount)])
        glt = self.sql.getRec(tables="gentrn", cols=col, where=whr)
        if not glt:
            # Test Date and Amount
            whr = sve[:]
            whr.extend([("glt_trdt", "=", date), ("glt_tramt", "=", amount)])
            glt = self.sql.getRec(tables="gentrn", cols=col, where=whr)
            if not glt:
                # Test Amount Only
                whr = sve[:]
                whr.append(("glt_tramt", "=", amount))
                glt = self.sql.getRec(tables="gentrn", cols=col, where=whr)
        if glt:
            if len(glt) > 1:
                # More than one transaction found, manually select which one
                get = self.doSelectEntry(date, memo, refno, amount, glt)
                if not get:
                    return "no"
                else:
                    glt = [glt[get]]
            # Set recon Flag on gentrn
            self.sql.updRec("gentrn", cols=["glt_recon"], data=[date/100],
                where=[("glt_seq", "=", glt[0][0])])
            self.doUpdateRct(seq)
            return
        if not amount:
            # Transaction amount is zero
            self.doUpdateRct(seq)
            return
        whr = [("grc_cono", "=", self.conum), ("grc_acno", "=", self.bh.acc)]
        if amount < 0:
            whr.append(("grc_rtn", "=", "P"))
        else:
            whr.append(("grc_rtn", "=", "D"))
        codes = self.sql.getRec(tables="genrcc", where=whr)
        if codes:
            # There are standard genrcc codes
            if self.bh.multi == "Y":
                self.curdt = date / 100
            else:
                self.curdt = self.bh.curdt
            for cono, acno, code, desc1, desc2, desc3, trtp, self.allcoy, self.accnum, acrs, achn, adrs, aage, self.vatcod in codes:
                self.alldet = CCD(desc1, "NA", 30).work
                if memo.find(desc1) != -1:
                    if desc2:
                        if memo.find(desc2) != -1:
                            if desc3:
                                if memo.find(desc3) != -1:
                                    found = True
                                else:
                                    found = False
                            else:
                                found = True
                        else:
                            found = False
                    else:
                        found = True
                else:
                    found = False
                if found:
                    # Check Control Details
                    if self.incoac:
                        chk = self.doChkLoadCtls()
                        if chk:
                            found = False
                            break
                    # Code matched, do automatic posting
                    self.trndet = self.alldet
                    if trtp == "P":
                        othrtn = 5
                        self.rtn = 2
                        self.vtyp = "I"
                    elif trtp == "D":
                        othrtn = 2
                        self.rtn = 6
                        self.vtyp = "O"
                    else:
                        self.rtn = 4
                        othrtn = 3
                        if amount < 0:
                            self.vtyp = "I"
                        else:
                            self.vtyp = "O"
                    self.trndat = date
                    self.trnref = refno
                    self.allamt = amount
                    no = self.doVatCalc(self.vatcod)
                    if no:
                        self.vatamt = 0
                    data = (self.conum, self.bh.acc, self.curdt, date,
                        self.rtn, self.trnref, self.bh.batno, self.allamt,
                        0.00, self.trndet, "", "", date/100, self.sysdtw, 0)
                    self.sql.insRec("gentrn", data=data)
                    amt = float(ASD(0) - ASD(self.allamt))
                    vat = float(ASD(0) - ASD(self.vatamt))
                    amt = float(ASD(amt) - ASD(vat))
                    self.restDebitCredit(amt, vat)
                    if acrs:
                        amt = float(ASD(0) - ASD(amt))
                        AgeAll(self.mf, system="crs", agetyp=aage,
                            agekey=[self.allcoy, acrs, othrtn, refno,
                            self.curdt, amt, 0])
                        data = [self.allcoy, acrs, othrtn, refno,
                            self.bh.batno, date, "", amt, 0, 0, self.curdt,
                            0, "", 0, self.trndet, "", "", self.sysdtw, 0]
                        self.sql.insRec("crstrn", data=data, unique="crt_ref1")
                    elif adrs:
                        AgeAll(self.mf, system="drs", agetyp=aage,
                            agekey=[self.allcoy, achn, adrs, othrtn, refno,
                            self.curdt, amt, 0])
                        data = [self.allcoy, achn, adrs, othrtn, refno,
                            self.bh.batno, date, "", amt, 0, self.curdt,
                            self.trndet, "", "", self.sysdtw, 0]
                        self.sql.insRec("drstrn", data=data, unique="drt_ref1")
                    self.doUpdateRct(seq)
                    break

    def doReadRct(self):
        whr = [("grt_cono", "=", self.conum), ("grt_acno", "=", self.bh.acc),
            ("grt_flag", "=", "N")]
        start = (self.bh.curdt * 100) + 1
        end = mthendDate(start)
        if self.bh.multi == "N":
            whr.append(("grt_date", "between", start, end))
        else:
            whr.append(("grt_date", "<=", end))
        return self.sql.getRec(tables="genrct", where=whr, order="grt_date")

    def doSelectEntry(self, date, memo, refno, amount, glt):
        # Multiple comparisons found, Select one
        state = self.df.disableButtons()
        self.mf.updateStatus("Select Transaction")
        date = CCD(date, "D1", 10)
        amount = CCD(amount, "SD", 13.2)
        titl = """Multiple Comparison Entries in the Ledger for:-

%s %s %s""" % (date.disp, amount.disp, memo)
        head = ("   Date   ", "Reference", "      Amount ", "Details")
        lines = []
        for seq in glt:
            acc = self.sql.getRec(tables="gentrn", cols=["glt_trdt",
                "glt_refno", "glt_tramt", "glt_desc"],
                where=[("glt_seq", "=", seq[0])], limit=1)
            dte = CCD(acc[0], "D1", 10)
            ref = CCD(acc[1], "Na", 9)
            amt = CCD(acc[2], "SD", 13.2)
            des = CCD(acc[3], "NA", 30)
            lines.append([dte.disp, ref.disp, amt.disp, des.disp])
        sc = SChoice(self.mf, scrn=self.df.mstFrame, titl=titl, head=head,
            data=lines, retn="I", mode="S")
        self.df.enableButtons(state=state)
        if sc.selection:
            return sc.selection[0]

    def doUpdateRct(self, seq):
        # Set genrct grt_flag to Y
        self.sql.updRec("genrct", cols=["grt_flag"], data=["Y"],
            where=[("grt_seq", "=", seq)])
        self.mf.dbm.commitDbase()

    def doSelections(self, widget=None):
        all = getSingleRecords(self.mf, "gentrn", ("glt_type", "glt_trdt",
            "glt_refno", "glt_tramt", "glt_desc"), head=("X", "T", "Trans-Date",
            "Reference", "Trans-Value", "Details"), where=[("glt_cono", "=",
            self.conum), ("glt_acno", "=", self.bh.acc), ("glt_recon", "=",
            0)], order="glt_trdt, glt_refno")
        if not all:
            return
        for rec in all:
            self.seq = rec[self.sql.gentrn_col.index("glt_seq")]
            self.updateRecon()

    def doTrnTyp(self, frt, pag, r, c, p, i, w):
        self.typ = w
        if not self.dorec:
            return "ok"
        chk = self.doCheckRecords()
        if chk:
            return chk
        if self.typ == "P":
            self.rtn = 2
        elif self.typ == "D":
            self.rtn = 6
        self.df.colf[1][1][8]["where"] = [("glt_cono", "=", self.conum),
            ("glt_acno", "=", self.bh.acc), ("glt_type", "=", self.rtn),
            ("glt_recon", "=", 0)]
        try:
            self.df.rs.selection = None
        except:
            pass

    def doCheckRecords(self):
        data = self.doReadRct()
        if data:
            showError(self.mf.body, "Import Error",
                "Unallocated Records in Previous Bank\n"\
                "           Import File\n\n"\
                "Please Select the 'Process Bank Data'\n"\
                "      Button to Process them")
            return "rf"

    def doTrnRef(self, frt, pag, r, c, p, i, w):
        if self.rtn == 3 and not w:
            a = self.sql.getRec(tables="gentrn", cols=["glt_refno"],
                where=[("glt_cono", "=", self.conum), ("glt_acno", "=",
                self.conacc),("glt_refno", "like", "PC_______")],
                group="glt_refno", order="glt_refno desc")
            auto = False
            for r in a:
                try:
                    w = "PC%07i" % (int(r[0][2:]) + 1)
                    auto = True
                    break
                except:
                    pass
            if not auto:
                w = "PC0000001"
            self.df.loadEntry(frt, pag, p, data=w)
        self.trnref = w
        if not self.dorec:
            return "ok"
        try:
            rs = self.df.rs.selection
        except:
            rs = None
        if not rs:
            wa = [("glt_type", "=", self.rtn), ("glt_refno", "=", self.trnref)]
            opt = {}
            for k in self.glt:
                if k == "where":
                    opt[k] = self.glt[k][:]
                    opt[k].extend(wa)
                elif type(self.glt[k]) == list:
                    opt[k] = self.glt[k][:]
                else:
                    opt[k] = self.glt[k]
            rs = self.df.selRecord(1, opt).selection
        if rs:
            self.recon = True
            if self.typ == "P":
                amt = float(ASD(0) - ASD(rs[4]))
            else:
                amt = rs[4]
            self.df.loadEntry("C", 1, p+1, data=rs[0])
            self.df.loadEntry("C", 1, p+2, data=amt)
            self.df.loadEntry("C", 1, p+3, data=rs[5])
            self.seq = rs[3]
            return "nd"
        else:
            self.recon = False

    def doTrnDat(self, frt, pag, r, c, p, i, w):
        if w < self.period[1][0] or w > self.period[2][0]:
            return "Invalid Date, Not in Financial Period"
        self.trndat = w

    def doTrnAmt(self, frt, pag, r, c, p, i, w):
        self.trnamt = w

    def doTrnDet(self, frt, pag, r, c, p, i, w):
        self.trndet = w
        self.df.colf[2][6][5] = w

    def doCoyNum(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=", w)])
        if not acc:
            return "Invalid Company"
        if self.incoac and w not in self.incoac:
            return "Invalid Company, No Intercompany Record 1"
        self.allcoy = w
        return self.doChkLoadCtls()

    def doChkLoadCtls(self):
        # Check for Intercompany Records
        if self.allcoy != self.conum:
            acc = self.sql.getRec(tables="genint", where=[("cti_cono", "=",
                self.allcoy), ("cti_inco", "=", self.conum)], limit=1)
            if not acc:
                return "Invalid Company, No Intercompany Record 2"
        # Get Company Details
        ctlmst = self.gc.getCtl("ctlmst", self.allcoy)
        if not ctlmst:
            return "rf"
        # Set Company VAT Default
        self.taxdf = ctlmst["ctm_taxdf"]
        if not self.taxdf:
            self.taxdf = "N"
        # Check for Integrated Systems
        mod = []
        for x in range(0, 21, 2):
            mod.append(ctlmst["ctm_modules"][x:x+2])
        # Load Ass Integration
        if mod.count("AR"):
            ctl = self.gc.getCtl("assctl", self.allcoy)
            if not ctl:
                return "rf"
            self.ass_gl = ctl["cta_glint"]
        else:
            self.ass_gl = "N"
        # Load Crs Integration
        if mod.count("CR"):
            ctl = self.gc.getCtl("crsctl", self.allcoy)
            if not ctl:
                return "rf"
            self.crs_gl = ctl["ctc_glint"]
        else:
            self.crs_gl = "N"
        # Load Drs Integration and Chain Store Flag
        if mod.count("DR"):
            ctl = self.gc.getCtl("drsctl", self.allcoy)
            if not ctl:
                return "rf"
            self.drs_gl = ctl["ctd_glint"]
            self.drs_ch = ctl["ctd_chain"]
        else:
            self.drs_gl = "N"
            self.drs_ch = "N"
        # Load Mem Integration
        if mod.count("ML"):
            ctl = self.gc.getCtl("memctl", self.allcoy)
            if not ctl:
                return "rf"
            self.mem_gl = ctl["mcm_glint"]
        else:
            self.mem_gl = "N"
        # Load Rtl Integration
        if mod.count("RT"):
            ctl = self.gc.getCtl("rtlctl", self.allcoy)
            if not ctl:
                return "rf"
            self.rtl_gl = ctl["ctr_glint"]
        else:
            self.rtl_gl = "N"
        # Check for Control Records e.g. VAT, Crs, Drs and Discounts
        self.ctlctl = self.gc.getCtl("ctlctl", self.allcoy)
        if not self.ctlctl:
            return "rf"
        ctls = ["vat_ctl"]
        if self.rtn in (2, 6, 7):
            if self.crs_gl == "Y":
                ctls.extend(["crs_ctl", "dis_rec"])
            if self.drs_gl == "Y":
                ctls.extend(["drs_ctl", "dis_all"])
            if self.mem_gl == "Y":
                ctls.append("mem_ctl")
                if not "dis_all" in ctls:
                    ctls.append("dis_all")
        if self.gc.chkRec(self.allcoy, self.ctlctl, ctls):
            return "rf"
        self.convat = self.ctlctl["vat_ctl"]
        # Get for Asset Accounts
        if self.ass_gl == "Y":
            self.assctl = []
            ass = self.sql.getRec(tables="assgrp", cols=["asg_assacc"],
                where=[("asg_cono", "=", self.allcoy)])
            for a in ass:
                self.assctl.append(a[0])
        # Get for Rental Accounts
        if self.rtl_gl == "Y":
            self.rtlctl = []
            rtl = self.sql.getRec(tables="rtlprm", cols=["rtp_rtlacc"],
                where=[("rtp_cono", "=", self.allcoy)])
            for a in rtl:
                self.rtlctl.append(a[0])

    def doAccNum(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables="genmst", where=[("glm_cono", "=",
            self.allcoy), ("glm_acno", "=", w)], limit=1)
        if not acc:
            return "Invalid Account"
        self.others = None
        self.othtot = 0
        self.vattot = 0
        dic = self.sql.genmst_dic
        if self.rtn in (2, 6) or self.dorec:
            if self.ass_gl == "Y" and w in self.assctl:
                self.others = "ass"
            elif self.crs_gl == "Y" and w == self.ctlctl["crs_ctl"]:
                self.others = "crs"
            elif self.drs_gl == "Y" and w == self.ctlctl["drs_ctl"]:
                self.others = "drs"
            elif self.mem_gl == "Y" and w == self.ctlctl["mem_ctl"]:
                self.others = "mem"
        if self.others:
            if self.rtn == 2:
                if self.others == "mem":
                    self.othrtn = 2
                else:
                    self.othrtn = 5
            else:
                if self.others == "mem":
                    self.othrtn = 5
                else:
                    self.othrtn = 2
        else:
            if self.ass_gl == "Y" and w in self.assctl:
                ctl = True
            elif self.crs_gl == "Y" and w == self.ctlctl["crs_ctl"]:
                ctl = True
            elif self.drs_gl == "Y" and w == self.ctlctl["drs_ctl"]:
                ctl = True
            elif self.mem_gl == "Y" and w == self.ctlctl["mem_ctl"]:
                ctl = True
            elif self.rtl_gl == "Y" and w in self.rtlctl:
                ctl = True
            else:
                ctl = False
            if ctl:
                cf = PwdConfirm(self.mf, conum=self.allcoy, system="GEN",
                    code="CtlAcc")
                if not cf.pwd or cf.flag == "no":
                    return "This Is An Integrated Control Account"
            elif acc[dic["glm_ind"][1]] == "N":
                cf = PwdConfirm(self.mf, conum=self.allcoy, system="GEN",
                    code="NoPost")
                if not cf.pwd or cf.flag == "no":
                    return "Postings Disallowed"
        self.accnum = w
        if not acc[dic["glm_vat"][1]]:
            if acc[dic["glm_type"][1]] == "P":
                self.taxgl = self.taxdf
            else:
                self.taxgl = "N"
        else:
            self.taxgl = acc[dic["glm_vat"][1]]
        self.df.loadEntry("C", 2, p+1, acc[dic["glm_desc"][1]])

    def doAllAmt(self, frt, pag, r, c, p, i, w):
        if not w:
            w = float(ASD(self.trnamt) - ASD(self.allocated))
            self.df.loadEntry("C", 2, p, w)
        self.allamt = w
        if self.others in ("drs", "crs", "mem"):
            self.vatcod = "N"
            self.vatamt = 0
            self.df.loadEntry("C", 2, p+1, self.vatcod)
            self.df.loadEntry("C", 2, p+2, self.vatamt)
            return "sk2"
        else:
            self.df.loadEntry("C", 2, p+1, self.taxgl)

    def doVatCod(self, frt, pag, r, c, p, i, w):
        no = self.doVatCalc(w)
        if no:
            return no
        self.df.loadEntry("C", 2, p+1, self.vatamt)
        if not self.vatamt:
            return "sk"

    def doVatCalc(self, code):
        vat = self.sql.getRec(tables="ctlvmf", cols=["vtm_new"],
            where=[("vtm_cono", "=", self.conum), ("vtm_code", "=", code)],
            limit=1)
        if not vat:
            return "Invalid V.A.T. Code"
        self.vatcod = code
        rate = float(vat[0])
        self.vatamt = round((self.allamt * rate / (rate + 100)), 2)

    def doVatAmt(self, frt, pag, r, c, p, i, w):
        if self.allamt < 0 and w > 0:
            self.vatamt = float(ASD(0) - ASD(w))
        elif self.allamt > 0 and w < 0:
            self.vatamt = float(ASD(0) - ASD(w))
        else:
            self.vatamt = w
        self.df.loadEntry(frt, pag, p, data=self.vatamt)

    def doAllDet(self, frt, pag, r, c, p, i, w):
        self.alldet = w

    def endPage1(self):
        self.cancel = False
        self.agecan = False
        self.batupd = False
        if self.dorec and self.recon:
            self.updateRecon()
            self.df.advanceLine(1)
        else:
            self.updateTables(1)
            self.updateBatch()
            if not self.trnamt:
                self.df.advanceLine(1)
            else:
                self.allocated = float(0.0)
                self.df.selPage("Allocation")
                self.df.loadEntry("T", 2, 0, data=self.trnamt)
                self.df.focusField("C", 2, 1)

    def doCancel(self, widget=None):
        if self.agecan:
            ok = "yes"
        else:
            ok = askQuestion(self.mf.body, head="Cancel",
                mess="Are You Certain You Want to Cancel This Entry?")
        if ok == "yes":
            self.cancel = True
            self.mf.dbm.rollbackDbase()
            if self.batupd:
                self.updateBatch(rev=True)
            for pag in range(2, (self.df.pag + 1)):
                self.df.clearFrame("C", pag)
            if self.df.pag > 2 or self.rctimp:
                self.mf.closeLoop()
            row = int((self.df.last[1][1] - 1) / self.df.colq[1])
            col = (row * self.df.colq[1]) + 1
            self.df.selPage("Transaction")
            self.df.focusField("C", 1, col)
        else:
            self.df.focusField(self.df.frt, self.df.pag, self.df.col)

    def updateRecon(self):
        self.sql.updRec("gentrn", cols=["glt_recon"], data=[self.bh.curdt],
            where=[("glt_seq", "=", self.seq)])
        self.mf.dbm.commitDbase()

    def endPage2(self):
        self.updateTables(2)
        if self.others == "crs":
            self.page3 = True
            self.df.selPage("Creditors")
            self.df.loadEntry("T", self.crspag, 0, data=self.allamt)
            self.df.focusField("C", self.crspag, 1)
            self.mf.startLoop()
            self.page3 = False
            if self.cancel:
                if self.rctimp:
                    self.mf.closeLoop()
                return
            self.df.clearFrame("C", self.crspag)
            self.df.selPage("Allocation")
        elif self.others == "drs":
            self.page4 = True
            if self.drs_ch == "N":
                self.df.colf[self.drspag][0][1] = "OUI"
                self.chain = 0
            else:
                self.df.colf[self.drspag][0][1] = "IUI"
            self.df.selPage("Debtors")
            self.df.loadEntry("T", self.drspag, 0, data=self.allamt)
            self.df.focusField("C", self.drspag, 1)
            self.mf.startLoop()
            self.page4 = False
            if self.cancel:
                if self.rctimp:
                    self.mf.closeLoop()
                return
            self.df.clearFrame("C", self.drspag)
            self.df.selPage("Allocation")
        elif self.others == "mem":
            self.page5 = True
            self.df.selPage("Members")
            self.df.loadEntry("T", self.mempag, 0, data=self.allamt)
            self.df.focusField("C", self.mempag, 1)
            self.mf.startLoop()
            self.page5 = False
            if self.cancel:
                if self.rctimp:
                    self.mf.closeLoop()
                return
            self.df.clearFrame("C", self.mempag)
            self.df.selPage("Allocation")
        elif self.others == "ass":
            self.page6 = True
            self.df.selPage("Assets")
            self.allvat = self.vatamt
            bal = float(ASD(self.allamt) - ASD(self.vatamt))
            self.df.loadEntry("T", self.asspag, 0, data=bal)
            self.df.focusField("C", self.asspag, 1)
            self.mf.startLoop()
            self.page5 = False
            if self.cancel:
                if self.rctimp:
                    self.mf.closeLoop()
                return
            self.df.clearFrame("C", self.asspag)
            self.df.selPage("Allocation")
        self.allocated = float(ASD(self.allocated) + ASD(self.allamt))
        if self.allocated == self.trnamt:
            self.df.clearFrame("C", 2)
            self.df.selPage("Transaction")
            self.df.advanceLine(1)
            if self.rctimp:
                self.rctupd = True
                self.mf.closeLoop()
            else:
                self.mf.dbm.commitDbase()
        else:
            bal = float(ASD(self.trnamt) - ASD(self.allocated))
            self.df.loadEntry("T", 2, 0, data=bal)
            self.df.advanceLine(2)

    def updateTables(self, pag):
        if self.bh.multi == "Y":
            self.curdt = self.trndat / 100
        else:
            self.curdt = self.bh.curdt
        if self.rtn == 1:                          # Sales
            self.rflag = 0
            self.vtyp = "O"
            self.debit(pag)
        elif self.rtn == 2:                        # Payments
            self.conacc = self.bh.acc
            if self.dorec:
                self.rflag = self.bh.curdt
            elif not self.trnamt:
                self.rflag = self.curdt
            else:
                self.rflag = 0
            self.vtyp = "I"
            self.credit(pag)
        elif self.rtn == 3:                        # Petty Cash
            self.rflag = 0
            self.vtyp = "I"
            self.credit(pag)
        elif self.rtn == 4:                        # Journal Entries
            pass
        elif self.rtn == 5:                        # Purchases
            self.rflag = 0
            self.vtyp = "I"
            self.credit(pag)
        elif self.rtn == 6:                        # Receipts
            self.conacc = self.bh.acc
            if self.dorec:
                self.rflag = self.bh.curdt
            elif not self.trnamt:
                self.rflag = self.curdt
            else:
                self.rflag = 0
            self.vtyp = "O"
            self.debit(pag)

    def debit(self, pag):
        if pag == 1:
            amt = self.trnamt
            data = (self.conum, self.conacc, self.curdt, self.trndat,
                self.rtn, self.trnref, self.bh.batno, amt, 0.00,
                self.trndet, "", "N", self.rflag, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)
        else:
            amt = float(ASD(0) - (ASD(self.allamt) - ASD(self.vatamt)))
            vat = float(ASD(0) - ASD(self.vatamt))
            self.restDebitCredit(amt, vat)

    def credit(self, pag):
        if pag == 1:
            amt = float(ASD(0) - ASD(self.trnamt))
            data = (self.conum, self.conacc, self.curdt, self.trndat,
                self.rtn, self.trnref, self.bh.batno, amt, 0.00,
                self.trndet, "", "N", self.rflag, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)
        else:
            amt = float(ASD(self.allamt) - ASD(self.vatamt))
            vat = self.vatamt
            self.restDebitCredit(amt, vat)

    def restDebitCredit(self, amt, vat):
        data = (self.allcoy, self.accnum, self.curdt, self.trndat, self.rtn,
            self.trnref, self.bh.batno, amt, vat, self.alldet, self.vatcod,
            "", 0, self.sysdtw, 0)
        self.sql.insRec("gentrn", data=data)
        if self.allcoy != self.conum:
            acc = self.sql.getRec(tables="genint", cols=["cti_acno"],
                where=[("cti_cono", "=", self.conum), ("cti_inco", "=",
                self.allcoy)], limit=1)[0]
            val = float(ASD(amt) + ASD(vat))
            data = (self.conum, acc, self.curdt, self.trndat, self.rtn,
                self.trnref, self.bh.batno, val, 0.00, self.alldet,
                self.vatcod, "", 0, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)
            acc = self.sql.getRec(tables="genint", cols=["cti_acno"],
                where=[("cti_cono", "=", self.allcoy), ("cti_inco", "=",
                self.conum)], limit=1)[0]
            val = float(ASD(0) - ASD(amt) - ASD(vat))
            data = (self.allcoy, acc, self.curdt, self.trndat, self.rtn,
                self.trnref, self.bh.batno, val, 0.00, self.alldet,
                self.vatcod, "", 0, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)
        if self.vatcod:
            if vat:
                data = (self.allcoy, self.convat, self.curdt, self.trndat,
                    self.rtn, self.trnref, self.bh.batno, vat, 0.00,
                    self.alldet, self.vatcod, "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
            data = (self.allcoy, self.vatcod, self.vtyp, self.curdt, "G",
                self.rtn, self.bh.batno, self.trnref, self.trndat,
                self.accnum, self.alldet, amt, vat, self.sysdtw, 0, 0)
            self.sql.insRec("ctlvtf", data=data)

    def doCrsAcc(self, frt, pag, r, c, p, i, w):
        self.othacno = w
        acc = self.sql.getRec(tables="crsmst", cols=["crm_name"],
            where=[("crm_cono", "=", self.allcoy), ("crm_acno", "=",
            self.othacno)], limit=1)
        if not acc:
            ok = askQuestion(screen=self.mf.body, head="New Creditor",
                mess="Account does not exist, Create?")
            if ok == "no":
                return "rf"
            self.doNewCrs()
            acc = self.sql.getRec(tables="crsmst", cols=["crm_name"],
                where=[("crm_cono", "=", self.allcoy), ("crm_acno", "=",
                self.othacno)], limit=1)
            if not acc:
                return "rf"
        self.df.loadEntry("C", pag, p+1, data=acc[0])
        self.allref = self.trnref

    def doCrsRef(self, frt, pag, r, c, p, i, w):
        self.othref = w

    def doCrsDis(self, frt, pag, r, c, p, i, w):
        self.othdis = w

    def doCrsAmt(self, frt, pag, r, c, p, i, w):
        self.othamt = w
        if not self.othamt:
            self.othamt = float(ASD(self.allamt) - ASD(self.othtot))
        self.df.loadEntry(frt, pag, p, data=self.othamt)

    def doDrsChn(self, frt, pag, r, c, p, i, w):
        self.chain = w

    def doDrsAcc(self, frt, pag, r, c, p, i, w):
        self.othacno = w
        acc = self.sql.getRec(tables="drsmst", cols=["drm_name"],
            where=[("drm_cono", "=", self.allcoy), ("drm_chain", "=",
            self.chain), ("drm_acno", "=", self.othacno)], limit=1)
        if not acc:
            ok = askQuestion(screen=self.mf.body, head="New Debtor",
                mess="Account does not exist, Create?")
            if ok == "no":
                return "rf"
            self.doNewDrs()
            acc = self.sql.getRec(tables="drsmst", cols=["drm_name"],
                where=[("drm_cono", "=", self.allcoy), ("drm_chain", "=",
                self.chain), ("drm_acno", "=", self.othacno)], limit=1)
            if not acc:
                return "rf"
        self.df.loadEntry("C", pag, p+1, data=acc[0])
        self.allref = self.trnref

    def doDrsRef(self, frt, pag, r, c, p, i, w):
        self.othref = w

    def doDrsDis(self, frt, pag, r, c, p, i, w):
        self.othdis = w

    def doDrsAmt(self, frt, pag, r, c, p, i, w):
        self.othamt = w
        if not self.othamt:
            self.othamt = float(ASD(self.allamt) - ASD(self.othtot))
        self.df.loadEntry(frt, pag, p, data=self.othamt)

    def doMemAcc(self, frt, pag, r, c, p, i, w):
        self.othacno = w
        acc = self.sql.getRec(tables="memmst", cols=["mlm_surname",
            "mlm_names"], where=[("mlm_cono", "=", self.allcoy),
            ("mlm_memno", "=", self.othacno)], limit=1)
        if not acc:
            return "Invalid Member Number"
        self.name = "%s, %s" % (acc[0], acc[1])
        self.df.loadEntry("C", pag, p+1, data=self.name)
        self.allref = self.trnref

    def doMemDis(self, frt, pag, r, c, p, i, w):
        self.othdis = w

    def doMemAmt(self, frt, pag, r, c, p, i, w):
        self.othamt = w
        if not self.othamt:
            self.othamt = float(ASD(self.allamt) - ASD(self.othtot))
        self.df.loadEntry(frt, pag, p, data=self.othamt)

    def endCrsDrsMem(self):
        if      (self.others == "crs" and self.rtn == 2) or \
                (self.others == "drs" and self.rtn == 6) or \
                (self.others == "mem" and self.rtn == 6):
            self.tramt = float(ASD(0.0) - ASD(self.othamt))
            self.dsamt = float(ASD(0.0) - ASD(self.othdis))
        else:
            self.tramt = self.othamt
            self.dsamt = self.othdis
        # Age Transaction
        state = self.df.disableButtons()
        self.mf.updateStatus("Choose an Ageing Option")
        if len(self.but) == 5:
            self.beg = 0
        elif len(self.but) == 8:
            self.beg = 4
        for b in range(self.beg, (self.beg + 4)):
            exec "self.df.setWidget(self.df.B%d, 'normal')" % b
        self.mf.startLoop()
        self.df.enableButtons(state=state)
        if self.agecan:
            self.doCancel()
            return
        self.othtot = float(ASD(self.othtot) + ASD(self.othamt))
        if self.others == "crs":
            # Creditors Ledger Transaction
            data = [self.allcoy, self.othacno, self.othrtn, self.allref,
                self.bh.batno, self.trndat, self.othref, self.tramt,
                self.vatamt, 0, self.curdt, 0, "Y", 0, self.alldet,
                self.vatcod, "", self.sysdtw, 0]
            self.sql.insRec("crstrn", data=data, unique="crt_ref1")
            if self.dsamt:
                # Creditors Ledger Transaction
                data[2] = 7
                data[7] = self.dsamt
                data[8] = 0
                self.sql.insRec("crstrn", data=data, unique="crt_ref1")
                # General Ledger Discount Received
                data = (self.allcoy, self.ctlctl["dis_rec"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, self.dsamt,
                    0, self.alldet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
                # General Ledger Creditors Control
                dis = float(ASD(0.0) - ASD(self.dsamt))
                data = (self.allcoy, self.ctlctl["crs_ctl"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, dis, 0,
                    self.trndet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
            if self.othtot != self.allamt:
                bal = float(ASD(self.allamt) - ASD(self.othtot))
                self.df.loadEntry("T", self.crspag, 0, data=bal)
                self.df.advanceLine(3)
            else:
                self.mf.closeLoop()
        elif self.others == "drs":
            # Debtors Ledger Transaction
            data = [self.allcoy, self.chain, self.othacno, self.othrtn,
                self.allref, self.bh.batno, self.trndat, self.othref,
                self.tramt, self.vatamt, self.curdt, self.alldet,
                self.vatcod, "", self.sysdtw, 0]
            self.sql.insRec("drstrn", data=data, unique="drt_ref1")
            if self.dsamt:
                # Debtors Ledger Transaction
                data[3] = 7
                data[8] = self.dsamt
                data[9] = 0
                self.sql.insRec("drstrn", data=data, unique="drt_ref1")
                # General Ledger Debtors Control
                data = (self.allcoy, self.ctlctl["drs_ctl"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, self.dsamt,
                    0, self.alldet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
                # General Ledger Discount Allowed
                dis = float(ASD(0.0) - ASD(self.dsamt))
                data = (self.allcoy, self.ctlctl["dis_all"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, dis, 0,
                    self.trndet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
            if self.othtot != self.allamt:
                bal = float(ASD(self.allamt) - ASD(self.othtot))
                self.df.loadEntry("T", self.drspag, 0, data=bal)
                self.df.advanceLine(4)
            else:
                self.mf.closeLoop()
        elif self.others == "mem":
            # Members Ledger Transaction
            data = [self.allcoy, self.othacno, self.othrtn, self.allref,
                self.bh.batno, self.trndat, self.tramt, self.vatamt,
                self.curdt, "", 0, self.alldet, self.vatcod, "", self.sysdtw, 0]
            self.sql.insRec("memtrn", data=data, unique="mlt_refno")
            if self.dsamt:
                # Members Ledger Transaction
                data[2] = 6
                data[6] = self.dsamt
                data[7] = 0
                self.sql.insRec("memtrn", data=data, unique="mlt_refno")
                # General Ledger Members Control
                data = (self.allcoy, self.ctlctl["mem_ctl"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, self.dsamt,
                    0, self.alldet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
                # General Ledger Discount Allowed
                dis = float(ASD(0.0) - ASD(self.dsamt))
                data = (self.allcoy, self.ctlctl["dis_all"], self.curdt,
                    self.trndat, 4, self.allref, self.bh.batno, dis, 0,
                    self.trndet, "", "", 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
            if self.othtot != self.allamt:
                bal = float(ASD(self.allamt) - ASD(self.othtot))
                self.df.loadEntry("T", self.mempag, 0, data=bal)
                self.df.advanceLine(4)
            else:
                self.mf.closeLoop()

    def doAgeNormal(self, widget=None):
        self.doAgeing("N")

    def doAgeHistory(self, widget=None):
        self.doAgeing("H")

    def doAgeAuto(self, widget=None):
        self.doAgeing("O")

    def doAgeCurrent(self, widget=None):
        self.doAgeing("C")

    def doAgeing(self, atype):
        # Disable All Ageing Buttons
        for b in range(self.beg, (self.beg + 4)):
            exec "self.df.setWidget(self.df.B%d, 'disabled')" % b
        self.mf.updateStatus("Select a Transaction to Allocate Against")
        if self.others == "crs":
            age = AgeAll(self.mf, system="crs", agetyp=atype,
                agekey=[self.allcoy, self.othacno, self.othrtn, self.allref,
                self.curdt, self.tramt, self.dsamt])
        elif self.others == "drs":
            age = AgeAll(self.mf, system="drs", agetyp=atype,
                agekey=[self.allcoy, self.chain, self.othacno, self.othrtn,
                self.allref, self.curdt, self.tramt, self.dsamt])
        elif self.others == "mem":
            age = AgeAll(self.mf, system="mem", agetyp=atype,
                agekey=[self.allcoy, self.othacno, self.othrtn, self.allref,
                self.curdt, self.tramt, self.dsamt])
        self.agecan = age.cancel
        self.mf.closeLoop()

    def doAssGrp(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables=["assgrp"], where=[("asg_cono", "=",
            self.allcoy), ("asg_group", "=", w)], limit=1)
        if not acc:
            return "Invalid Asset Group (1)"
        self.assgrp = w
        self.depcod = acc[self.sql.assgrp_col.index("asg_depcod")]
        self.assacc = acc[self.sql.assgrp_col.index("asg_assacc")]
        self.depacc = acc[self.sql.assgrp_col.index("asg_depacc")]
        self.expacc = acc[self.sql.assgrp_col.index("asg_expacc")]
        if self.assacc != self.accnum:
            return "Invalid Asset Group (2)"

    def doAssCod(self, frt, pag, r, c, p, i, w):
        self.asscod = w
        self.assbuy = 0
        self.asssel = 0
        cols = ["asm_desc"]
        whr = [("asm_cono", "=", self.allcoy), ("asm_group", "=", self.assgrp),
            ("asm_code", "=", w)]
        acc = self.sql.getRec(tables=["assmst"], where=whr, cols=cols, limit=1)
        if not acc:
            ok = askQuestion(screen=self.mf.body, head="New Asset",
                mess="Asset does not exist, Create?")
            if ok == "no":
                return "rf"
            self.doNewAsset()
            acc = self.sql.getRec(tables=["assmst"], where=whr, cols=cols,
                limit=1)
            if not acc:
                return "rf"
            self.newass = True
        else:
            self.newass = False
        pur = self.sql.getRec(tables=["asstrn"], cols=["ast_date"],
            where=[("ast_cono", "=", self.allcoy), ("ast_group", "=",
            self.assgrp), ("ast_code", "=", w)], limit=1)
        if pur:
            self.assbuy = pur[0]
        sel = self.sql.getRec(tables=["asstrn"], cols=["ast_date"],
            where=[("ast_cono", "=", self.allcoy), ("ast_group", "=",
            self.assgrp), ("ast_code", "=", w), ("ast_mtyp", "=", 5)],
            limit=1)
        if sel:
            self.asssel = sel[0]
        self.df.loadEntry("C", pag, p+1, data=acc[0])
        if self.newass:
            self.othmov = 1
            self.df.loadEntry(frt, pag, p+2, data=self.othmov)
            return "sk2"

    def doAssMov(self, frt, pag, r, c, p, i, w):
        if type(w) == list:
            w = w[0] + 1
        # Do some tests to see if not new again or already sold etc.
        if self.rtn == 2 and w not in (1, 2):
            return "Invalid Choice For Payment"
        if self.rtn == 6 and w not in (5,):
            return "Invalid Choice For Recipt"
        if w == 1 and self.assbuy:
            return "Asset Already Purchased"
        if w in (2, 5) and self.asssel:
            return "Asset Already Sold"
        if w in (2, 5) and not self.assbuy:
            return "Asset Not Yet Purchased"
        self.df.loadEntry(frt, pag, p, data=w)
        self.othmov = w

    def doAssAmt(self, frt, pag, r, c, p, i, w):
        self.othamt = w
        if not self.othamt:
            self.othamt = float(
                ASD(self.allamt) - ASD(self.vatamt) - ASD(self.othtot))
            self.othvat = float(ASD(self.vatamt) - ASD(self.vattot))
        else:
            self.othvat = float(ASD(self.allamt) - ASD(self.vatamt))
            self.othvat = round(((self.othamt / self.othvat) * self.vatamt), 2)
        self.df.loadEntry(frt, pag, p, data=self.othamt)
        self.df.loadEntry(frt, pag, p+1, data=self.trndet)

    def endAssets(self):
        if self.rtn == 2:
            self.othrtn = 1
            self.tramt = self.othamt
            self.vtamt = self.othvat
        else:
            self.othrtn = 2
            self.tramt = float(ASD(0.0) - ASD(self.othamt))
            self.vtamt = float(ASD(0.0) - ASD(self.othvat))
        self.othtot = float(ASD(self.othtot) + ASD(self.othamt))
        self.vattot = float(ASD(self.vattot) + ASD(self.othvat))
        tot = float(ASD(self.allamt) - ASD(self.vatamt))
        if self.othtot == tot and self.vattot != self.vatamt:
            self.othvat = float(
                ASD(self.othvat) + ASD(self.vatamt) - ASD(self.vattot))
        # Asset Register Transaction
        data = [self.allcoy, self.assgrp, self.asscod, self.othrtn,
            self.trnref, self.bh.batno, self.trndat, self.othmov,
            self.tramt, self.tramt, self.othvat, self.curdt,
            self.alldet, self.vatcod, "", self.sysdtw, 0]
        self.sql.insRec("asstrn", data=data)
        if self.othtot != tot:
            bal = float(
                ASD(self.allamt) - ASD(self.vatamt) - ASD(self.othtot))
            self.df.loadEntry("T", self.asspag, 0, data=bal)
            self.df.advanceLine(5)
        else:
            self.mf.closeLoop()

    def doNewDrs(self, widget=None):
        callModule(self.mf, self.df, "dr1010", coy=(self.conum, self.conam),
            period=None, user=self.capnm, args=(self.chain, self.othacno))

    def doNewCrs(self, widget=None):
        callModule(self.mf, self.df, "cr1010", coy=(self.conum, self.conam),
            period=None, user=self.capnm, args=self.othacno)

    def doNewAsset(self):
        tit = ("Create New Asset", )
        r1s = (("Straight Line","S"),("Diminishing Balance","D"))
        self.fld = (
            (("T",0,2,0),"INA",30,"Description","Description",
                "","N",None,None,None,("notblank",)),
            (("T",0,3,0),"INa",3,"Dep Code","Depreciation Code",
                self.depcod,"N",self.doDepCode,None,None,("notblank",)),
            (("T",0,1,0),"ONA",34,""))
        tnd = ((self.doNewEnd,"N"), )
        txt = (self.doNewXit, )
        state = self.df.disableButtons()
        self.na = TartanDialog(self.mf, newscn=(0,1), tops=True, title=tit,
            eflds=self.fld, tend=tnd, txit=txt)
        self.mf.startLoop()
        self.df.enableButtons(state=state)

    def doDepCode(self, frt, pag, r, c, p, i, w):
        chk = self.sql.getRec(tables="assdep", cols=["asd_desc"],
            where=[("asd_cono", "=", self.conum), ("asd_code", "=", w)],
            limit=1)
        if not chk:
            return "Invalid Depreciation Code"
        self.na.loadEntry(frt, pag, p+1, data=chk[0])

    def doNewEnd(self):
        dat = [self.allcoy, self.assgrp, self.asscod]
        for x in range(0, len(self.na.t_work[0][0])):
            dat.append(self.na.t_work[0][0][x])
        self.sql.insRec("assmst", data=dat)
        self.doNewXit()

    def doNewXit(self):
        self.na.closeProcess()
        self.mf.closeLoop()

    def updateBatch(self, rev=False):
        if rev:
            self.bh.batqty = self.bh.batqty - 1
            self.bh.batval = float(ASD(self.bh.batval) - ASD(self.trnamt))
        else:
            self.batupd = True
            self.bh.batqty = self.bh.batqty + 1
            self.bh.batval = float(ASD(self.bh.batval) + ASD(self.trnamt))
        self.df.loadEntry("T", 0, 0, data=self.bh.batqty)
        self.df.loadEntry("T", 0, 1, data=self.bh.batval)

    def exitPage1(self, widget=None):
        if not self.rctimp:
            self.df.closeProcess()
            if not self.dorec:
                self.bh.doBatchTotal()
        self.mf.closeLoop()

    def exitPage2(self):
        self.df.focusField("C", 2, self.df.col)

    def exitPage3(self):
        self.df.focusField("C", self.crspag, self.df.col)

    def exitPage4(self):
        self.df.focusField("C", self.drspag, self.df.col)

    def exitPage5(self):
        self.df.focusField("C", self.mempag, self.df.col)

    def exitPage6(self):
        self.df.focusField("C", self.asspag, self.df.col)
