#!/usr/bin/env python
# -*- coding: latin-1 -*-

"""
SYNOPSIS
    This is the Main Tartan module.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import base64, cStringIO, getpass, getopt, glob, os, sys, time
from TartanClasses import CCD, FileDialog, MainFrame, PwdConfirm, ScrollText
from TartanClasses import Sql, TarBckRes, TartanCalendar, TartanConfig
from TartanClasses import TartanDialog, TartanMenu, TartanUser, RunCommand
from tartanFunctions import aboutTartan, askQuestion, callModule, dateDiff
from tartanFunctions import doGtkImport, ftpDownload, getPeriods, getPrgPath
from tartanFunctions import loadRcFile, projectDate, sendMail, showError
from tartanFunctions import showException, showWarning

# Ubuntu Unity uses the Global Menu which breaks Tartan's Menu
if sys.platform == "linux2":
    os.environ["UBUNTU_MENUPROXY"] = "0"

class msc000:
    def __init__(self, version, opts, args):
        self.current = [version, "%s.%s.%s" % version]
        self.mf = None
        help = False
        self.conam = ""
        self.conum = None
        self.rday = None
        self.dbcmd = None
        self.display = True
        self.image = True
        self.maxi = False
        self.menfle = None
        self.module = None
        self.pernum = None
        self.check = True
        if sys.platform == "win32":
            self.output = True
        else:
            self.output = False
        self.rcfile = None
        self.rcdir = None
        self.rtn = None
        self.sysfle = None
        self.user = {}
        self.version = False
        self.fixed = True
        self.debug = False
        self.bpwd = None
        self.force = False
        self.tables = None
        self.rload = False
        self.reformat = False
        self.script = None
        self.console = False
        for o, v in opts:
            if o in ("-h", "--help"):
                help = True
            elif o == "-c":
                self.conum = int(v)
            elif o in ("-d", "--debug"):
                self.debug = True
            elif o == "-f":
                self.pernum = int(v)
            elif o in ("-i", "--image"):
                self.image = False
            elif o == "-l":
                self.maxi = True
            elif o == "-m":
                self.menfle = v
            elif o in ("-n", "--nocheck"):
                self.check = False
            elif o in ("-o", "--output"):
                if sys.platform == "win32":
                    self.output = False
                else:
                    self.output = True
            elif o == "-p":
                self.module = v
                self.fixed = False
            elif o == "-r":
                self.rcfile = v
            elif o == "-R":
                self.rcdir = os.path.abspath(os.path.normpath(v))
                if not os.path.isdir(self.rcdir):
                    print "Invalid Rc Directory: %s" % self.rcdir
                    self.doExit()
            elif o == "-s":
                self.sysfle = v
            elif o == "-t":
                self.rtn = int(v)
            elif o == "-u":
                u = v.split(".")
                self.user["name"] = u[0]
                if len(u) == 2:
                    self.user["pwd"] = u[1]
                else:
                    self.user["pwd"] = ""
                self.user["last"] = 0
                self.user["coy"] = ""
                self.user["lvl"] = 0
            elif o in ("-v", "--version"):
                self.version = True
            elif o == "-w":
                self.fixed = False
            elif o == "-x":
                if not args:
                    self.dbcmd = v # Execute sql commands in a file
                else:
                    self.dbcmd = []
                    for a in args:
                        if a == ";":
                            self.dbcmd.append(v)
                            v = ""
                            continue
                        if not v:
                            v = a
                        else:
                            v = "%s %s" % (v, a.replace("\\", ""))
                    self.dbcmd.append(v)
            elif o == "--bpwd":
                self.bpwd = v
            elif o == "--nodisplay":
                self.display = False
            elif o == "--reload":
                self.rload = True
            elif o == "--force":
                self.force = True
            elif o == "--reformat":
                self.reformat = True
            elif o == "--tables":
                self.tables = v
            elif o == "--script":
                self.script = v
            elif o == "--console":
                self.console = True
        if self.output:
            ############################################
            # Redirect stdout                          #
            ############################################
            for pid in range(1000):
                if self.debug:
                    name = "debug_%s.txt" % pid
                else:
                    name = "stdout_%s.txt" % pid
                self.stdout = os.path.join(getPrgPath(), name)
                if not os.path.exists(self.stdout):
                    sys.stdout = open(self.stdout, "w")
                    break
            ############################################
        if self.console:
            # RAW Python Console - Execute commands with F5 or Ctl-E
            gui, gtk, gobject, pango = doGtkImport()
            if gui:
                from PyGTKShell.RawConsole import Window
                from PyGTKShell.RawConsole import RawConsoleWithIntro
                win = Window()
                win(RawConsoleWithIntro())
                win.set_default_size(600,450)
                win.connect("destroy", gtk.main_quit)
                gtk.main()
            self.doExit()
        if self.script:
            exec "import %s" % self.script
            self.doExit()
        if help:
            print """
Usage:      ms0000.py [options]

General Options:
            -d, --debug             Create the TARTANDB environ variable
            -h, --help              This Help Message
            -i, --image             Do not display Tartan Image
            -l                      Start with a Maximized Window
            -m [file]               Menu file to import e.g. stdmen
            -n, --nocheck           Do not do ctlsys, ctlmst, version checking
            -o, --output            Toggle stdout redirection to stdout.txt
                                        Linux   defaults to False
                                        Windows defaults to True
            -r [path]               Path of Tartan RC File to use
            -R [directory]          Directory of Available Tartan RC Files
            -s [system]             System (.sys) file e.g. tartan
            -v, --version           Display Version Number Only
            -w                      Variable windows

Direct Options:
            -c [nnn]                Company Number
            -f [n]                  Financial Period Code, defaults to latest
            -p [program]            Program Module e.g. gl1010
            -t [n]                  Transaction Code if Applicable e.g. 1
            -u [user[.password]]    User name and password e.g. user.password

File Format Updates:
            --nodisplay             Do Not Display Dialog
            --force                 Force the update irrespective of version
            --reformat              Reformat tables
            --tables                Tables to reformat, defaults to all

Backup Function:
            --nodisplay             Do Not Display Dialog
            --bpwd [password]       The Backup Password

Execute SQL Commands:
            -u [user[.password]]    User name and password e.g. user.password
            -x [string]             A path to the file containing the commands
                                    or a string of commands seperated by ' ; '
                                    in which case this must be the final option
                                    used because all options after this one
                                    will be treated as sql commands. In linux,
                                    if there are any ;'s or *'s in the string,
                                    the whole string must be enclosed in double
                                    quotes and the ;'s and *'s must be escaped
                                    i.e. \; or \*. In addition all single
                                    quotes, excepting those enclosing the whole
                                    string, must also be escaped
                                    e.g. "Select ... where xyz like \\'%ABC%\\'"

Extended Options:
            --console               Open a RAW Console
            --reload                Re-import modules
            --script [name]         The name of the script to execute. Please
                                    remember that the script must be in the
                                    PYTHONPATH variable

ENVIRONMENT VARIABLES:
            PHOTODIR                Alternate Members Photo Directory
            TARTANDB                Set the DEBUG feature on
            TARTANRC                Alternate tartanrc file
            """
            self.doExit()
        if self.version:
            print "Tartan Version: %s" % self.current[1]
            self.doExit()
        if self.debug:
            os.environ["TARTANDB"] = "1"
            import trace
            igm = [
                "TartanClasses", "tartanFunctions", "rl_codecs", "rl_config",
                "rl_temp", "canvas", "pdfdoc", "ttfonts", "_fontdata", "utils",
                "pdfmetrics", "colors", "textobject", "fonts", "pdfutils",
                "units", "pagesizes", "logger", "rltempfile", "pathobject",
                "pdfgeom", "boxstuff", "__init__"]
            if sys.platform == "win32" and os.path.basename(sys.path[0]) == "library.zip":
                igd = []
                for mod in sys.modules:
                    igm.append(mod)
            else:
                igd = [sys.prefix, sys.exec_prefix]
            self.tracer = trace.Trace(ignoredirs=igd, ignoremods=igm,
                trace=1, count=0)
        # Create all variables
        self.setVariables()
        # if rcdir, select rcfile to use
        if self.rcdir:
            dialog = FileDialog(title="Select rcfile", initd=self.rcdir)
            self.rcfile = dialog.askopenfilename()
            if not self.rcfile:
                self.doExit()
        # Load the rcfile if it exists else run TartanConfig to create one
        self.rcdic = None
        while not self.rcdic:
            self.rcdic = loadRcFile(self.rcfile)
            if self.rcdic == "error":
                self.doExit()
            elif not self.rcdic:
                if not self.display:
                    print "No tartanrc File and display=False, Aborting"
                    self.doExit()
                cfg = TartanConfig(rcfile=self.rcfile, level=9)
                if not cfg.rcfile:
                    self.doExit()
                self.rcfile = cfg.rcfile
        # Create the MainFrame class with or without a window
        self.mf = MainFrame(rcdic=self.rcdic, display=self.display,
            fixed=self.fixed, maxi=self.maxi)
        if self.display and not self.mf.body:
            self.doExit()
        # Check if ti1010 is to be run i.e. Convert and Import from Visual COBOL
        if self.module == "ti1010":
            err = self.doDirect()
            if err:
                print "\nDirect Error: %s\n" % err
            self.doExit()
        # Check if the database has been created by trying to open it
        err = self.doOpenDbase(check=True)
        if err:
            self.doExit()
        # Check if DBCreate is to be run
        if self.module == "DBCreate":
            err = self.doDirect()
            if err:
                print "\nDirect Error: %s\n" % err
            self.doExit()
        if self.check:
            # Check for ctlsys and if missing call msc110
            err = self.doCheckSys()
            if err:
                self.doExit()
            # Check for ctlmst and if missing call ms1010
            err = self.doCheckMst()
            if err:
                self.doExit()
            # Check if registered and, if not, register and advise
            smtp = self.getSMTP(regdte=True)
            if smtp and smtp[2] and not smtp[-1]:
                try:
                    sendMail(smtp[2], smtp[3],
                        ["register@tartan.co.za"],
                        "%s - Registration" % smtp[1],
                        "%s - Registration" % smtp[1],
                        timeout=10)
                    t = time.localtime()
                    sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
                    sql = Sql(self.mf.dbm, "ctlsys", prog="msc000", errs=False)
                    sql.updRec("ctlsys", cols=["sys_regdte"], data=[sysdtw])
                    self.mf.dbm.commitDbase()
                except:
                    pass
        # Check if a correct user name and password have been entered
        if self.user:
            self.doReadCheckUser(self.user["name"], pwd=self.user["pwd"],
                pwdchk=True)
            if not self.user:
                print "Invalid User Details"
                self.doExit()
            else:
                self.updateCtlLog("login", "msc000")
        else:
            self.doLogin()
            if not self.user:
                self.doExit()
        self.mf.dbm.closeDbase()
        if self.user and self.dbcmd:
            if self.lvl == 9:
                err = self.doDbCommand()
            else:
                err = "Invalid Security Level"
            if err:
                showError(self.mf.body, "Data Base Error",
                "\nDbCommand Error: %s\n" % err)
        elif self.user and self.module:
            err = self.doDirect()
            if err:
                print "\nDirect Error: %s\n" % err
        else:
            self.doProcess()
        self.doExit(sysexit=False)

    def doExit(self, sysexit=True):
        if self.output:
            # Display and close stdout
            try:
                sys.stdout.close()
                if not self.debug:
                    if os.path.getsize(self.stdout):
                        text = open(self.stdout, "r")
                        mess = ""
                        for txt in text:
                            mess = "%s%s" % (mess, txt)
                        text.close()
                        ScrollText(title="Standard Output", mess=mess)
                    # Housekeeping
                    for pid in range(1000):
                        try:
                            os.remove(os.path.join(getPrgPath(),
                                "stdout_%s.txt" % pid))
                        except:
                            pass
            except:
                pass
        if sysexit:
            os._exit(1)

    def setVariables(self):
        if not self.user:
            self.display = True
        if self.menfle:
            try:
                exec "import %s as usrmen" % self.menfle
            except:
                import stdmen as usrmen
        else:
            import stdmen as usrmen
        self.sys = {}
        self.men = []
        self.mod = []
        for s in usrmen.systems:
            self.sys[s[0]] = s[1]
            try:
                exec "self.men.extend(usrmen.%2smen)" % s[0]
                exec "self.mod.extend(usrmen.%2smod)" % s[0]
            except:
                pass
        self.direct = False

    def doProcess(self):
        if self.image:
            self.img = os.path.join(self.rcdic["prgdir"], "img", "tartan.gif")
        else:
            self.img = None
        if self.debug:
            self.cmd = lambda wid, cmd: self.tracer.runfunc(self.execModule,
                (wid, cmd))
        else:
            self.cmd = lambda wid, cmd: self.execModule((wid, cmd))
        self.menu = TartanMenu(self.mf, self.user["name"], self.usrmen,
            self.usrmod, lvl=self.user["lvl"], cmd=self.cmd, img=self.img)
        if not self.doVersionCheck(vtype="chk"):
            self.menu.checkNotes(self.user["name"])
        self.menu.drawMenu()
        self.mf.startLoop()

    def doLogin(self):
        if not self.user:
            self.doReadCheckUser()
            if self.user and not self.user["pwd"]:
                self.updateCtlLog("login", "msc000")
                return
        if self.user:
            login = self.user["name"]
        else:
            try:
                login = getpass.getuser()
                if not self.doReadCheckUser(login, userchk=True):
                    raise Exception
            except:
                login = ""
        self.mf.head.set_text("Tartan Systems - User Validation")
        tit = ("User Validation",)
        fld = (
            (("T",0,0,0),"INA",30,"User Name","Your User Name",
                login,"N",self.usrNam,None,None,None,None,
                "Your log-in Name"),
            (("T",0,1,0),"IHA",30,"User Password","Your Password",
                "","N",self.usrPwd,None,None,None,None,
                "Your password. You are allowed a Maximum of 3 attempts."))
        but = (("Quit", None, self.noUser, 1, None, None),)
        self.df = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            butt=but, tend=((self.endUser, "n"),), txit=(self.noUser,))
        self.mf.startLoop()

    def usrNam(self, frt, pag, r, c, p, i, w):
        self.doReadCheckUser(w)
        if not self.user:
            return "Invalid User"
        if not self.user["pwd"]:
            return "sk1"
        self.times = 0

    def usrPwd(self, frt, pag, r, c, p, i, w):
        err = self.doCheckPassword(w)
        if not err:
            return "nc"
        elif self.times == 2:
            return "xt"
        else:
            self.times += 1
            return "Invalid Password"

    def noUser(self, widget=None):
        self.user = {}
        self.endUser()

    def endUser(self):
        if self.user:
            self.updateCtlLog("login", "msc000")
        self.df.closeProcess()
        self.mf.closeLoop()

    def doLogout(self):
        self.mf.dbm.openDbase()
        self.updateCtlLog("logout", "msc000")
        self.mf.dbm.closeDbase()
        self.user = {}

    def doReadCheckUser(self, user=None, userchk=False, pwd=None, pwdchk=False):
        sql = Sql(self.mf.dbm, ["ctlmst", "ctlpwu", "ctlpwm"], prog="msc000")
        if not user:
            chk = sql.getRec(tables="ctlpwu")
            if len(chk) == 1:
                user = chk[0][sql.ctlpwu_col.index("usr_name")]
        usr = sql.getRec(tables="ctlpwu", where=[("usr_name", "=", user)],
            limit=1)
        if userchk:
            return usr
        if not usr:
            self.user = {}
        else:
            self.user = {
                "name": user,
                "pwd":  usr[sql.ctlpwu_col.index("usr_pwd")]}
            if "usr_last" in sql.ctlpwu_col:
                self.user["last"] = usr[sql.ctlpwu_col.index("usr_last")]
            else:
                self.user["last"] = 0
            self.user["coy"] = usr[sql.ctlpwu_col.index("usr_coy")]
            self.user["lvl"] = usr[sql.ctlpwu_col.index("usr_lvl")]
            if pwdchk and self.doCheckPassword(pwd):
                self.user = {}
            if self.user:
                self.cono = []
                self.usrnam = self.user["name"]
                self.pwd    = self.user["pwd"]
                last        = self.user["last"]
                coy         = self.user["coy"]
                self.lvl    = self.user["lvl"]
                if coy:
                    coy = coy.split(",")
                else:
                    coy = []
                for co in coy:
                    if int(co):
                        self.cono.append(int(co))
                nos = []
                # All systems not in the company modules
                for sys in self.sys:
                    if sys not in ("ca", "ms", "rp"):
                        nos.append(sys)
                ctm = sql.getRec(tables="ctlmst", cols=["ctm_modules"])
                for m in ctm:
                    for x in range(0, len(m[0]), 2):
                        if m[0][x:x+2].lower() in nos:
                            nos.remove(m[0][x:x+2].lower())
                # All systems excluded for the user
                tmp = sql.getRec(tables="ctlpwm", cols=["mpw_sys"],
                    where=[("mpw_usr", "=", self.user["name"]),
                    ("mpw_prg", "=", "")])
                if tmp:
                    for s in tmp:
                        if not s in nos:
                            nos.append(s[0])
                nop = sql.getRec(tables="ctlpwm", cols=["mpw_sys", "mpw_prg"],
                    where=[("mpw_usr", "=", self.user["name"]), ("mpw_prg",
                    "<>", ""), ("mpw_pwd", "=", "")])
                vop = sql.getRec(tables="ctlpwm", cols=["mpw_sys", "mpw_prg"],
                    where=[("mpw_usr", "=", self.user["name"]), ("mpw_prg",
                    "<>", ""), ("mpw_pwd", "<>", "")])
                if vop:
                    for sys, mod in vop:
                        if sys in nos:
                            nos.remove(sys)
                            for prg in self.mod:
                                if prg[2][:2] == sys:
                                    if prg[2][2:] == mod:
                                        continue
                                    nop.append([prg[2][:2], prg[2][2:]])
                self.usrmen = []
                self.usrmod = []
                for men in self.men:
                    add = True
                    for sys in nos:
                        if men[1] == "mm_%s" % sys:
                            add = False
                    if add:
                        self.usrmen.append(men)
                for mod in self.mod:
                    add = True
                    for sys in nos:
                        if mod[2][:2] == sys:
                            add = False
                    for sys, prg in nop:
                        if mod[2] == "%s%s" % (sys, prg):
                            add = False
                    if add:
                        self.usrmod.append(mod)

    def doCheckPassword(self, pwd):
        try:
            crypt = base64.decodestring(self.user["pwd"])
        except:
            crypt = self.user["pwd"]
        if pwd == self.mf.override:
            return
        elif pwd in (self.user["pwd"], crypt):
            pwlife = self.getCtlSys(["sys_pwlife"])
            if not pwlife or not self.user["last"]:
                return
            t = time.localtime()
            dte = (t[0] * 10000) + (t[1] * 100) + t[2]
            chk = projectDate(self.user["last"], pwlife)
            dif = dateDiff(dte, chk, ptype="days")
            if not dif:
                showWarning(self.mf.body, "Password Expired",
                    "Your Password Has Expired.\n\nPlease Contact Your "\
                    "System Administrator!!!")
                return "Expired Password"
            elif dif < 8:
                showWarning(self.mf.body, "Password Expiring",
                    "Your Password Expires in %s Day(s)!!!\n\nYou Must "\
                    "Renew It As Soon As Possible!!!" % dif)
        else:
            return "Invalid Password"

    def doCheckCono(self, coy):
        if self.cono and not self.cono.count(coy):
            return "Unavailable Company Number"
        self.conum = coy
        col, ctl = self.readCono(self.conum)
        if not ctl:
            return "Invalid Company Number"
        self.conum = ctl[col.index("ctm_cono")]
        self.conam = ctl[col.index("ctm_name")]
        self.email = ctl[col.index("ctm_email")]
        self.modul = ctl[col.index("ctm_modules")]

    def doDbCommand(self):
        if type(self.dbcmd) is list:
            flenam = self.dbcmd
        else:
            name = os.path.abspath(self.dbcmd)
            if os.path.isfile(name):
                flenam = open(name, "r")
            else:
                return "Invalid File"
        self.mf.dbm.openDbase()
        for line in flenam:
            line = line.replace(chr(10), "").replace(chr(13), "")
            if not line or line[0] == "#":
                continue
            comm = line.split()
            sel = False
            qty = None
            if comm and comm[0].lower() == "select":
                sel = True
                if comm[1][:3].lower() in ("avg", "max", "min", "sum"):
                    qty = 1
                elif comm[1].lower() == "count(*)":
                    qty = 1
                for num, cmd in enumerate(comm):
                    if num < 2:
                        continue
                    if cmd == "limit":
                        qty = int(comm[num + 1])
                        break
            try:
                if comm[0] == "commit":
                    self.mf.dbm.commitDbase()
                else:
                    if sel:
                        ret = self.mf.sqlRec(line, limit=qty)
                        mess = ""
                        for r in ret:
                            if type(r) is list:
                                r = str(r)[1:-1]
                            else:
                                r = str(r)
                            if not mess:
                                mess = r
                            else:
                                mess = mess + "\n" + r
                        if self.mf.display:
                            ScrollText(scrn=self.mf.body, mess=mess)
                        else:
                            print mess
                    else:
                        self.mf.sqlRec(line)
            except:
                self.mf.dbm.closeDbase()
                return "Error in SQL Statement\n\n%s" % line
        self.mf.dbm.closeDbase()

    def doDirect(self):
        if self.module == "DBCreate":
            # Database maintenance utility
            from TartanClasses import DBCreate
            opts = [("-i", True)]
            if self.sysfle:
                opts.append(("-s", self.sysfle))
            DBCreate(mf=self.mf, opts=opts)
            return
        elif self.module == "tarUpd":
            # table formats update utility
            self.tarUpd()
            return
        elif self.module == "ti1010":
            # Visual COBOL isam files convert and import
            self.doImport("ti1010").ti1010(self.mf, self.current[1])
            return
        elif self.module == "tb1010":
            # ffield and ftable maintenance
            if self.sysfle:
                sys = self.sysfle
            else:
                sys = "tartan"
            self.doImport("tb1010").tb1010(self.mf, sys)
            return
        elif self.module in ("tb1030", "tb3010"):
            # Table data maintenance utilities
            self.mf.dbm.openDbase()
            exec "self.doImport(self.module).%s(self.mf)" % self.module
            self.mf.dbm.closeDbase()
            return
        elif self.module == "TartanConfig":
            # tartanrc file maintenance utility
            from TartanClasses import TartanConfig
            TartanConfig(mf=self.mf)
            return
        self.direct = True
        self.menu = TartanMenu(self.mf, self.user["name"], self.usrmen,
            self.usrmod, lvl=self.lvl)
        if self.rtn:
            module = "%s%s" % (self.module[:-1], self.rtn)
        else:
            module = self.module
        try:
            typ = self.menu.moddic[module][0]
        except:
            return "Unknown or Invalid Module (%s)" % module
        if typ == "C":
            try:
                os.spawnv(os.P_WAIT, module, (module,))
            except:
                return "Command (%s) Error" % module
        elif typ == "F":
            try:
                self.updateCtlLog("start", module)
                exec "self.%s()" % module
                self.updateCtlLog("finis", module)
            except:
                self.updateCtlLog("error", module)
                return "Function (%s) Error" % module
        else:
            # Check if a correct company number has been entered
            if self.conum:
                self.mf.dbm.openDbase()
                err = self.doCheckCono(self.conum)
                self.mf.dbm.closeDbase()
                if err:
                    return err
            if self.debug:
                self.tracer.runfunc(self.execModule, (None, module))
            else:
                self.execModule((None, module))

    def execModule(self, *args):
        # Extract module
        module = args[0][1]
        # Remove the menu with or without the image
        if not self.direct:
            self.menu.removeMenu()
        self.mf.head.set_text("")
        self.mf.updateStatus("")
        if module == "sysEnd":
            self.doLogout()
            self.doHouseKeeping()
            self.mf.closeLoop()
            self.doExit()
        self.head = "%s (%s)" % (self.menu.moddic[module][1], module)
        if self.menu.moddic[module][0][0] == "C":            # Command
            self.mf.head.set_text(self.head)
            try:
                RunCommand(self.mf.body, module, self.menu.moddic[module][0][1])
            except:
                showException(self.mf.body, self.rcdic["wrkdir"],
                    "Command (%s) Error" % module, dbm=self.mf.dbm)
        elif self.menu.moddic[module][0][0] == "F":          # Function
            self.mf.head.set_text(self.head)
            try:
                self.updateCtlLog("start", module)
                exec "self.%s()" % module
                self.updateCtlLog("finis", module)
            except:
                self.updateCtlLog("error", module)
                showException(self.mf.body, self.rcdic["wrkdir"],
                    "Function (%s) Error" % module, dbm=self.mf.dbm)
        elif self.menu.moddic[module][0][0] == "S":          # Standalone
            self.mf.head.set_text(self.head)
            exec "self.doImport(module).%s(self.mf.body)" % module
        elif self.menu.moddic[module][0][0] == "U":          # User Details
            self.mf.head.set_text(self.head)
            exec "self.doImport(module).%s(self.mf, self.user)" % module
        else:                                                # Tartan Module
            err1 = self.doOpenDbase()
            err2 = self.doVersionCheck(vtype="chk")
            if err1 or err2:
                self.mf.dbm.closeDbase()
                if not self.direct:
                    self.menu.drawMenu()
                return
            typ, coy, self.per, self.usr = self.menu.moddic[module][0]
            cancel = False
            if coy == "Y":
                if self.direct and self.conum:
                    if self.per == "N":
                        self.pernum = None
                    elif self.pernum == None:
                        self.getLastPeriod()
                else:
                    self.getConum()
                if not self.conum:
                    cancel = True
                elif self.per == "Y" and self.pernum == None:
                    cancel = True
                else:
                    err = self.chkSystem(module)
                    if err:
                        cancel = True
                    else:
                        self.rtn = self.menu.moddic[module][2]
            else:
                self.conum = None
                self.email = None
                self.rtn = self.menu.moddic[module][2]
            if not cancel:
                p = self.chkPassword(module)
                if p == "cc":
                    pass
                elif p == "no":
                    showError(self.mf.body, "Permission Denied",
                        "User Not Validated for:\n\n%s %s (%s)" % \
                        (self.sys[module[:2]], self.menu.moddic[module][1],
                        module))
                else:
                    self.runModule(module)
            self.mf.dbm.closeDbase()
        if not self.direct:
            if module == "ms1010":  # Company Record Maintenance
                self.chgUsr()
            self.menu.checkNotes(self.user["name"])
            self.menu.drawMenu()

    def chkPassword(self, module):
        """
        To disable a whole system:
            Coy      = 0 or a valid company number
            SS       = gl
            Prog     = Blank
            Password = Blank
        To disable a module:
            Coy      = 0 or a valid company number
            SS       = mm
            Prog     = y010
            Password = Blank
        To control a module:
            Coy      = 0 or a valid company number
            SS       = mm
            Prog     = y010
            Password = Password
        """
        usr = self.user["name"]
        sys = module[:2]
        prg = module[2:]
        sql = Sql(self.mf.dbm, "ctlpwm", prog="msc000")
        pwd = sql.getRec(tables="ctlpwm", cols=["mpw_pwd"], where=[("mpw_usr",
            "=", usr), ("mpw_coy", "=", 0), ("mpw_sys", "=", sys), ("mpw_prg",
            "=", prg)], limit=1)
        if not pwd and self.conum:
            pwd = sql.getRec(tables="ctlpwm", cols=["mpw_pwd"],
                where=[("mpw_usr", "=", usr), ("mpw_coy", "=", self.conum),
                ("mpw_sys", "=", sys), ("mpw_prg", "=", prg)], limit=1)
        if not pwd:
            ok = "ok"
        elif not pwd[0]:
            ok = "no"
        else:
            pwd = base64.decodestring(pwd[0])
            self.doChkPwd()
            if not self.ppwd:
                ok = "cc"
            elif self.ppwd == self.mf.override:
                ok = "ok"
            elif pwd.strip() == self.ppwd.strip():
                ok = "ok"
            else:
                ok = "no"
        return ok

    def doChkPwd(self):
        tit = ("Password Validation",)
        fld = ((("T",0,0,0),"IHA",30,"Password","Password","","N",
                self.doPwdGet,None,None,None,None),)
        but = (("Cancel", None, self.doPwdCancel, 1, None, None),)
        self.df = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            butt=but, tend=((self.doPwdEnd, "n"),), txit=(self.doPwdCancel,))
        self.mf.startLoop()

    def doPwdGet(self, frt, pag, r, c, p, i, pwd):
        self.ppwd = pwd
        return "nc"

    def doPwdCancel(self, widget=None):
        self.ppwd = None
        self.doPwdEnd()

    def doPwdEnd(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def runModule(self, module):
        try:
            mod = "%s%s" % (module[:5], 0)
            self.updateCtlLog("start", mod)
            if self.conum == None:
                self.conum = 0
                self.mf.head.set_text(self.head)
                if self.usr == "N":
                    exec "self.doImport(mod).%s(self.mf)" % mod
                else:
                    exec "self.doImport(mod).%s(self.mf, "\
                        "self.user['name'])" % mod
            elif self.pernum == None:
                self.head = "%s for %s" % (self.head, self.conam)
                self.mf.head.set_text(self.head)
                if self.usr == "N":
                    exec "self.doImport(mod).%s(self.mf, self.conum, "\
                        "self.conam)" % mod
                else:
                    exec "self.doImport(mod).%s(self.mf, self.conum, "\
                        "self.conam, self.user['name'])" % mod
            else:
                if module in ("gl3030", "gl3040", "gl3050", "gl3080",
                              "gl4010", "gl4020", "gl6030", "gl6040"):
                    check = True
                else:
                    check = False
                start, end, fin = getPeriods(self.mf, self.conum, self.pernum,
                    check=check)
                if check and (start, end, fin) == (None, None, None):
                    pass
                elif start == None or end == None:
                    pass
                elif self.rtn and fin == "Y":
                    showError(self.mf.body, "Period Error",
                        "This Financial Period Has Already Been Finalised")
                else:
                    self.head = "%s for %s for Period %s to %s" % (self.head,
                        self.conam, start.disp, end.disp)
                    self.mf.head.set_text(self.head)
                    period = (self.pernum, (start.work, start.disp),
                        (end.work, end.disp))
                    if not self.rtn:
                        if self.usr == "N":
                            exec "self.doImport(mod).%s(self.mf, self.conum, "\
                                "self.conam, period)" % mod
                        elif module == "ml1010":
                            exec "self.doImport(mod).%s(self.mf, self.conum, "\
                                "self.conam, period, self.user['name'], "\
                                "self.lvl)" % mod
                        else:
                            exec "self.doImport(mod).%s(self.mf, self.conum, "\
                                "self.conam, period, self.user['name'])" % mod
                    elif self.usr == "N":
                        exec "self.doImport(mod).%s(self.mf, self.conum, "\
                            "self.conam, period, self.rtn)" % mod
                    else:
                        exec "self.doImport(mod).%s(self.mf, self.conum, "\
                            "self.conam, period, self.rtn, "\
                            "self.user['name'])" % mod
            self.updateCtlLog("finis", mod)
        except:
            self.updateCtlLog("error", mod)
            showException(self.mf.body, self.rcdic["wrkdir"],
                "Module %s Error" % module, dbm=self.mf.dbm)

    def updateCtlLog(self, utype, module):
        if not self.mf.dbm.dbopen:
            opendb = True
            self.mf.dbm.openDbase()
        else:
            opendb = False
        #===================================================================
        # Rollback any uncommitted transactions before updating the log file
        #===================================================================
        try:
            self.mf.dbm.rollbackDbase()
        except:
            pass
        #===================================================================
        sql = Sql(self.mf.dbm, "ctllog", prog="msc000")
        if not sql.error:
            try:
                log = getpass.getuser()
            except:
                log = "unknown"
            if self.user:
                nam = self.user["name"]
            else:
                nam = log
            if self.rtn:
                rtn = self.rtn
            else:
                rtn = 0
            if self.conum:
                coy = self.conum
            else:
                coy = 0
            if self.pernum != None:
                per = self.pernum
            else:
                per = 0
            if utype == "login":
                self.login = long("%04i%02i%02i%02i%02i%02i" % \
                    time.localtime()[:-3])
                sql.insRec("ctllog", data=[log, nam, module, 0, 0, 0,
                    self.login, 0, ""])
            elif utype == "logout":
                self.end = long("%04i%02i%02i%02i%02i%02i" % \
                    time.localtime()[:-3])
                whr = [("clg_login", "=", log), ("clg_user", "=", nam),
                    ("clg_prog", "=", module), ("clg_start", "=", self.login)]
                dat = [self.end, "N"]
                col = ["clg_finis", "clg_error"]
                sql.updRec("ctllog", where=whr, data=dat, cols=col)
            elif utype == "start":
                self.start = long("%04i%02i%02i%02i%02i%02i" % \
                    time.localtime()[:-3])
                sql.insRec("ctllog", data=[log, nam, module, rtn, coy,
                    per, self.start, 0, ""])
            else:
                self.end = long("%04i%02i%02i%02i%02i%02i" % \
                    time.localtime()[:-3])
                if utype == "finis":
                    err = "N"
                else:
                    err = "Y"
                whr = [("clg_login", "=", log), ("clg_user", "=", nam),
                    ("clg_prog", "=", module), ("clg_start", "=", self.start)]
                dat = [self.end, err]
                col = ["clg_finis", "clg_error"]
                sql.updRec("ctllog", where=whr, data=dat, cols=col)
            #===============================================================
            # Commit ctllog transaction
            #===============================================================
            self.mf.dbm.commitDbase()
        if opendb:
            self.mf.dbm.closeDbase()

    def getConum(self):
        sql = Sql(self.mf.dbm, "ctlmst", prog="msc000")
        if self.cono:
            whr = [("ctm_cono", "in", tuple(self.cono))]
        else:
            whr = None
        coy = sql.getRec(tables="ctlmst", where=whr, order="ctm_cono")
        if not coy:
            showError(self.mf.body, "Error", "No Valid Company Records")
            return
        self.coys = len(coy)
        if self.coys == 1:
            # Single Company
            self.conum = coy[0][sql.ctlmst_col.index("ctm_cono")]
            self.conam = coy[0][sql.ctlmst_col.index("ctm_name")]
            self.email = coy[0][sql.ctlmst_col.index("ctm_email")]
            self.modul = coy[0][sql.ctlmst_col.index("ctm_modules")]
            if self.per == "N":
                self.pernum = None
                return
        if not self.conum:
            if self.cono:
                self.conum = self.cono[0]
            else:
                self.conum = 1
        csel = {
            "stype": "R",
            "tables": ("ctlmst",),
            "cols": (
                ("ctm_cono", "", 0, "Com"),
                ("ctm_name", "", 0, "Name", "Y")),
            "order": "ctm_cono"}
        if self.cono:
            csel["where"] = [("ctm_cono", "in", tuple(self.cono))]
        if self.coys == 1:
            # Single Company
            tit = (self.conam,)
            fld = []
        else:
            tit = ("Company Details",)
            fld = [
                [("T",0,0,0),"IUI",3,"Company Number","Company Number",
                    self.conum,"N",self.coNum,csel,None,("notzero",)],
                [("T",0,1,0),"ONA",30,"Company Name","",
                    self.conam,"N",None,None,None,None]]
        if self.per == "Y":
            self.psel = {
                "stype": "R",
                "tables": ("ctlynd",),
                "cols": (
                    ("cye_period", "", 0, "Prd"),
                    ("cye_start", "", 0, "Start"),
                    ("cye_end", "", 0, "End")),
                "where": []}
            fld.append(
                [["T",0,2,0],"IUI",3,"Financial Period","Financial Period",
                    "","N",self.finPeriod,self.psel,None,None])
            if self.coys == 1:
                self.psel["where"] = [("cye_cono", "=", self.conum)]
                fld[0][0][2] = 0
                self.getLastPeriod()
                fld[0][5] = self.pernum
        elif self.per == "L":
            self.getLastPeriod()
            if self.coys == 1:
                return
        else:
            self.pernum = None
            if self.coys == 1:
                return
        but = (("Cancel", None, self.coExit, 1, ("T",0,1), ("T",0,0)),)
        self.df = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            butt=but, tend=((self.coEnd, "y"),), txit=(self.coExit,))
        if self.conum:
            try:
                self.df.loadEntry("T", 0, 0, data=self.conum)
                self.df.loadEntry("T", 0, 1, data=self.conam)
                if self.per == "Y" and self.pernum != None:
                    self.df.topf[0][2][5] = self.pernum
                    self.df.loadEntry("T", 0, 2, data=self.pernum)
            except:
                pass
        self.df.focusField("T", 0, 1)
        self.mf.startLoop()

    def coNum(self, frt, pag, r, c, p, i, w):
        err = self.doCheckCono(w)
        if err:
            return err
        self.df.loadEntry("T", pag, p+1, data=self.conam)
        if self.per in ("L", "N"):
            return "ok"
        self.psel["where"] = [("cye_cono", "=", self.conum)]
        if self.pernum == None:
            self.getLastPeriod()
        self.df.topf[0][2][5] = self.pernum

    def getLastPeriod(self):
        sql = Sql(self.mf.dbm, "ctlynd", prog="msc000")
        p = sql.getRec(tables="ctlynd", cols=["max(cye_period)"],
            where=[("cye_cono", "=", self.conum)])
        self.pernum = int(p[0][0])

    def finPeriod(self, frt, pag, r, c, p, i, w):
        sql = Sql(self.mf.dbm, "ctlynd", prog="msc000")
        r = sql.getRec(tables="ctlynd", cols=["cye_period"], where=[("cye_cono",
            "=", self.conum), ("cye_period", "=", w)], limit=1)
        if not r:
            return "Invalid Financial Period"
        self.pernum = w

    def coExit(self, widget=None):
        self.conum = None
        self.coEnd()

    def coEnd(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def chkSystem(self, module):
        if module[:2] in ("bm", "ca", "ms", "rc"):
            # Not system modules e.g. gl, dr, cr etc.
            return
        elif module[:2] == "sl":
            # Hack for Staff Loans modules
            mod = "WG"
        else:
            mod = module[:2].upper()
        if mod in self.modul:
            return
        showError(self.mf.body, "System Error",
            "This System (%s) is NOT in the Company Record" % mod)
        return "System Not In Company Record"

    def readCono(self, cono):
        sql = Sql(self.mf.dbm, "ctlmst", prog="msc000")
        ctl = sql.getRec(tables="ctlmst", where=[("ctm_cono", "=", cono)],
            limit=1)
        return sql.ctlmst_col, ctl

    def chgPwd(self):
        tit = ("Change Password",)
        fld = (
            (("T",0,0,0),"IHA",30,"Old Password","Old Password",
                "","N",self.doOldPwd,None,None,None,None),
            (("T",0,1,0),"IHA",30,"New Password","New Password",
                "","N",self.doNewP1,None,None,None,None),
            (("T",0,2,0),"IHA",30,"New Password","New Password Again",
                "","N",self.doNewP2,None,None,None,None))
        but = (("Cancel", None, self.doPwdEnd, 1, None, None),)
        self.doOpenDbase()
        self.df = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            butt=but, tend=((self.doPwdEnd, "n"),), txit=(self.doPwdEnd,))
        self.mf.startLoop()
        self.mf.dbm.closeDbase()

    def doOldPwd(self, frt, pag, r, c, p, i, pwd):
        crypt = base64.encodestring(pwd)
        if pwd == self.pwd or pwd == base64.decodestring(self.pwd):
            pass
        else:
            return "Invalid Old Password"

    def doNewP1(self, frt, pag, r, c, p, i, pwd):
        if pwd == self.pwd or pwd == base64.decodestring(self.pwd):
            return "Same Password"
        pwmust, pwsize = self.getCtlSys(["sys_pwmust", "sys_pwsize"])
        if not pwd and pwmust == "Y":
            return "You Must Have a Password"
        if pwsize and len(pwd) < pwsize:
            return "Minimum of %s Characters Required" % pwsize
        self.new = pwd

    def doNewP2(self, frt, pag, r, c, p, i, pwd):
        if pwd != self.new:
            return "Invalid New Password"
        self.pwd = base64.encodestring(pwd)
        t = time.localtime()
        dte = (t[0] * 10000) + (t[1] * 100) + t[2]
        sql = Sql(self.mf.dbm, "ctlpwu", prog="msc000")
        sql.updRec("ctlpwu", cols=["usr_pwd", "usr_last"], data=[self.pwd, dte],
            where=[("usr_name", "=", self.user["name"])])
        self.mf.dbm.commitDbase()

    def doPwdEnd(self, widget=None):
        self.user["pwd"] = self.pwd
        self.df.closeProcess()
        self.mf.closeLoop()

    def tarUsr(self):
        sys = []
        prg = {}
        keys = self.sys.keys()
        keys.sort()
        for s in keys:
            sys.append((self.sys[s], s))
            exec "%sm = []" % s
            for mod in self.mod:
                if mod[0][0] == "P" and mod[2][:2] == s:
                    if len(mod) == 6:
                        mod[2] = "%s%s" % (mod[2][:5], mod[5])
                    exec "%sm.append((mod[3], mod[2][2:], mod[4]))" % s
            exec "prg['%s'] = %sm" % (s, s)
        self.doOpenDbase()
        TartanUser(self.mf, self.mf.body, sys, prg)
        self.mf.dbm.closeDbase()

    def chgUsr(self):
        self.doLogout()
        self.mf.dbm.openDbase()
        self.doLogin()
        self.mf.dbm.closeDbase()
        if self.user:
            self.menu = TartanMenu(self.mf, self.user["name"], self.usrmen,
                self.usrmod, lvl=self.user["lvl"], cmd=self.cmd, img=self.img)
        else:
            self.doExit()

    def tarCfg(self):
        self.mf.dbm.openDbase()
        cfg = TartanConfig(mf=self.mf, rcdic=self.rcdic, level=self.lvl)
        self.mf.dbm.closeDbase()
        self.rcdic = loadRcFile(self.rcfile)

    def sysUpd(self):
        tit = ("System Upgrade",)
        typ = (("Internet", "I"), ("Local", "L"))
        fld = (
            (("T",0,0,0),("IRB",typ),0,"Upgrade Type","",
                "I","N",self.doSysLoc,None,None,None,None),
            (("T",0,1,0),"ONA",9,"Current Version","",
                "","N",None,None,None,None,None),
            (("T",0,2,0),"ONA",9,"Upgrade Version","",
                "","N",None,None,None,None,None))
        but = (
            ("Upgrade", None, self.doSysUpgrade, 0, ("T",0,0), ("T",0,1)),
            ("Cancel", None, self.doSysUpdXit, 1, None, None))
        self.su = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            butt=but, tend=None, txit=(self.doSysUpdXit,))
        self.mf.startLoop()

    def doSysLoc(self, frt, pag, r, c, p, i, w):
        self.updtyp = w
        if self.updtyp == "I":
            err = self.doSysChkUpgrade()
            if err:
                showError(self.mf.body, err[0], err[1])
                return "Upgrade Error"
            else:
                self.su.loadEntry("T", 0, 1, data=self.current[1])
                self.su.loadEntry("T", 0, 2, data=self.nv)
                return
        if sys.platform == "win32":
            ftype = [("Upgrade Files", "Tartan_upd_*.exe")]
        else:
            ftype = [("Upgrade Files", "Tartan_*.tar.gz")]
        dialog = FileDialog(title="Select Upgrade File",
            initd=self.rcdic["upgdir"], ftype=ftype)
        self.updfle = dialog.askopenfilename()
        if not self.updfle:
            return "Invalid File"
        self.su.loadEntry("T", 0, 1, data=self.current[1])
        nv = os.path.basename(self.updfle).split("_")
        if sys.platform == "win32":
            nv = nv[2].split(".")
        else:
            nv = nv[1].split(".")
        self.newver = (int(nv[0]), int(nv[1]), int(nv[2]))
        self.nv = "%s.%s.%s" % self.newver
        if sys.platform == "win32" and self.oldver < (4, 0, 0) and self.newver > (4, 0, 0):
            showError(self.mf.body, "Upgrade", "This version of Tartan, %s, cannot be Upgraded to version %s. You will have to Un-Install your Current version of Tartan, download Tartan_full_%s.exe from ftp://tartan.co.za and then run it." % (self.ov, self.nv, self.nv))
            return "Invalid Version"
        self.su.loadEntry("T", 0, 2, data=self.nv)
        if int(nv[0]) > self.current[0][0]:
            return
        elif int(nv[0]) == self.current[0][0]:
            if int(nv[1]) > self.current[0][1]:
                return
            elif int(nv[1]) == self.current[0][1]:
                if int(nv[2]) > self.current[0][2]:
                    return
        # Check for Bug Fix e.g. 3.4.31.1.exe or 3.4.31.1.tar.gz
        if sys.platform == "win32":
            if len(nv) == 5:
                return
        elif len(nv) == 6:
            return
        self.updfle = None
        showError(self.mf.body, "Invalid Version",
            "Not Later Than Current Version")
        return "Version Error"

    def doSysChkUpgrade(self):
        try:
            self.doOpenDbase()
            col, ctl = self.readCono(1)
            self.mf.dbm.closeDbase()
            if not ctl or not ctl[col.index("ctm_email")]:
                self.emadd = "info@tartan.co.za"
            else:
                self.emadd = ctl[col.index("ctm_email")]
            flenam = cStringIO.StringIO()
            error = ftpDownload("tartan.co.za", "current", word=self.emadd,
                dest=flenam, close=False)
            if error:
                raise Exception
            v = flenam.getvalue()
            flenam.close()
            # Read and create Tartan file
            v = v.replace(chr(10), "").replace(chr(13), "").split(".")
            self.newver = (int(v[0]), int(v[1]), int(v[2]))
            self.nv = "%s.%s.%s" % self.newver
            if self.oldver < (4, 0, 0) and self.newver > (4, 0, 0):
                return ("Upgrade Not Allowed",
                    """This version of Tartan %s Cannot be Upgraded to version %s. You will have to Un-Install your Current version of Tartan and then Download the Latest version from the Tartan site and then Install it.""" % (self.ov, self.nv))
            if self.newver[0] > self.current[0][0]:
                return
            elif self.newver[0] == self.current[0][0]:
                if self.newver[1] > self.current[0][1]:
                    return
                elif self.newver[1] == self.current[0][1]:
                    if self.newver[2] > self.current[0][2]:
                        return
            return ("Upgrade", "No Upgrade Available")
        except:
            return ("Connection Error", """Please Ensure that You Are
Connected to the Internet.


"If Not, Please Connect and then Try Again!""")

    def doSysUpgrade(self, widget=None):
        self.mf.updateStatus("")
        self.su.closeProcess()
        sp = ScrollText(scrn=self.mf.body, width=750, height=500, mess="""
                        Performing the Upgrade

  After the update has completed you must perform the following:

  1) If you are running Linux the downloaded file will be in your
     'Tartan/wrk' directory. The file's name will be something like
     'Tartan_3.x.xx.tar.gz'.

                                 or

  1) If you are running Windows the downloaded file will be in your
     'Tartan\wrk' directory. The file's name will be something like
     'Tartan_upd_3.x.xx.exe'.
  2) Restart Tartan.
  3) Execute the 'Update File Formats' routine on the 'System' menu.  
  4) If Tartan is installed on other workstations, copy the downloaded
     file to those other workstations and either, in the case of linux,
     extract the file into the Tartan/prg directory or, in the case of
     Windows, execute the file.""")
        try:
            if self.updtyp == "I":
                if sys.platform == "win32":
                    fnam = "Tartan_upd_%s.%s.%s.exe" % self.newver
                else:
                    fnam = "Tartan_%s.%s.%s.tar.gz" % self.newver
                fle = os.path.join(self.mf.wrkdir, fnam)
                error = ftpDownload("ftp.tartan.co.za", fnam, word=self.emadd,
                    dest=fle)
                if error:
                    raise Exception
            else:
                fle = self.updfle
            # Unpack New Tartan file
            if sys.platform == "win32":
                os.spawnv(os.P_NOWAIT, fle, (fle,))
            else:
                os.spawnv(os.P_NOWAIT, "/bin/tar",
                    ("tar", "-xzf", fle, "-C", self.mf.prgdir))
        except:
            showError(self.mf.body, "Get Error",
                "Upgrade File Could Not be Retrieved!\n\n"\
                "Your System Has Not Been Upgraded!\n\n"\
                "Please Contact Your IT Manager!")
        # Exit msc000
        sys.exit()

    def doSysUpdXit(self, widget=None):
        self.su.closeProcess()
        self.mf.closeLoop()
        
    def tarUpd(self):
        if not self.force:
            # Check on status of Database
            if self.doVersionCheck(vtype="upd"):
                return
        else:
            self.doVersionCheck(vtype="get")
        if self.mf.display:
            ok = askQuestion(self.mf.body, "Backup",
                "Have You Backed Up the Database?\nYes to Continue else No!",
                default="no")
        else:
            ok = raw_input("Have You Backed Up the Database? (yes/no) ")
        if ok != "yes":
            return
        import tb1020
        pth = ("-d", os.path.join(self.mf.prgdir, "csv"))
        gui = ("-g", True)
        mfm = ("-m", self.mf)
        pwd = ("-p", self.mf.dbpwd)
        rcf = ("-r", self.rcfile)
        sys = ("-s", self.sysfle)
        usr = ("-u", self.mf.dbuser)
        ver = ("-v", self.current[1])
        # Check if Version earlier than 3.1.6 and if so enforce a full
        # data format of all tables. In version 3.1.6 there was a change
        # to the Document and Purchases orders format i.e. Na to UI
        if self.oldver < (3, 1, 6) or self.reformat:
            fmt = ("-f", "y")
        else:
            fmt = ("-f", "n")
        if self.tables:
            tab = ("-t", self.tables)
        else:
            tab = ("-t", None)
        self.doOpenDbase()
        upd = tb1020.tb1020(opts=[gui,mfm,pth,usr,pwd,rcf,sys,ver,fmt,tab])
        self.mf.dbm.closeDbase()

    def tarBck(self):
        self.doOpenDbase()
        cf = PwdConfirm(self.mf, scrn=self.mf.body, conum=0, system="MST",
            code="TarBck", passwd=self.bpwd)
        if cf.flag == "ok":
            if self.display:
                smtp = self.getSMTP()
            else:
                smtp = None
            TarBckRes(self.mf, mode="B", smtp=smtp)
        self.mf.dbm.closeDbase()

    def tarRes(self):
        self.doOpenDbase()
        cf = PwdConfirm(self.mf, scrn=self.mf.body, conum=0, system="MST",
            code="TarRes", passwd=self.bpwd)
        if cf.flag == "ok":
            TarBckRes(self.mf, mode="R", ver=self.current[1], sys=self.sysfle)
        self.mf.dbm.closeDbase()

    def showCal(self):
        self.mf.updateStatus("")
        TartanCalendar(screen=self.mf.body)

    def doAbout(self):
        aboutTartan(self.mf.window, self.current[1])

    def doManual(self):
        doc = os.path.join(self.rcdic["prgdir"], "doc", "Manual.html")
        web = "http://tartan.co.za/cgi-bin/simple_web.py/Documentation"
        try:
            import webbrowser
            try:
                webbrowser.open(doc)
            except:
                try:
                    import urllib2
                    f = urllib2.urlopen(urllib2.Request(web))
                    webbrowser.open(web)
                except:
                    raise Exception
        except:
            showError(self.mf.body, "Error", "Browser Error")

    def doOpenDbase(self, levl=1, check=False):
        if check:
            while not self.mf.dbm.checkDbase():
                ok = askQuestion(self.mf.body, "Missing Database",
                    "Do You Want To Create Your Database Now ?")
                if ok == "yes":
                    from TartanClasses import DBCreate
                    opts = [
                        ("-c", "i"),
                        ("-d", "%s" % os.path.join(self.mf.prgdir, "csv")),
                        ("-l", True),
                        ("-p", self.rcdic["dbpwd"]),
                        ("-u", self.rcdic["dbuser"]),
                        ("-w", False),
                        ("-x", False)]
                    if self.sysfle:
                        opts.append(("-s", self.sysfle))
                    else:
                        opts.append(("-s", "tartan"))
                    create = DBCreate(mf=self.mf, opts=opts)
                    if not create.put:
                        return "error"
                        break
                    t = time.localtime()
                    sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
                    sql = Sql(self.mf.dbm, ["verupd"], prog="msc000")
                    sql.insRec("verupd", data=[self.current[1], sysdtw])
                    self.mf.dbm.commitDbase()
                else:
                    if self.rcdic["dbase"] == "SQLite":
                        dbdsn = os.path.join(self.rcdic["dbdir"],
                            self.rcdic["dbname"])
                        if os.path.isfile(dbdsn):
                            os.remove(dbdsn)
                    return "error"
                    break
        self.mf.dbm.openDbase(levl=levl)

    def doCheckSys(self):
        sql = Sql(self.mf.dbm, "ctlsys", prog=__name__)
        if sql.error:
            return "error"
        if sql.getRec(tables="ctlsys", limit=1):
            return
        self.mf.head.set_text("System Record Maintenance (msc110)")
        callModule(self.mf, None, "msc110", user="admin")
        rec = sql.getRec(tables="ctlsys", limit=1)
        if not rec:
            return "error"

    def doCheckMst(self):
        sql = Sql(self.mf.dbm, "ctlmst", prog="msc000")
        rec = sql.getRec(tables="ctlmst", where=[("ctm_cono", "=", 1)], limit=1)
        if rec:
            return
        self.mf.head.set_text("Company Records Maintenance (ms1010)")
        callModule(self.mf, None, "ms1010", user="admin")
        rec = sql.getRec(tables="ctlmst", where=[("ctm_cono", "=", 1)], limit=1)
        if not rec:
            return "error"

    def doHouseKeeping(self):
        files = glob.glob(os.path.join(self.rcdic["wrkdir"], "*.ps"))
        for flenam in files:
            try:
                os.remove(flenam)
            except:
                pass
        files = glob.glob(os.path.join(self.rcdic["wrkdir"], "*.pdf"))
        files.extend(glob.glob(os.path.join(self.rcdic["wrkdir"], "*.gif")))
        files.extend(glob.glob(os.path.join(self.rcdic["wrkdir"], "*.jpg")))
        files.extend(glob.glob(os.path.join(self.rcdic["wrkdir"], "*.odt")))
        files.extend(glob.glob(os.path.join(self.rcdic["wrkdir"], "*.csv")))
        files.extend(glob.glob(os.path.join(self.rcdic["wrkdir"], "*.xls")))
        if files:
            ask = askQuestion(self.mf.body, "Report Files",
            "Do you want to keep the .pdf, .csv and .xls report files "\
            "in the wrk directory?", default="no")
            if ask == "no":
                for flenam in files:
                    try:
                        os.remove(flenam)
                    except:
                        pass

    def doVersionCheck(self, vtype="chk"):
        error = None
        self.oldver = (0, 0, 0)
        self.ov = "%s.%s.%s" % self.oldver
        upddate = CCD(0, "d1", 10)
        if not self.display:
            scrn = None
        else:
            scrn = self.mf.body
        if not self.mf.dbm.dbopen:
            opendb = True
            self.mf.dbm.openDbase()
        else:
            opendb = False
        sql = Sql(self.mf.dbm, "verupd", prog="msc000")
        if sql.error:
            if vtype == "chk":
                return "missing"
            else:
                return
        ver = sql.getRec(tables="verupd", cols=["max(ver_version)"], limit=1)
        if ver[0]:
            v = ver[0].split(".")
            self.oldver = (int(v[0]), int(v[1]), int(v[2]))
            self.ov = "%s.%s.%s" % self.oldver
            upd = sql.getRec(tables="verupd", cols=["ver_upddate"],
                where=[("ver_version", "=", ver[0])], limit=1)
            if upd:
                upddate = CCD(upd[0], "d1", 10)
        if opendb:
            self.mf.dbm.closeDbase()
        if vtype == "get":
            return
        if vtype == "chk" and self.check:
            if self.oldver == (0, 0, 0):
                error = "not"
            elif self.oldver < self.current[0]:
                error = "newer"
            elif self.oldver > self.current[0]:
                error = "older"
        elif vtype == "upd":
            if self.oldver == (0, 0, 0):
                pass
            elif self.oldver == self.current[0]:
                error = "updated"
            elif self.oldver > self.current[0]:
                error = "older"
        if error == "not":
            ok = showError(scrn, "Version Error", """'Update File Formats'
Not Yet Done for this Version of TARTAN.""")
        elif error == "newer":
            showError(scrn, "Version Error", """Your Version of TARTAN (%s) Is Newer than the File Formats!

Please Update Your File Formats""" % self.current[1])
        elif error == "older":
            showError(scrn, "Version Error", """Your Version of TARTAN (%s) Is Older than the File Formats!

Please Upgrade Your Version of TARTAN.""" % self.current[1])
        elif error == "updated":
            ok = askQuestion(scrn, "Version Error", """Your File Formats for TARTAN (%s) were Already Updated on %s.

Do You Want to Run the Update Again?""" % (self.current[1], upddate.disp), default="no")
            if ok == "yes":
                error = None
        return error

    def doImport(self, module):
        if self.rload and module in sys.modules:
            del(sys.modules[module])
        exec "import %s as imported" % module
        return imported

    def getSMTP(self, regdte=False):
        try:
            sql = Sql(self.mf.dbm, "ctlsys", prog="msc000", errs=False)
            if sql.error:
                raise Exception
            cols = ["sys_budays", "ctm_name", "sys_smtp", "ctm_email"]
            if regdte:
                cols.append("sys_regdte")
            acc = sql.getRec(tables=["ctlmst", "ctlsys"], cols=cols,
                where=[("ctm_cono", "=", 1)], limit=1, excp=True)
            if acc == "Exception":
                return None
            else:
                return acc
        except:
            return None

    def getCtlSys(self, cols):
        try:
            sql = Sql(self.mf.dbm, "ctlsys", prog="msc000", errs=False)
            if sql.error:
                raise Exception
            sys = sql.getRec(tables="ctlsys", cols=cols, limit=1)
            if not sys:
                raise Exception
            if len(cols) == 1:
                return sys[0]
            else:
                return sys
        except:
            sys = []
            for col in cols:
                sys.append(None)
            return sys
