#! /usr/bin/python

"""
SYNOPSIS
    Financial Year End routine.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt

It must be executed at the end of every financial period and it creates
the necessary records in the ctlynd table and also generates the correct
opening balances in the general ledger system if applicable.

It must also be executed whenever alterations to a previous financial period,
have been made, to update the opening balances of later financial periods.
"""

import time
from TartanClasses import ASD, CCD, GetCtl, PwdConfirm, Sql, TartanDialog
from tartanFunctions import getPeriods, mthendDate, showError

class msy010:
    def __init__(self, mf, conum, conam, period, args=None):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        no = self.setVariables()
        if not no:
            if args:
                self.final = args[0]
                self.commit = args[1]
                self.auto = True
                self.doProcess()
            else:
                self.commit = True
                self.auto = False
                self.drawScreen()
                self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlynd", "genbal", "genmst", "gentrn"],
            prog=__name__)
        if self.sql.error:
            return self.sql.error
        gc = GetCtl(self.mf)
        ctlmst = gc.getCtl("ctlmst", self.conum)
        if not ctlmst:
            return "no"
        mod = []
        for x in range(0, 21, 2):
            mod.append(ctlmst["ctm_modules"][x:x+2])
        if mod.count("GL"):
            self.genleg = True
            ctlctl = gc.getCtl("ctlctl", self.conum)
            if not ctlctl:
                return "no"
            if gc.chkRec(self.conum, ctlctl, ["ret_inc"]):
                return "no"
            self.ri_acc = ctlctl["ret_inc"]
        else:
            self.genleg = False
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.c_per = int(self.period[0])
        self.start = self.period[1][0]
        self.end = self.period[2][0]
        self.l_per = self.sql.getRec(tables="ctlynd", cols=["max(cye_period)"],
            where=[("cye_cono", "=", self.conum)], limit=1)[0]
        self.last, self.oldf = self.sql.getRec(tables="ctlynd",
            cols=["cye_last", "cye_final"], where=[("cye_cono", "=",
            self.conum), ("cye_period", "=", self.period[0])], limit=1)[:2]
        if self.oldf == "Y":
            cf = PwdConfirm(self.mf, conum=0, system="MST", code="YearEnd")
            if cf.pwd and cf.flag == "ok":
                return
            if not cf.pwd:
                showError(self.mf.body, "Year-End Error",
                    "This Financial Period Has Already Been Finalised!!!")
            return "no"

    def drawScreen(self):
        tit = ("%-3d %s" % (self.conum, self.conam), "Financial Year End")
        r1s = (("Yes", "Y"), ("No", "N"))
        fld = [
            (("T",0,0,0),"OD1",10,"Starting Date","",
                self.start,"N",None,None,None,None,("efld",)),
            [("T",0,1,0),"ID1",10,"Ending Date ","Ending Date",
                self.end,"N",self.doEnd,None,None,None,("efld",)],
            (("T",0,2,0),("IRB",r1s),0,"Finalise","Finalise",
                self.oldf,"N",self.doFinal,None,None,None,None,"""Finalising a financial period prevents any further data capture for that period as well as any previous periods i.e. if you finalise a period all previous financial periods are also automatically finalised.""")]
        if self.last:
            fld[1][1] = "OD1"
        but = (
            ("Apply",None,self.doProcess,1,("T",0,0),("T",0,1)),
            ("Cancel",None,self.doExit,1,("T",0,2),("T",0,1)))
        self.df = TartanDialog(self.mf, title=tit, eflds=fld, tend=(None,),
            txit=(self.doExit,), butt=but)

    def doEnd(self, frt, pag, r, c, p, i, w):
        self.end = w
        if self.end <= self.start:
            return "Invalid Year End Date"

    def doFinal(self, frt, pag, r, c, p, i, w):
        self.final = w

    def doProcess(self, widget=None):
        for per in range(0, self.c_per + 1):
            self.retinc = 0
            self.doYearEnd(per)
        if self.commit:
            self.mf.dbm.commitDbase()
        if not self.auto:
            self.doExit()

    def doYearEnd(self, per):
        self.sql.updRec("ctlynd", cols=["cye_last", "cye_final"],
            data=[self.sysdtw, self.final], where=[("cye_cono", "=",
            self.conum), ("cye_period", "=", per)])
        if per == self.c_per:
            start_c = self.start
            end_c = self.end
            if (per + 1) > self.l_per:
                newy = end_c / 10000
                newm = ((end_c / 100) % 100) + 1
                if newm > 12:
                    newy += 1
                    newm -= 12
                news = (newy * 10000) + (newm * 100) + 1
                newy = (end_c / 10000) + 1
                newm = (end_c / 100) % 100
                newe = (newy * 10000) + (newm * 100) + 1
                newe = mthendDate(newe)
                self.sql.insRec("ctlynd", data=[self.conum, (per + 1), news,
                    newe, 0, "N"])
        else:
            s, e, f = getPeriods(self.mf, self.conum, per)
            if s == None or e == None:
                return
            start_c = s.work
            end_c = e.work
        sp = start_c / 100
        ep = end_c / 100
        s, e, f = getPeriods(self.mf, self.conum, (per + 1))
        start_n = s.work
        end_n = e.work
        if not self.genleg:
            return
        self.sql.delRec("genbal", where=[("glo_cono", "=", self.conum),
            ("glo_trdt", "=", start_n)])
        gm = self.sql.getRec(tables="genmst", cols=["glm_acno, glm_type"],
            where=[("glm_cono", "=", self.conum)])
        if gm:
            for ac in gm:
                ov = self.sql.getRec(tables="genbal", cols=["glo_cyr"],
                    where=[("glo_cono", "=", self.conum), ("glo_acno", "=",
                    ac[0]), ("glo_trdt", "=", start_c)], limit=1)
                if not ov:
                    ov = CCD(0, "SD", 13.2)
                else:
                    ov = CCD(ov[0], "SD", 13.2)
                cy = self.sql.getRec(tables="gentrn",
                    cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono",
                    "=", self.conum), ("glt_acno", "=", ac[0]), ("glt_curdt",
                    "between", sp, ep)], limit=1)
                if not cy[0]:
                    cy = CCD(0, "SD", 13.2)
                else:
                    cy = CCD(cy[0], "SD", 13.2)
                bal = float(ASD(ov.work) + ASD(cy.work))
                if ac[1] == "P":
                    self.retinc = float(ASD(self.retinc) + ASD(bal))
                    bal = 0.00
                self.sql.insRec("genbal", data=[self.conum,ac[0],start_n,bal])
            ri_bal = self.sql.getRec(tables="genbal", cols=["glo_cyr"],
                where=[("glo_cono", "=", self.conum), ("glo_acno", "=",
                self.ri_acc), ("glo_trdt", "=", start_n)], limit=1)
            if not ri_bal:
                self.sql.insRec(tables="genbal", data=[self.conum,
                self.ri_acc, start_n, self.retinc])
            else:
                bal = float(ASD(ri_bal[0]) + ASD(self.retinc))
                self.sql.updRec("genbal", cols=["glo_cyr"], data=[bal],
                    where=[("glo_cono", "=", self.conum), ("glo_acno",
                    "=", self.ri_acc), ("glo_trdt", "=", start_n)])

    def doExit(self, widget=None):
        self.df.closeProcess()
        self.mf.closeLoop()

if __name__ == "__main__":
    import getopt, sys
    from TartanClasses import MainFrame
    try:
        opts, args = getopt.getopt(sys.argv[1:],"c:p:f:")
    except:
        print "Missing Options"
        sys.exit()
    coy = 1
    num = None
    fin = "N"
    for o, v in opts:
        if o == "-c":
            coy = int(v)
        elif o == "-p":
            num = int(v)
        elif o == "-f":
            fin = v.upper()
    if not num:
        print "Missing Period"
        sys.exit()
    mf = MainFrame(display=False)
    mf.dbm.openDbase()
    per = getPeriods(mf, coy, num)
    if per[1] and per[2]:
        per = (num, (per[0].work, per[0].disp), (per[1].work, per[1].disp))
        ex = msy010(mf, coy, "", per, (fin, True))
    else:
        print "Invalid Period %s for Company %s" % (num, coy)
    mf.dbm.closeDbase()
