#! /usr/bin/env python

"""
SYNOPSIS
    Stores Stock Movement Report.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Balances, CCD, ProgressBar, Sql, TartanDialog
from tartanFunctions import doReportlab, doWriteExport, getModName, doPrinter
from tartanFunctions import showError

class st3060:
    def __init__(self, mf, coy, nam):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        no = self.setVariables()
        if not no:
            self.mainProcess()
            self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["strgrp", "strloc", "strmf1"],
            prog=__name__)
        self.pdfnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sysdtd = "%i/%02i/%02i" % (t[0], t[1], t[2])
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i)" % \
            (t[0], t[1], t[2], t[3], t[4])
        self.stot = [0] * 8
        self.gtot = [0] * 8
        self.t = None

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "Stock Movement Report (%s)" % __name__)
        loc = {
            "stype": "R",
            "tables": ("strloc",),
            "cols": (
                ("srl_loc", "", 0, "L"),
                ("srl_desc", "", 0, "Description", "Y")),
            "where": [("srl_cono", "=", self.conum)]}
        grp = {
            "stype": "R",
            "tables": ("strgrp",),
            "cols": (
                ("gpm_group", "", 0, "Grp"),
                ("gpm_desc", "", 0, "Description", "Y")),
            "where": [("gpm_cono", "=", self.conum)]}
        r1s = (("Yes", "Y"), ("No", "N"))
        fld = (
            (("T",0,0,0),"ID2",7,"Start Period","Start Period",
                (self.sysdtw/100),"Y",self.doSdate,None,None,("efld",)),
            (("T",0,1,0),"ID2",7,"End Period","End Period",
                (self.sysdtw/100),"Y",self.doEdate,None,None,("efld",)),
            (("T",0,2,0),"IUA",1,"Location","Location",
                "","Y",self.doLoc,loc,None,("efld",)),
            (("T",0,3,0),"IUA",3,"Product Group","Product Group",
                "","Y",self.doGroup,grp,None,None),
            (("T",0,4,0),("IRB",r1s),0,"Exclude Zeros","Exclude Zeros",
                "Y","Y",self.doZero,None,None,None))
        tnd = ((self.doEnd,"Y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("Y","V"), mail=("Y","N"))

    def doSdate(self, frt, pag, r, c, p, i, w):
        self.sdate = w
        self.sdatd = self.df.t_disp[0][0][i]

    def doEdate(self, frt, pag, r, c, p, i, w):
        self.edate = w
        self.edatd = self.df.t_disp[0][0][i]

    def doLoc(self, frt, pag, r, c, p, i, w):
        self.loc = w
        if not w:
            self.locd = "ALL"
        else:
            acc = self.sql.getRec(tables="strloc", cols=["srl_desc"],
                where=[("srl_cono", "=", self.conum), ("srl_loc", "=", w)],
                limit=1)
            if not acc:
                return "Invalid Location"
            self.locd = acc[0]

    def doGroup(self, frt, pag, r, c, p, i, w):
        if not w:
            self.group = ""
        else:
            acc = self.sql.getRec(tables="strgrp", cols=["gpm_desc"],
                where=[("gpm_cono", "=", self.conum), ("gpm_group", "=", w)],
                limit=1)
            if not acc:
                return "Invalid Group"
            self.group = w

    def doZero(self, frt, pag, r, c, p, i, w):
        self.zero = w

    def doEnd(self):
        self.df.closeProcess()
        if not self.group:
            self.sgrp = ""
            self.egrp = "zzz"
        else:
            self.sgrp = self.group; self.egrp = self.group
        all = self.sql.getRec(tables="strmf1", cols=["st1_group", "st1_code",
            "st1_desc", "st1_uoi"], where=[("st1_cono", "=", self.conum),
            ("st1_group", ">=", self.sgrp), ("st1_group", "<=", self.egrp)],
            order="st1_group, st1_code")
        if not all:
            showError(self.mf.body, "Processing Error",
            "No Records Selected")
        elif self.df.repprt[0][2] == "export":
            self.exportReport(all)
        else:
            self.printReport(all)
        self.closeProcess()

    def exportReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        expnam = getModName(self.mf.wrkdir, __name__, self.conum)
        self.expheads = ["%03u %-30s %s" % (self.conum, self.conam,
            self.sysdttm), "Stock Movement Report as at %s for Period "\
            "%s to %s" % (self.sysdtd, self.sdatd, self.edatd)]
        self.expcolsh = [["", "", "", "", ["*** Opening Balances ***", 4, 5],
            ["******* Receipts *******", 6, 7], ["******** Issues ********",
            8, 9], ["*** Closing Balances ***", 10, 11]], ["Grp",
            "Product-Code", "Description", "U.O.I", "Quantity", "Value",
            "Quantity", "Value", "Quantity", "Value", "Quantity", "Value"]]
        self.expforms = [("UA", 3), ("NA", 20), ("NA", 30), ("NA", 10),
            ("SD", 12,2), ("SD", 12.2), ("SD", 12.2), ("SD", 12.2),
            ("SD", 12,2), ("SD", 12.2), ("SD", 12.2), ("SD", 12.2)]
        self.expdatas = []
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                p.closeProgress()
                return
            vals = self.getValues(dat)
            if not vals:
                continue
            self.expdatas.append(["BODY", [vals[0].work, vals[1].work,
                vals[2].work, vals[3].work, vals[4].work, vals[5].work,
                vals[6].work, vals[7].work, vals[8].work, vals[9].work,
                vals[10].work, vals[11].work]])
        p.closeProgress()
        self.grandTotal()
        doWriteExport(xtype=self.df.repprt[0][1], name=expnam, sheet=__name__,
            heads=self.expheads, colsh=self.expcolsh, forms=self.expforms,
            datas=self.expdatas, rcdic=self.mf.rcdic)

    def printReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        self.head = ("%03u %-30s %44s %33s %44s %6s" % (self.conum, self.conam,
            "", self.sysdttm, "", __name__))
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, self.head)
        old_grp = ""
        self.pgnum = 0
        self.pglin = 99
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                if self.t:
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                break
            vals = self.getValues(dat)
            if not vals:
                continue
            grp, code, desc, uoi, t0, t1, t2, t3, t4, t5, t6, t7 = vals
            if old_grp and old_grp != grp.work:
                self.groupTotal()
                self.pglin = 99
            if self.pglin > self.lpp:
                if self.pgnum:
                    self.c.drawText(self.t)
                    self.c.showPage()
                self.pageHeading()
            self.t.textLine("%s %s %s %s %s %s %s %s %s %s %s" % \
                (code.disp, desc.disp, uoi.disp, t0.disp, t1.disp,
                t2.disp, t3.disp, t4.disp, t5.disp, t6.disp, t7.disp))
            self.pglin = self.pglin + 1
            old_grp = grp.work
        p.closeProgress()
        if self.t:
            if can != "yes":
                self.groupTotal()
                self.grandTotal()
            self.c.drawText(self.t)
        self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head=self.tit, view=self.df.repprt[0][1],
            prtnam=self.df.repprt[0][2], mail=self.df.repprt[1][2])
        self.closeProcess()

    def getValues(self, data):
        grp = CCD(data[0], "UA", 3)
        self.groupd = grp.disp
        code = CCD(data[1], "NA", 20)
        desc = CCD(data[2], "NA", 30)
        uoi = CCD(data[3], "NA", 10)
        bals = Balances(self.mf, "STR", self.conum, self.edate,
            keys=(grp.work, code.work, self.loc))
        figs = [0] * 8
        m_ob,m_mv,m_cb,y_ob,y_mv,y_cb,ac,lc = bals.doStrBals(self.sdate)
        figs[0] = y_ob[0]
        figs[1] = y_ob[1]
        for bal in y_mv:
            if bal[0] in (1, 3, 5):
                figs[2] = float(ASD(figs[2]) + ASD(bal[1]))
                figs[3] = float(ASD(figs[3]) + ASD(bal[2]))
            else:
                figs[4] = float(ASD(figs[4]) - ASD(bal[1]))
                figs[5] = float(ASD(figs[5]) - ASD(bal[2]))
        figs[6] = y_cb[0]
        figs[7] = y_cb[1]
        for y in range(0, 8):
            exec "t%s = CCD(figs[y], 'SD', 12.2)" % y
            exec "self.stot[y] = float(ASD(self.stot[y]) + ASD(t%s.work))" % y
            exec "self.gtot[y] = float(ASD(self.gtot[y]) + ASD(t%s.work))" % y
        if self.zero == "Y" and not t0.work and not t1.work and \
            not t2.work and not t3.work and not t4.work and \
            not t5.work and not t6.work and not t7.work: return
        return (grp, code, desc, uoi, t0, t1, t2, t3, t4, t5, t6, t7)
            
    def pageHeading(self):
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.pgnum = self.pgnum + 1
        self.t.textLine(self.head)
        self.t.textLine()
        self.t.textLine("%-68s %90s %5s" % \
            ("Stock Movement Report as at %s for Period "\
            "%s to %s" % (self.sysdtd, self.sdatd, self.edatd),
            "Page", self.pgnum))
        self.t.textLine()
        self.t.textLine("%-5s %3s      %-8s %s  %s" % \
            ("Group", self.groupd, "Location", self.loc, self.locd))
        self.t.textLine()
        self.t.textLine("%-62s %24s  %24s  %24s  %24s" % ("",
            "*** Opening Balances ***", "******* Receipts *******",
            "******** Issues ********", "*** Closing Balances ***"))
        self.t.textLine("%-20s %-30s %-10s %11s  %11s  %11s  %11s  %11s  " \
            "%11s  %11s  %11s " % ("Product-Code", "Description", "U.O.I",
            "Quantity", "Value", "Quantity", "Value",
            "Quantity", "Value", "Quantity", "Value"))
        self.t.textLine("%s" % (self.suc * len(self.head)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        self.pglin = 11

    def groupTotal(self):
        for x in range(0, 8):
            exec "t%s = CCD(self.stot[x], 'SD', 12.2)" % x
        self.t.textLine()
        self.t.textLine("%-20s %-41s %12s %12s %12s %12s %12s %12s %12s " \
            "%12s" % (" ", "Group Totals", t0.disp, t1.disp, t2.disp,
            t3.disp, t4.disp, t5.disp, t6.disp, t7.disp))
        self.stot = [0] * 8

    def grandTotal(self):
        if self.df.repprt[0][2] == "export":
            self.expdatas.append(["ULINES"])
            gtot = ["TOTAL", ["", "", "Grand Totals", ""]]
            gtot[1].extend(self.gtot)
            self.expdatas.append(gtot)
            self.expdatas.append(["ULINED"])
            return
        for x in range(0, 8):
            exec "t%s = CCD(self.gtot[x], 'SD', 12.2)" % x
        self.t.textLine()
        self.t.textLine("%-20s %-41s %12s %12s %12s %12s %12s %12s %12s " \
            "%12s" % (" ", "Grand Totals", t0.disp, t1.disp, t2.disp,
            t3.disp, t4.disp, t5.disp, t6.disp, t7.disp))

    def doExit(self):
        self.df.closeProcess()
        self.closeProcess()

    def closeProcess(self):
        self.mf.closeLoop()

