"""
SYNOPSIS
    These are various functions used in TARTAN.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""
from __future__ import generators

def getPrgPath():
    import os, sys
    prgdir = sys.path[0]
    while not os.path.isdir(prgdir):
        prgdir = os.path.normpath(prgdir.replace(os.path.basename(prgdir), ""))
    return prgdir

########################################
# Graphical User Interface Functions
########################################

def doGtkImport():
    import os, sys
    try:
        # If Linux and DISPLAY variable does not exist, abort
        if sys.platform != "win32" and not os.environ["DISPLAY"]:
            raise Exception
        import pygtk
        if sys.platform == "win32":
            try:
                gtkdir = os.environ["GTK_BASEPATH"]
                if not os.path.isdir(gtkdir):
                    raise Exception
            except:
                gtkdir = os.path.join(getPrgPath(), "gtk")
            if gtkdir and os.path.exists(gtkdir):
                os.environ["GTK_PATH"] = gtkdir
                os.environ["GTK_BASEPATH"] = gtkdir
                if not os.environ["PATH"].count(gtkdir):
                    os.environ["PATH"] = \
                        os.path.join(gtkdir, "lib;") + os.environ["PATH"]
                    os.environ["PATH"] = \
                        os.path.join(gtkdir, "bin;") + os.environ["PATH"]
        else:
            pygtk.require("2.0")
        import gtk, gobject, pango
        return True, gtk, gobject, pango
    except:
        return False, None, None, None

# Standard Message Functions
def showDialog(dtype=None, screen=None, head="", mess="", default=""):
    gui, gtk, gobject, pango = doGtkImport()
    if not gui:
        if dtype == "Q":
            return raw_input("%s " % mess)
        else:
            print mess
            return
    if dtype:
        if screen:
            screen = screen.get_toplevel()
        if dtype == "C":
            tit = "Choice"
            dia = gtk.MESSAGE_QUESTION
            but = gtk.BUTTONS_NONE
        elif dtype == "Q":
            tit = "Question"
            dia = gtk.MESSAGE_QUESTION
            but = gtk.BUTTONS_YES_NO
        else:
            if dtype == "E":
                tit = "Error Message"
                dia = gtk.MESSAGE_ERROR
            elif dtype == "I":
                tit = "Information"
                dia = gtk.MESSAGE_INFO
            elif dtype == "W":
                tit = "Warning"
                dia = gtk.MESSAGE_WARNING
            but = gtk.BUTTONS_CLOSE
        flg = gtk.DIALOG_DESTROY_WITH_PARENT
        dd = gtk.MessageDialog(parent=screen, type=dia, buttons=but, flags=flg)
        dd.set_title(tit)
        dd.set_markup("<b><big>%s</big></b>\n\n%s" % \
            (head.replace("&", "&amp;"), mess.replace("&", "&amp;")))
        if dtype == "C":
            return dd
        if default == "yes":
            dd.set_default_response(gtk.RESPONSE_YES)
        elif default == "no":
            dd.set_default_response(gtk.RESPONSE_NO)
        ok = gtk.RESPONSE_DELETE_EVENT
        while ok == gtk.RESPONSE_DELETE_EVENT:
            ok = dd.run()
        dd.destroy()
        while gtk.events_pending():
            gtk.main_iteration()
        if dtype == "Q":
            if ok == gtk.RESPONSE_YES:
                return "yes"
            else:
                return "no"

def showError(screen=None, head="", mess=""):
    showDialog("E", screen, head, mess)

def showInfo(screen=None, head="", mess=""):
    showDialog("I", screen, head, mess)

def showWarning(screen=None, head="", mess=""):
    showDialog("W", screen, head, mess)

def askQuestion(screen=None, head="", mess="", default="yes"):
    return showDialog("Q", screen, head, mess, default)

def askChoice(screen=None, head="", mess="", butt=None, default=None):
    if not butt:
        return
    diag = showDialog("C", screen, head, mess, default)
    for but in butt:
        diag.add_button(but[0], but[1])
    diag.set_default_response(default)
    ok = None
    while not ok:
        ok = diag.run()
    diag.destroy()
    return ok

def doLabel(gtk, win, text="", size=None, style="etched_out", color=True, font=None):
    # Create eventbox, label and set size, font, style and color
    evt = gtk.EventBox()
    if size:
        fmt = "%-" + str(size) + "s"
        text = fmt % text
    lab = gtk.Label(text.replace("&", "&amp;"))
    lab.set_use_markup(True)
    if font:
        lab.modify_font(font)
    if style:
        frm = gtk.Frame()
        frm.add(lab)
        if style == "out":
            frm.set_shadow_type(gtk.SHADOW_OUT)
        elif style == "in":
            frm.set_shadow_type(gtk.SHADOW_IN)
        elif style == "etched_out":
            frm.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        elif style == "etched_in":
            frm.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        else:
            frm.set_shadow_type(gtk.SHADOW_NONE)
        evt.add(frm)
    else:
        evt.add(lab)
    if color:
        ss = win.get_style().copy()
        evt.modify_bg(gtk.STATE_NORMAL, ss.bg[gtk.STATE_SELECTED])
        lab.modify_fg(gtk.STATE_NORMAL, ss.fg[gtk.STATE_SELECTED])
    return (evt, lab)

def getSingleRecords(mf, table, cols, title=None, head=None, where=None, group=None, order=None, selcol=None, ttype="T"):
    """
    This is a gui used to select individual records from a table
    returning all columns for the selected record

    table   The table(s) from which to return the records.
            In the case of multiple tables, only the first table's columns
            will be returned, all other tables must only be used for the
            where statement.
    cols    A list of columns to display in the select view.
    title   The title.
    head    A list of column headings.
    where   A list of where conditions else all records or a list of data to
            use in the selection.
    group   True or False
    order   The order in which the records are to be displayed else
            the order will be cols.
    selcol  The column for keyboard selection.
    ttype   T = The where above is a list of where conditions.
            D = The where above is a list of data.
    """
    from TartanClasses import CCD, SChoice, Sql

    if not title:
        title = "Available Records"
    if type(table) == str:
        table = [table]
    sql = Sql(mf.dbm, table, prog=__name__)
    if ttype == "D":
        all = where
        col = cols
        exec "tmp = sql.%s_dic" % table[0]
        dic = {}
        for key in tmp:
            if key in col:
                dic[key] = tmp[key]
    else:
        exec "dic = sql.%s_dic" % table[0]
        exec "col = sql.%s_col" % table[0]
        if not order:
            ord = ""
            for c in cols:
                if not ord:
                    ord = c
                else:
                    ord = "%s, %s" % (ord, c)
        else:
            ord = order
        if group:
            grp = ""
            for c in col:
                if not grp:
                    grp = c
                else:
                    grp = "%s, %s" % (grp, c)
        else:
            grp = None
        all = sql.getRec(tables=table,cols=col,where=where,group=grp,order=ord)
    if not all:
        showError(mf.body, "Error", "No Records Available")
        return []
    sel = []
    lines = []
    types = [("CB", 1)]
    for num, acc in enumerate(all):
        lin = []
        for c in cols:
            if not num:
                types.append((dic[c][2], dic[c][3]))
            lin.append(CCD(acc[col.index(c)], dic[c][2], dic[c][3]).work)
        lines.append(lin)
    if selcol:
        selc = list(cols).index(selcol)
    else:
        selc = None
    if mf.fixed:
        deco = False
    else:
        deco = True
    sc = SChoice(mf, scrn=mf.body, deco=deco, titl=title, head=head,
        typs=types, data=lines, rows=None, retn="I", mode="M", selc=selc)
    if sc.selection:
        for idx in sc.selection:
            sel.append(all[idx])
    return sel

def aboutTartan(window, version):
    gui, gtk, gobject, pango = doGtkImport()
    if not gui:
        return

    def showChanges(*args):
        try:
            from tarchg import changes
            from TartanClasses import ScrollText
            ScrollText(title="List of Changes", mess=changes)
        except:
            pass

    def aboutClose(*args):
        if args[1] == 99:
            showChanges()
            return True
        args[0].destroy()
        gtk.main_quit()

    title = "TARTAN SYSTEMS"
    body = \
        "For information and or support for this\n" \
        "product please contact:\n"
    contact = "\nPaul Malherbe\n\n"
    phone = "Phone:     27-21-6711866\n"
    fax =   "Fax:       27-21-6711866\n"
    email = "Email: info@tartan.co.za\n"
    info = body + contact + phone + fax + email
    copyright = "Copyright (C) 2004-2012 Paul Malherbe."
    dia = gtk.AboutDialog()
    dia.set_name(title)
    dia.set_version(version)
    dia.set_copyright(copyright)
    dia.set_website("http://www.tartan.co.za")
    dia.set_authors(["Paul Malherbe"])
    dia.set_comments(info)
    dia.set_license("""
Free use of this granted under the terms of the GNU General Public License (GPL) as
published by the Free Software Foundation, either version 2 of the version 2
of the License, or (at your option) any later version.

TARTAN is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY, without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
TARTAN, if not, write to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA or view/download it at
http://www.gnu.org/licenses/gpl.txt
""")
    dia.connect("response", aboutClose)
    dia.add_button("Changes", 99)
    dia.set_transient_for(window)
    dia.set_decorated(False)
    dia.show()
    gtk.main()

def setTheme(gtk, theme, fontm, fontd):
    import os
    # Parse theme
    my_theme = None
    thmdir = gtk.rc_get_theme_dir()
    for thm in (theme, "Clearlooks", "Metal", "Default"):
        pth = os.path.join(thmdir, thm, "gtk-2.0", "gtkrc")
        if os.path.isfile(pth):
            my_theme = pth
            break
    if my_theme:
        style = "include '%s'\n" % my_theme
        gtk.rc_parse_string(style)
    else:
        style = ""
    # Set personalised fonts and colours
    window = gtk.Window()
    window.show_all()
    cmap = window.get_colormap()
    csty = window.get_style().copy()
    for c in ("base", "text"):
        exec "r = cmap.alloc_color(csty.%s[gtk.STATE_NORMAL]).red" % c
        exec "g = cmap.alloc_color(csty.%s[gtk.STATE_NORMAL]).green" % c
        exec "b = cmap.alloc_color(csty.%s[gtk.STATE_NORMAL]).blue" % c
        exec "%s = [int(r), int(g), int(b)]" % c
    window.destroy()
    args = [fontd] + base + text + base + text + ["gray", "lightgray", "white", fontm]
    style = style + """
style "tartan"
{
  font_name = "%s"
  bg[INSENSITIVE] = {%s, %s, %s}
  fg[INSENSITIVE] = {%s, %s, %s}
  base[INSENSITIVE] = {%s, %s, %s}
  text[INSENSITIVE] = {%s, %s, %s}
  GtkWidget::interior_focus = 1
}
widget "*" style "tartan"

style "tarbut"
{
  bg[NORMAL] = "%s"
  bg[PRELIGHT] = "%s"
  bg[INSENSITIVE] = "%s"
}
widget "*MyButton*" style "tarbut"

style "tarmen"
{
  font_name = "%s"
}
widget "*Menu*" style "tarmen"

binding "gtk-unix-text-entry"
{
  bind "<ctrl>u" { "delete-from-cursor" (paragraph-ends, -1) }
  bind "<ctrl>h" { "delete-from-cursor" (chars, -1) }
  bind "<ctrl>w" { "delete-from-cursor" (word-ends, -1) }
}
class "GtkEntry" binding "gtk-unix-text-entry"
class "GtkTextView" binding "gtk-unix-text-entry"
""" % tuple(args)
    gtk.rc_parse_string(style)
    return gtk.settings_get_default()

########################################
# Standard, non GUI Tartan Functions
########################################

def showException(scrn, path, mess, maxTBlevel=0, exit=None, dbm=None):
    """
    Display an exception

    scrn = The Screen on which to Display the Exception or None to print
    path = The place where the log file resides
    mess = The First Line of the Message
    exit = A command to be executed on exiting the exception frame
    dbm  = The database class
    """
    import os, sys, time, traceback

    def doSaveExc():
        t = time.localtime()
        t = time.strftime("Date: %Y-%m-%d Time: %H:%M:%S", t)
        fnam = os.path.join(path, "errors.txt")
        if os.path.isfile(fnam):
            flenam = open(fnam, "a")
        else:
            flenam = open(fnam, "w")
        flenam.write(t + "\n\n")
        flenam.write(mess + "\n")
        flenam.write("*******************************" + "\n")
        flenam.close()
        if dbm:
            from TartanClasses import Sql
            dbm.openDbase()
            try:
                if sys.platform == "win32":
                    import win32api
                    host = win32api.GetComputerName()
                    user = win32api.GetUserName()
                else:
                    import getpass
                    host = os.uname()[1]
                    user = getpass.getuser()
                sql = Sql(dbm, tables=["ctlmst", "ctlsys"], prog=__name__)
                if sql.error:
                    raise Exception
                smtp = sql.getRec(tables=["ctlmst", "ctlsys"], cols=["ctm_name",
                    "sys_smtp", "ctm_email"], where=[("ctm_cono", "=", 1)],
                    limit=1, excp=True)
                sendMail(smtp[1], smtp[2], ["errors@tartan.co.za"],
                    "Exception Message from Company: %s Host: %s User: %s" % \
                    (smtp[0], host, user), attach=[fnam])
                os.remove(fnam)
            except:
                pass
            dbm.closeDbase()

    title = str(sys.exc_type)
    cla, exc, trbk = sys.exc_info()
    if type(cla) == str:
        excName = cla
    else:
        excName = cla.__name__
    if type(exc) == str or not exc.__dict__:
        excArgs = exc
    else:
        try:
            excArgs = exc.__dict__["args"]
        except KeyError:
            excArgs = "<no args>"
    if maxTBlevel:
        excTb = traceback.format_tb(trbk, maxTBlevel)
    else:
        excTb = traceback.format_tb(trbk)
    excMess = ""
    for ta in excTb:
        excMess = excMess + ta + "\012"
    mess = mess + "\012\012%s\012%s\012%s" % (excMess, excName, excArgs)
    if path:
        doSaveExc()
    if not scrn:
        gui, gtk, gobject, pango = doGtkImport()
        if not gui:
            print mess
            return
    from TartanClasses import ScrollText
    st = ScrollText(scrn=scrn, title=title, mess=mess, select=True, exit=exit)

def doReportlab(prgdir, name, head=None, orient="P", fonts=None):
    """
    A routine to set the parameters for reportlab form generation.

    prgdir = The base directory for ttf font directory (fnt)
    name   = The pdf file name to create
    head   = The report heading used to obtain the width of the report
    orient = The report layout i.e. P=Portrait, L=Landscape
    fonts  = A tuple of ttf fonts to use i.e. (Font-Normal, Font-Bold)

    It returns some or all of the following:

        cvs  = canvas
        ogn  = origination
        cpi  = characters per inch
        ldg  = vertical offset between lines
        lpp  = lines per page
        suc  = single undeline character
        duc  = double undeline character
        inch = number of pixels per inch
    """
    import os, sys
    from reportlab.lib import pagesizes
    from reportlab.lib.units import inch
    from reportlab.pdfgen import canvas
    from reportlab.pdfbase.pdfmetrics import registerFont, Font, stringWidth
    from reportlab.pdfbase.ttfonts import TTFont

    # Register Fonts
    encLabel = "WinAnsiEncoding"
    if not fonts:
        fonts = (
            os.path.join(prgdir, "fnt", "cour.ttf"),
            os.path.join(prgdir, "fnt", "courbd.ttf"))
    try:
        registerFont(TTFont("Tartan", fonts[0]))
        registerFont(TTFont("Tartan-Bold", fonts[1]))
        # Single and double underline characters
        suc = u"\u2500"
        duc = u"\u2550"
    except:
        registerFont(Font("Tartan", "Courier", encLabel))
        registerFont(Font("Tartan-Bold", "Courier-Bold", encLabel))
        # Single and double underline characters
        suc = duc = chr(151).decode(encLabel, "ignore").encode("utf8")

    # Defaults
    ogn = (15, 20)
    cpi = 10
    ldg = 12.7
    lpp = 60

    if not orient or orient == "P":
        page = pagesizes.portrait(pagesizes.A4)
    elif orient == "L":
        page = pagesizes.landscape(pagesizes.A4)
    else:
        exec "page = pagesizes.portrait(pagesizes.%s)" % orient
    if not head:
        cvs = canvas.Canvas(name, bottomup=0, pagesize=page)
        cvs.setFont("Tartan", cpi, ldg)
        return(cvs, cpi, inch, suc, duc)
    if head:
        # Adjust to heading width
        if len(head) > 120:
            page = pagesizes.landscape(pagesizes.A4)
            ogn = (20, 20)
            ldg = 9
            lpp = 60
        # Adjust font size
        for npi in range(120, 59, -1):
            cpi = round(npi / 10.0, 1)
            if stringWidth(head, "Tartan", cpi) <= (page[0] - 30):
                break
        cvs = canvas.Canvas(name, bottomup=0, pagesize=page)
        cvs.setFont("Tartan", cpi, ldg)
        return(cvs, ogn, cpi, ldg, lpp, suc, duc)

def makeArray(col, row, dth, typ="I"):
    """
    A routine to generate an array of lists

    col = Number of Columns
    row = Number of Rows
    dth = Number of Levels
    typ = "D" for Float, "I" for Integer, "S" for String
    """
    import copy
    a = []
    b = []
    c = []
    if typ == "S":
        d = ""
    elif typ == "D":
        d = 0.0
    else:
        d = 0
    for x in range(col):
        a.append(d)
    for x in range(row):
        b.append(copy.copy(a))
    if dth < 2:
        c = b
    else:
        for x in range(dth):
            c.append(copy.deepcopy(b))
    return(c)

def getPeriods(mf, conum, period, check=False):
    """
    A routine to extract a financial period's start and end date for a Company

    mf     = The MainFrame object
    conum  = The company number
    period = The period number
    check  = Check if a Year End needs to be run
    """
    from TartanClasses import CCD, Sql

    if mf.display:
        sc = mf.body
    else:
        sc = None
    sql = Sql(mf.dbm, ["ctlynd", "gentrn"], prog=__name__)
    p = sql.getRec(tables="ctlynd", cols=["cye_start", "cye_end", "cye_final"],
        where=[("cye_cono", "=", conum), ("cye_period", "=", period)], limit=1)
    if not p:
        showError(sc, "Period Error",
            "Invalid Financial Period %s" % period)
        return (None, None, None)
    if period and check:
        chk1 = sql.getRec(tables="ctlynd", cols=["cye_last"],
            where=[("cye_cono", "=", conum), ("cye_period", "=", (period - 1))],
            limit=1)
        chk2 = sql.getRec(tables="gentrn", cols=["max(glt_capdt)"],
            where=[("glt_cono", "=", conum), ("glt_curdt", "<", (p[0] / 100))],
            limit=1)
        if chk2[0] and chk2[0] > chk1[0]:
            ok = askQuestion(sc, "Year End Error", """A Year End Must be Performed for Period %s as Transactions Have Been Captured After the last Year End Run for that Period!\n\nIf Not Performed, certain Reports and Queries will have Incorrect Balances!\n\nDo You Want to Continue Anyway?""" % (period-1), default="no")
            if ok == "no":
                return (None, None, None)
    s_date = CCD(p[0], "D1", 10.0)
    e_date = CCD(p[1], "D1", 10.0)
    return (s_date, e_date, p[2])

def getModName(pth, prg, key, ext=None, wait=0):
    """
    A routine to return a name for a printer output file which is unique ??

    pth   = The path where the file is to be created
    prg   = The calling module name
    key   = The key of the report e.g. Company Number/Account Number
    ext   = The file extension e.g. pdf or None
    wait  = The number of seconds to pause bwfore returning the name
    """
    import os, time

    if wait:
        time.sleep(wait)
    t = time.localtime()
    tim = "%02i%02i%02i%02i%02i%02i" % \
        (int(str(t[0])[2:]), t[1], t[2], t[3], t[4], t[5])
    key = str(key).replace("/","-").replace("\\","-").replace(" ", "_")
    if not ext:
        return os.path.join(pth, "%s_%s_%s" % (prg, key, tim))
    if ext and type(ext) == str:
        return os.path.join(pth, "%s_%s_%s.%s" % (prg, key, tim, ext))
    nam = []
    for e in ext:
        nam.append(os.path.join(pth, "%s_%s_%s.%s" % (prg, key, tim, e)))
    return nam

def doPrinter(mf=None, conum=None, rep=None, head=None, view=None, prtnam=None, fadd=None, mail=None, splash=True, skip=False, noprt=False):
    """
    This module Views or Prints and/or Mails a Report

     mf     = MainFrame object
     conum  = Company Number
     rep    = Report PDF File Name
     head   = Heading Used as Subject of Mail
     view   = Output Report to (V)iew, (P)rinter ot e(X)port
     prtnam = The Printer Name, view, export or None
     fadd   = The from address if not the default company one.
     mail   = Mail Addresses delimited by , or ("N", "" or None) for No Mail.
              This can also be a list of mail addresses.
     splash = Splash Screen
     skip   = Skip mail errors
     noprt  = Do not print the document, mail only
     """
    import getpass, os, sys, time
    from TartanClasses import AskForData, Dbase, SplashScreen, Sql

    # Email Document
    if not mail in (None, "", "N", "n"):
        if not fadd:
            dbm = Dbase(mf.rcdic)
            dbm.openDbase()
            sql = Sql(dbm, tables=["ctlsys", "ctlmst", "emllog"], prog=__name__)
            if not conum:
                conum = 1
            smtp = sql.getRec(tables=["ctlmst", "ctlsys"], cols=["sys_smtp",
                "ctm_email"], where=[("ctm_cono", "=", conum)], limit=1)
            fadd = smtp[1]
        if not fadd:
            try:
                fadd = "%s@%s" % (getpass.getuser(), smtp[1])
            except:
                pass
        if type(mail) in (tuple, list):
            pass
        else:
            mail = mail.split(",")
        ok = None
        sp = None
        if type(head) in (tuple, list):
            if len(head) == 1:
                head = head[0]
            elif len(head) == 2:
                head = "%s %s" % tuple(head)
            elif len(head) == 3:
                head = "%s %s %s" % tuple(head)
        while not ok:
            if mf.display and splash:
                sp = SplashScreen(mf.body,
                    "E-Mailing the Report to %s\n\nPlease Wait........" %
                        mail[0])
            try:
                att = [rep]
                sendMail(smtp[0], fadd, mail, head, head, att)
                ok = "OK"
            except (KeyboardInterrupt, SystemExit):
                ok = "ABORTED"
            except:
                if skip:
                    ok = "SKIPPED"
                else:
                    if mf.display:
                        sc = mf.body
                    else:
                        sc = None
                    ok = askQuestion(sc, "E-Mail Error", "Problem Delivering "\
                        "This Message.\n\nTo: %s\nSubject: %s\n\nWould You "\
                        "Like to Retry?" % (mail, head))
                    if ok == "yes":
                        ok = None
                    else:
                        ok = "FAILED"
            if sp:
                sp.closeSplash()
        # Try and Log the email status into table emllog
        try:
            for tadd in mail:
                sql.insRec("emllog", data=[fadd.strip(), tadd.strip(),
                    head.strip(), "%04i-%02i-%02i %02i:%02i" %
                    time.localtime()[0:5], ok], excp=True)
            dbm.commitDbase()
        except:
            pass
        dbm.closeDbase()
        if noprt:
            return
    # Print Document
    sp = None
    try:
        if view.lower() == "x" or prtnam == "export":
            return
        if view.lower() == "v" or prtnam == "view":
            exe, cmd = parsePrg(mf.rcdic["vwr"])
            # Execute viewer command
            cmd.append(rep)
            os.spawnv(os.P_WAIT, exe, tuple(cmd))
        elif prtnam and prtnam.lower() != "none":
            if mf.display and splash:
                sp = SplashScreen(mf.body,
                    "Printing the Report\n\n           Please Wait....")
            if not mf.rcdic["cnv"] or mf.rcdic["cnv"] == ".":
                pfl = rep
            else:
                # Convert pdf to ps
                exe, cmd = parsePrg(mf.rcdic["cnv"])
                cmd.append(rep)
                pfl = rep.replace(".pdf", ".ps")
                cmd.append(pfl)
                os.spawnv(os.P_WAIT, exe, tuple(cmd))
            exe, cmd = parsePrg(mf.rcdic["prn"])
            if sys.platform == "win32":
                # Windows
                if prtnam == "Default":
                    import win32print
                    prtnam = win32print.GetDefaultPrinter()
                if exe.lower().count("acrord32"):
                    cmd.extend(["/t", pfl, '"%s"' % prtnam])
                elif exe.lower().count("foxit"):
                    cmd.extend(["/t", pfl, '"%s"' % prtnam])
                elif exe.lower() == "gsprint.exe":
                    cmd.extend(["-printer", '"%s"' % prtnam, pfl])
                elif exe.lower().count("pdfp"):
                    cmd.extend(["-p", '"%s"' % prtnam, pfl])
                elif exe.lower() == "redpr.exe":
                    cmd.extend(["-P", '"%s"' % prtnam, pfl])
                elif exe.lower().count("sumatra"):
                    cmd.extend(["-print-to", '"%s"' % prtnam, pfl])
                else:
                    cmd.append(pfl)
            # Linux
            elif prtnam == "Default":
                cmd.append(pfl)
            else:
                cmd.extend([("-P%s" % prtnam), (pfl)])
            # Execute print command
            os.spawnv(os.P_NOWAIT, exe, tuple(cmd))
            if sp:
                sp.closeSplash()
    except:
        if sp:
            sp.closeSplash()
        if mf.display:
            sc = mf.body
        else:
            sc = None
        showException(sc, mf.rcdic["wrkdir"], "Output Error")
        return

def parsePrg(text):
    prg = text.split(" -")
    exe = prg[0].strip()
    if prg[0].count(" "):
        cmd = ['"%s"' % prg[0].strip()]
    else:
        cmd = [prg[0].strip()]
    if len(prg) > 1:
        d = False
        for x in prg[1:]:
            if x == "-":
                d = True
            elif d:
                d = False
                cmd.append("--%s" % x.strip())
            else:
                cmd.append("-%s" % x.strip())
    return exe, cmd

def textFormat(text, width=30):
    """
    Format a text block

    Arguments:
        `text`   -- the string to be reformatted
        `width`  -- the maximum width of formatted text
    """
    if text:
        import textwrap
        return textwrap.wrap(text, width=width)
    else:
        return [""]

def findFile(start=".", name=None, ftyp="f", case="n"):
    """
    A routine used to find a file given the starting path, name and whether
    name refers to a (f)ile or (d)irectory and whether case sensitive or not.
    """
    import os

    if type(start) == str:
        start = [start]
    name = os.path.basename(name)
    for d in start:
        for root, dirs, files in os.walk(d):
            if ftyp == "d":
                for pth in dirs:
                    if case == "y" and os.path.basename(pth) == name:
                        return os.path.join(root, pth)
                    elif os.path.basename(pth).lower() == name.lower():
                        return os.path.join(root, pth)
            else:
                for flenam in files:
                    if case == "y" and flenam == name:
                        return os.path.join(root, flenam)
                    elif flenam.lower() == name.lower():
                        return os.path.join(root, flenam)

def sendMail(server, ex, to, subj, mess="", attach=[], embed=[], timeout=30):
    """
    A routine to email a message plus a file as an attachment

        server  = The smtp server e.g. smtp.abc.com|username|password
        ex      = The from address
        to      = A list of recipients
        subj    = The subject
        mess    = The Message
        attach  = A list of attachments
        embed   = A list of attachments to embed in the body of the message.
        timeout = The number of seconds before timing out defaulting to 60.
    """
    import mimetypes, os, smtplib, socket, sys
    if sys.version.split()[0] < "2.6.0":
        from email import Encoders
        from email.MIMEBase import MIMEBase
        from email.MIMEText import MIMEText
        from email.MIMEImage import MIMEImage
        from email.MIMEMultipart import MIMEMultipart
        from email.Utils import COMMASPACE, formatdate
    else:
        from email import encoders as Encoders
        from email.mime.base import MIMEBase
        from email.mime.text import MIMEText
        from email.mime.image import MIMEImage
        from email.mime.multipart import MIMEMultipart
        from email.Utils import COMMASPACE, formatdate

    if type(to) == str:
        to = [to]
    if type(attach) == str:
        attach = [attach]
    if type(embed) == str:
        embed = [embed]
    msgRoot = MIMEMultipart("related")
    msgRoot["From"] = ex
    msgRoot["To"] = COMMASPACE.join(to)
    msgRoot["Date"] = formatdate(localtime=True)
    msgRoot["Subject"] = subj
    msgAlternative = MIMEMultipart("alternative")
    msgRoot.attach(msgAlternative)
    if mess:
        if type(mess) in (list, tuple):
            msgText = MIMEText(mess[0], "plain", "utf-8")
            mess = mess[1]
        else:
            msgText = MIMEText(mess, "plain", "utf-8")
        msgAlternative.attach(msgText)
    if embed:
        seq = 1
        for flenam in embed:
            name = os.path.basename(flenam)
            ctype, encoding = mimetypes.guess_type(flenam)
            maintype, subtype = ctype.split("/", 1)
            if maintype == "image":
                try:
                    fp = open(flenam, "rb")
                    msgImage = MIMEImage(fp.read(), _subtype=subtype, name=name)
                    fp.close()
                    msgImage.add_header("Content-ID", "<image%s>" % seq)
                    msgRoot.attach(msgImage)
                    mess = mess + "<img src='cid:image%s'><br><br/>" % seq
                    seq += 1
                except:
                    pass
    msgText = MIMEText(mess, "html", "utf-8")
    msgAlternative.attach(msgText)
    if attach:
        for flenam in attach:
            if os.path.isfile(os.path.normpath(flenam)):
                name = os.path.basename(flenam)
                ctype, encoding = mimetypes.guess_type(flenam)
                if ctype:
                    maintype, subtype = ctype.split("/", 1)
                else:
                    maintype, subtype = ("application", "octet-stream")
                part = MIMEBase(maintype, subtype, name=name)
                part.set_payload( open(os.path.normpath(flenam),"rb").read() )
                Encoders.encode_base64(part)
                part.add_header("Content-Disposition",
                    'attachment; filename="%s"' % os.path.basename(flenam))
                msgRoot.attach(part)
    server = server.split("|")
    if server[0] == "smtp.gmail.com" and len(server) == 3:
        smtp = smtplib.SMTP(server[0], 587)
        smtp.ehlo()
        smtp.starttls()
        smtp.login(server[1], server[2], timeout=timeout)
    elif len(server) == 3:
        smtp = smtplib.SMTP(server[0])
        smtp.login(server[1], server[2], timeout=timeout)
    else:
        smtp = smtplib.SMTP(server[0], timeout=timeout)
    smtp.sendmail(ex, to, msgRoot.as_string())
    smtp.quit()

def mthendDate(date):
    """
    Function to return the month-end date of a date
    """
    from tartanWork import mthnam

    yy = date / 10000
    mm = (date % 10000) / 100
    if mm == 2 and not (yy % 4):
        date = (yy * 10000) + (mm * 100) + 29
    else:
        date = (yy * 10000) + (mm * 100) + mthnam[mm][2]
    return date

def projectDate(date, days, act="+"):
    """
    Function to return a date which is the result of a date +- days
    """
    import copy
    from tartanWork import mthnam

    chkmth = copy.deepcopy(mthnam)
    adate = str(date)
    yy = int(adate[0:4])
    mm = int(adate[4:6])
    if act == "+":
        dd = int(adate[6:8]) + days
        while days:
            if (yy % 4) == 0:
                chkmth[2][2] = 29
            else:
                chkmth[2][2] = 28
            if dd > chkmth[mm][2]:
                dd -= chkmth[mm][2]
                mm += 1
                if mm > 12:
                    mm = 1
                    yy += 1
            else:
                break
    else:
        dd = int(adate[6:8]) - days
        while days:
            if dd <= 0:
                mm -= 1
                if mm == 0:
                    mm = 12
                    yy -= 1
                if (yy % 4) == 0:
                    chkmth[2][2] = 29
                else:
                    chkmth[2][2] = 28
                dd += chkmth[mm][2]
            else:
                break
    prjdate = (yy * 10000) + (mm * 100) + dd
    return prjdate

def paymentDate(base, stday, terms, trdt):
    """
    Function to return the due date of a transaction
    """
    if not terms:
        paydt = trdt
    elif base == "D":
        paydt = projectDate(trdt, terms)
    else:
        styr = trdt / 10000
        stmt = trdt / 100 % 100
        if stday in (0, 30):
            stday = 31
        if trdt % 100 > stday:
            stmt += 1
            if stmt > 12:
                styr += 1
                stmt = 1
        lmed = mthendDate((styr*10000) + (stmt*100) + 1)
        if stday == 31 or stday > lmed % 100:
            stdt = lmed
        else:
            stdt = (styr*10000) + (stmt*100) + stday
        pyyr = stdt / 10000
        pymt = stdt / 100 % 100
        if terms in (30, 31):
            pymt += 1
            if pymt > 12:
                pyyr += 1
                pymt = 1
            lmed = mthendDate((pyyr*10000) + (pymt*100) + 1)
            paydt = (pyyr*10000) + (pymt*100) + stday
            if stday == 31 or paydt > lmed:
                paydt = lmed
        else:
            paydt = projectDate(stdt, terms)
    return paydt

def checkSunday(paydt):
    import time
    try:
        if time.strftime("%A",time.strptime(str(paydt),"%Y%m%d")) == "Sunday":
            pyyr = paydt / 10000
            pymt = ((paydt / 100) % 100)
            pydy = (paydt % 100) - 1
            if not pydy:
                pymt -= 1
                paydt = mthendDate((pyyr * 10000) + (pymt * 100) + 1)
            else:
                paydt = (pyyr * 10000) + (pymt * 100) + pydy
    except:
        paydt = 0
    return paydt

def dateToJulian(date):
    """
    Function to return a julian date from an iso date
    """
    cmon = [0,31,59,90,120,151,181,212,243,273,304,334]
    adate = str(date)
    dt = int(adate)
    y = dt / 10000
    m = (dt / 100) - (y * 100)
    d = dt % 100
    l = y / 4
    r = y % 4
    if r != 0:
        d = d + 1
    jdate = (y * 365) + l + cmon[m - 1] + d
    if r == 0 and m > 1:
        jdate = jdate + 1
    return jdate

def julianToDate(jdate):
    """
    Function to return an iso date from a julian date
    """
    import copy
    from tartanWork import mthnam

    chkmth = copy.deepcopy(mthnam)
    adate = str(jdate)
    dt = int(adate)
    y = dt / 365
    d = (dt % 365)
    l = y / 4
    d = d - l
    r = y % 4
    if r != 0:
        d = d - 1
    while d <= 0:
        y = y - 1
        r = y % 4
        if r == 0:
            d = d + 366
        else:
            d = d + 365
    r = y % 4
    m = 0
    if r == 0:
        chkmth[2][2] = 29
    else:
        chkmth[2][2] = 28
    while d > chkmth[m][2]:
        m = m + 1
        d = d - chkmth[m - 1][2]
    date = (y * 10000) + (m * 100) + d
    return date

def dateDiff(date1, date2, ptype="years"):
    """
    Function to return the differance between to dates.
        date1 = Starting Date
        date2 = Ending Date
    """
    import datetime, sys, time

    def M(d):
        return ((d.year * 12) + d.month)

    def D(d1, d2):
        return (d1.toordinal() - d2.toordinal())

    def I(ymd):
        return datetime.date(*time.strptime(ymd, "%Y%m%d")[:3])

    date1 = I(str(date1))
    date2 = I(str(date2))
    days = D(date2, date1)
    if days <= 0:
        return 0
    months = M(date2) - M(date1)
    if ptype.lower() == "years":
        if date2.day < date1.day:
            months -= 1
        return months / 12
    if ptype.lower() == "months":
        return  months
    if ptype.lower() == "days":
        return  days

def getUnderline(labs, unl, txt, sav=False):
    """
    labs = dictionaries of already assigned underline characters
    unl  = dictionary name, of already assigned underline characters, to use
    txt  = text to find underline for
    sav  = keep existing underlined character
    """
    if sav and txt.count("_"):
        return labs, txt
    for c in txt:
        # If the label is Exit, try and make x the hot key
        if txt == "Exit" and c == "E":
            continue
        if (ord(c) in range(65, 91) or \
            ord(c) in range(97, 123)) and \
            (labs[unl].count(c.lower()) == 0 and \
            labs[unl].count(c.upper()) == 0):
            labs[unl].append(c)
            pos = txt.index(c)
            lab = txt[:pos] + "_" + txt[pos:]
            return labs, lab
    return labs, txt

def showPrinters():
    """
    Function to return all avalaible printers.
    """
    import sys
    data = ["None", "Default"]
    if sys.platform == "win32":
        import win32print
        try:
            lst = win32print.EnumPrinters(2)
            for l in lst:
                data.append(l[2].strip())
        except:
            pass
    else:
        import subprocess
        try:
            proc = subprocess.Popen("lpstat -a", shell=True, bufsize=0,
                stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True)
            lst = proc.stdout.readlines()
            for l in lst:
                l = l.replace("\n", "").replace('"', "").replace("'", "")
                if l.count("accepting requests"):
                    p = l.split()
                    data.append(p[0].strip())
        except:
            pass
    return data

def loadRcFile(rcfile=None, default=False):
    """
    Function to load a tartanrc file and return a dictionary.

        rcfile - The path of the file to load defaulting to tartanrc.
        default - Whether to return default settings if no file.
    """

    import getpass, os, sys
    # Directories and Paths
    dirs = ("dbdir","prgdir","arcdir","bupdir","wrkdir","vwr","prn","cnv","exp")
    # Read, load rcfile
    if not rcfile:
        rcfile = os.environ.get("TARTANRC")
        if not rcfile:
            if sys.platform == "win32":
                if os.path.isfile(os.path.normpath("C:/tartanrc")):
                    rcfile = os.path.normpath("C:/tartanrc")
                else:
                    rcfile = os.path.normpath("C:/Tartan/tartanrc")
            else:
                home = os.environ.get("HOME")
                if not home:
                    home = "/home/%s" % getpass.getuser()
                rcfile = os.path.join(home, ".tartanrc")
    rcdic = {"name": rcfile}
    try:
        f = open(rcfile, "r")
        data = f.readlines()
        f.close()
        for a in data:
            if a[0] in ("#", ";"):
                continue
            a = a.replace("'", "")
            a = a.replace("[", "").replace("]", "")
            a = a.replace(chr(10), "").replace(chr(13), "")
            a = a.split(" = ")
            if a[0].strip() in ("w1", "w2", "bg", "eb", "ef"):
                continue
            if a[0].strip() == "t1":
                a[0] = "mft"
            elif a[0].strip() == "s1":
                a[0] = "mfs"
            elif a[0].strip() == "t2":
                a[0] = "dft"
            elif a[0].strip() == "s2":
                a[0] = "dfs"
            elif a[0].strip() == "dbdsn":
                a[0] = "dbdir"
            if a[0].strip() and a[0].strip() in dirs:
                path = os.path.normpath(a[1].strip())
                if path == ".":
                    rcdic[a[0].strip()] = ""
                else:
                    rcdic[a[0].strip()] = os.path.normpath(a[1].strip())
            else:
                rcdic[a[0].strip()] = a[1].strip()
        if not "exp" in rcdic:
            rcdic["exp"] = ""
        if not "cnf" in rcdic:
            rcdic["cnf"] = "y"
        if not "acnf" in rcdic:
            rcdic["acnf"] = "y"
        if not "ttip" in rcdic:
            rcdic["ttip"] = "y"
        if not "txtv" in rcdic:
            rcdic["txtv"] = "y"
        if not "efrm" in rcdic:
            rcdic["efrm"] = "n"
        if not "errs" in rcdic:
            rcdic["errs"] = "y"
        if not "theme" in rcdic:
            rcdic["theme"] = "Default"
        svedir = rcdic["dbdir"].replace(os.path.basename(rcdic["dbdir"]), "sve")
        if not "arcdir" in rcdic:
            rcdic["arcdir"] = svedir
        if not "bupdir" in rcdic:
            rcdic["bupdir"] = svedir
        if not "upgdir" in rcdic:
            rcdic["upgdir"] = rcdic["wrkdir"]
        check = []
        for pth in ("dbdir", "prgdir", "arcdir", "bupdir", "wrkdir", "upgdir"):
            if pth in rcdic and not os.path.isdir(rcdic[pth]):
                if not rcdic[pth] in check:
                    check.append(rcdic[pth])
        if check:
            dirs = ""
            for d in check:
                dirs = "\n%s" % d
            ok = askQuestion(None, "Missing Directories", "The Following Directories are Missing, Would You Like to Create them Now or Exit and Retry Later:\n%s" % dirs, default="yes")
            if ok == "yes":
                try:
                    for d in check:
                        os.makedirs(d)
                except:
                    return "error"
            else:
                return "error"
    except:
        if not default:
            return
        rcdic["dbase"] = "SQLite"
        rcdic["dbname"] = "tartan"
        rcdic["dbhost"] = "localhost"
        rcdic["dbport"] = ""
        try:
            rcdic["dbuser"] = getpass.getuser()
        except:
            rcdic["dbuser"] = ""
        rcdic["dbpwd"] = ""
        rcdic["prgdir"] = os.path.normpath(getPrgPath())
        if sys.platform == "win32":
            rcdic["mft"] = "Sans"
            rcdic["dft"] = "Courier New"
            rcdic["theme"] = "Raleigh"
            rcdic["dbdir"] = os.path.normpath("C:/Tartan/fle")
            rcdic["arcdir"] = os.path.normpath("C:/Tartan/arc")
            rcdic["bupdir"] = os.path.normpath("C:/Tartan/bup")
            rcdic["wrkdir"] = os.path.normpath("C:/Tartan/wrk")
            rcdic["upgdir"] = os.path.normpath("C:/Tartan/wrk")
            rcdic["vwr"] = os.path.normpath("C:/Tartan/prg/uty/SumatraPDF.exe")
            rcdic["prn"] = os.path.normpath("C:/Tartan/prg/uty/SumatraPDF.exe")
            rcdic["cnv"] = ""
            rcdic["exp"] = ""
        else:
            rcdic["mft"] = "Sans"
            rcdic["dft"] = "Courier"
            rcdic["theme"] = "Default"
            home = os.environ.get("HOME")
            rcdic["dbdir"] = os.path.normpath("%s/fle" % home)
            rcdic["arcdir"] = os.path.normpath("%s/arc" % home)
            rcdic["bupdir"] = os.path.normpath("%s/bup" % home)
            rcdic["wrkdir"] = os.path.normpath("%s/wrk" % home)
            rcdic["upgdir"] = os.path.normpath("%s/wrk" % home)
            rcdic["vwr"] = os.path.normpath("/usr/bin/xpdf")
            rcdic["prn"] = os.path.normpath("/usr/bin/lpr")
            rcdic["cnv"] = os.path.normpath("/usr/bin/pdftops")
            rcdic["exp"] = os.path.normpath("/usr/bin/ooffice")
        rcdic["cnf"] = "y"
        rcdic["acnf"] = "y"
        rcdic["ttip"] = "y"
        rcdic["txtv"] = "y"
        rcdic["efrm"] = "n"
        rcdic["errs"] = "y"
        gui, gtk, gobject, pango = doGtkImport()
        if gui:
            w = gtk.gdk.screen_width()
        else:
            w = 800
        if w >= 1280:
            rcdic["geo"] = "1024x768"
            rcdic["mfs"] = "10"
            rcdic["dfs"] = "12"
        elif w == 1024:
            rcdic["geo"] = "800x600"
            rcdic["mfs"] = "8"
            rcdic["dfs"] = "10"
        elif w == 800:
            rcdic["geo"] = "640x540"
            rcdic["mfs"] = "7"
            rcdic["dfs"] = "8"
        else:
            rcdic["geo"] = "800x600"
            rcdic["mfs"] = "8"
            rcdic["dfs"] = "10"
    return rcdic

def deepList(data):
    """fully copies trees of tuples to a tree of lists.
       deep_list( (1,2,(3,4)) ) returns [1,2,[3,4]]"""
    if type(data) != tuple:
        return data
    return [ deepList(d) for d in data ]

def removeFunctions(nam, dec=0):
    """removes all aggregate details from a column name
        nam = the column detail
        dec = decimal places"""
    if not nam.count("("):
        return nam
    nam = nam.lower().replace("avg(", "")
    nam = nam.lower().replace("count(", "")
    nam = nam.lower().replace("max(", "")
    nam = nam.lower().replace("min(", "")
    nam = nam.lower().replace("round(", "")
    nam = nam.lower().replace("stddev(", "")
    nam = nam.lower().replace("sum(", "")
    nam = nam.lower().replace("variance(", "")
    nam = nam.replace("),%s)" % dec, "")
    nam = nam.replace("), %s)" % dec, "")
    nam = nam.replace(")", "")
    return(nam.strip())

def getTrn(dbm, sys, dte=None, jon=None, whr=None, ord=None, neg=True, zer="Y", lim=None):
    """
    This function returns a list of column names and a list of lists of data.

    sys = crs - Creditor's Ledger
          drs - Debtor's Ledger
          mem - Member's Ledger
    dte = A cut-off curdt period where any transactions zeroed before this
          period will not be returned.
    jon = An addition to the join statement e.g. "cra_curdt <= 200612"
    whr = A list of tuples of where statements e.g. [("drt_cono", "=", 7)]
    ord = An order replacement.
    neg = True or False to include or exclude negative transaction balances
    zer = Y, A, N to include or exclude zero transaction balances
    lim = The number of records to return, defaults to all
    """
    from TartanClasses import ASD, Sql
    if sys == "crs":
        tab = "crstrn"
        sql = Sql(dbm, tab, prog=__name__)
        col = sql.crstrn_col[:]
        join = "left outer join crsage on cra_cono=crt_cono and "\
            "cra_acno=crt_acno and cra_type=crt_type and cra_ref1=crt_ref1"
        if dte:
            if whr:
                whr.append(("crt_curdt", "<=", dte))
            else:
                whr = [("crt_curdt", "<=", dte)]
            join = "%s and cra_curdt <= %s" % (join, dte)
        if jon:
            join = "%s and %s" % (join, jon)
    elif sys == "drs":
        tab = "drstrn"
        sql = Sql(dbm, tab, prog=__name__)
        col = sql.drstrn_col[:]
        join = "left outer join drsage on dra_cono=drt_cono and "\
            "dra_chain=drt_chain and dra_acno=drt_acno and "\
            "dra_type=drt_type and dra_ref1=drt_ref1"
        if dte:
            if whr:
                whr.append(("drt_curdt", "<=", dte))
            else:
                whr = [("drt_curdt", "<=", dte)]
            join = "%s and dra_curdt <= %s" % (join, dte)
        if jon:
            join = "%s and %s" % (join, jon)
    elif sys == "mem":
        tab = "memtrn"
        sql = Sql(dbm, tab, prog=__name__)
        col = sql.memtrn_col[:]
        join = "left outer join memage on mta_cono=mlt_cono and "\
            "mta_memno=mlt_memno and mta_type=mlt_type and mta_refno=mlt_refno"
        if dte:
            if whr:
                whr.append(("mlt_curdt", "<=", dte))
            else:
                whr = [("mlt_curdt", "<=", dte)]
            join = "%s and mta_curdt <= %s" % (join, dte)
        if jon:
            join = "%s and %s" % (join, jon)
    else:
        return None, None
    grp = ""
    for c in col:
        grp = "%s%s, " % (grp, c)
    grp = grp[:-2]
    if sys == "crs":
        col.append("max(cra_curdt)")
        col.append("round(sum(cra_amnt), 2)")
        if not ord:
            ord = "crt_trdt, crt_type, crt_ref1"
    elif sys == "drs":
        col.append("max(dra_curdt)")
        col.append("round(sum(dra_amnt), 2)")
        if not ord:
            ord = "drt_trdt, drt_type, drt_ref1"
    else:
        col.append("max(mta_curdt)")
        col.append("round(sum(mta_amnt), 2)")
        if not ord:
            ord = "mlt_trdt, mlt_type, mlt_refno"
    all = sql.getRec(tables=tab, join=join, cols=col, where=whr, group=grp,
        order=ord, limit=lim)
    col[-2] = "curdt"
    col[-1] = "paid"
    col.append("balance")
    dat = []
    if all and lim == 1:
        all = [all]
    tot = 0
    for rec in all:
        if sys == "crs":
            amt = rec[col.index("crt_tramt")]
        elif sys == "drs":
            amt = rec[col.index("drt_tramt")]
        else:
            amt = rec[col.index("mlt_tramt")]
        if rec[col.index("curdt")] == None:
            rec[col.index("curdt")] = 0
        pay = rec[col.index("paid")]
        if pay == None:
            pay = 0.0
            rec[col.index("paid")] = pay
        bal = float(ASD(amt) - ASD(pay))
        tot = float(ASD(tot) + ASD(bal))
        rec.append(bal)
        if bal or zer == "A":
            dat.append(rec)
        elif zer == "Y" and (not dte or not rec[col.index("curdt")] < dte):
            dat.append(rec)
    if not neg and tot < 0:
        return col, []
    return col, dat

def chkCtlAcc(mf, coy, acc, pwd=True):
    """
    This function returns None if the account is not an integrated control else
    it asks for a Password and returns an Error Message if not correct.
    """
    from TartanClasses import GetCtl, PwdConfirm, Sql

    gc = GetCtl(mf)
    # Get Company Master Record
    ctlmst = gc.getCtl("ctlmst", coy)
    if not ctlmst:
        return "Invalid Company Number"
    # Check for Company Control Records
    ctlctl = gc.getCtl("ctlctl", coy)
    if not ctlctl:
        return "Missing Control Records"
    # Check If Account Is a Control Record
    mod = []
    for x in range(0, 21, 2):
        mod.append(ctlmst["ctm_modules"][x:x+2])
    ctlacc = []
    # Load Ass Control
    if mod.count("AR"):
        ctl = gc.getCtl("assctl", coy)
        if not ctl:
            return "Missing Control Record - assctl"
        if ctl["cta_glint"] == "Y":
            sql = Sql(mf.dbm, "assgrp", prog=__name__)
            ass = sql.getRec(tables="assgrp", cols=["asg_assacc"],
                where=[("asg_cono", "=", coy)])
            for a in ass:
                ctlacc.append(a[0])
    # Load Crs Control
    if mod.count("CR"):
        ctl = gc.getCtl("crsctl", coy)
        if not ctl:
            return "Missing Control Record - crsctl"
        if ctl["ctc_glint"] == "Y":
            ctlacc.append(ctlctl["crs_ctl"])
    # Load Drs Control
    if mod.count("DR"):
        ctl = gc.getCtl("drsctl", coy)
        if not ctl:
            return "Missing Control Record - drsctl"
        if ctl["ctd_glint"] == "Y":
            ctlacc.append(ctlctl["drs_ctl"])
    # Load Mem Control
    if mod.count("ML"):
        ctl = gc.getCtl("memctl", coy)
        if not ctl:
            return "Missing Control Record - memctl"
        if ctl["mcm_glint"] == "Y":
            ctlacc.append(ctlctl["mem_ctl"])
    # Load Rtl Integration
    if mod.count("RT"):
        ctl = gc.getCtl("rtlctl", coy)
        if not ctl:
            return "Missing Control Record - rtlctl"
        if ctl["ctr_glint"] == "Y":
            sql = Sql(mf.dbm, "rtlprm", prog=__name__)
            rtl = sql.getRec(tables="rtlprm", cols=["rtp_rtlacc"],
                where=[("rtp_cono", "=", coy)])
            for a in rtl:
                ctlacc.append(a[0])
    if acc in ctlacc:
        if pwd:
            cf = PwdConfirm(mf, conum=coy, system="GEN", code="NoPost")
            if not cf.pwd or cf.flag == "no":
                return "%s Is An Integrated Control Account" % acc
        else:
            return "%s Is An Integrated Control Account" % acc

def chkAggregate(fld):
    for typ in ("avg(","count(","max(","min(","sum("):
        if fld.lower().count(typ):
            return True

def callModule(mf, df, mod, coy=None, period=None, user=None, args=None):
    if not mf.fixed:
        width, height = mf.window.get_size()
    if df:
        state = df.disableButtons()
        df.mstAlw.hide()
    exec "import %s" % mod
    if not period:
        if not user:
            if not args:
                if not coy:
                    exec "%s.%s(mf)" % (mod, mod)
                else:
                    exec "%s.%s(mf, coy[0], coy[1])" % (mod, mod)
            else:
                if not coy:
                    exec "%s.%s(mf, args)" % (mod, mod)
                else:
                    exec "%s.%s(mf, coy[0], coy[1], args=args)" % (mod, mod)
        elif not args:
            if not coy:
                exec "%s.%s(mf, user)" % (mod, mod)
            else:
                exec "%s.%s(mf, coy[0], coy[1], user)" % (mod, mod)
        else:
            if not coy:
                exec "%s.%s(mf, user, args)" % (mod, mod)
            else:
                exec "%s.%s(mf, coy[0], coy[1], user, args=args)" % (mod, mod)
    elif not user:
        if not args:
            if not coy:
                exec "%s.%s(mf, period)" % (mod, mod)
            else:
                exec "%s.%s(mf, coy[0], coy[1], period)" % (mod, mod)
        else:
            if not coy:
                exec "%s.%s(mf, period, args=args)" % (mod, mod)
            else:
                exec "%s.%s(mf, coy[0], coy[1], period, args=args)" % (mod, mod)
    elif not args:
        if not coy:
            exec "%s.%s(mf, period, user)" % (mod, mod)
        else:
            exec "%s.%s(mf, coy[0], coy[1], period, user)" % (mod, mod)
    else:
        if not coy:
            exec "%s.%s(mf, period, user, args=args)" % (mod, mod)
        else:
            exec "%s.%s(mf,coy[0],coy[1], period, user, args=args)" % (mod, mod)
    if df:
        df.mstAlw.show()
        df.enableButtons(state=state)
        df.focusField(df.frt, df.pag, df.col)
    if not mf.fixed:
        mf.window.resize(width, height)

def doChkCatChg(mf, cono, memno, yrnd, nxtdt):
    from TartanClasses import Sql

    sql = Sql(mf.dbm, ["memctc", "memmst", "memcat"])
    # Set nxtdt variables
    year = nxtdt / 10000
    month = nxtdt / 100 % 100
    day = nxtdt % 100
    # Get age and years as a member
    if type(memno) in (list, tuple):
        dob, doe = memno
    else:
        mst = sql.getRec(tables="memmst", where=[("mlm_cono", "=", cono),
            ("mlm_memno", "=", memno)], limit=1)
        if not mst:
            return
        dob = mst[sql.memmst_col.index("mlm_dob")]
        doe = mst[sql.memmst_col.index("mlm_entry")]
    if not dob or not doe:
        return
    age = dateDiff(dob, yrnd, "years")
    mem = dateDiff(doe, yrnd, "years")
    # Get current category
    cat = sql.getRec(tables="memcat", where=[("mlc_cono", "=", cono),
        ("mlc_memno", "=", memno), ("mlc_type", "=", "B")], limit=1)
    if not cat:
        return
    ocode = cat[sql.memcat_col.index("mlc_code")]
    start = cat[sql.memcat_col.index("mlc_start")]
    end = cat[sql.memcat_col.index("mlc_end")]
    last = cat[sql.memcat_col.index("mlc_last")]
    # Get current category limits and next code (ncode)
    rec = sql.getRec(tables="memctc", where=[("mcc_cono", "=", cono),
        ("mcc_type", "=", "B"), ("mcc_code", "=", ocode)], limit=1)
    if not rec:
        return
    desc = rec[sql.memctc_col.index("mcc_desc")]
    freq = rec[sql.memctc_col.index("mcc_freq")]
    glac = rec[sql.memctc_col.index("mcc_glac")]
    ncode = rec[sql.memctc_col.index("mcc_ncode")]
    age_l = rec[sql.memctc_col.index("mcc_age_l")]
    and_s = rec[sql.memctc_col.index("mcc_and_s")]
    or_s = rec[sql.memctc_col.index("mcc_or_s")]
    # See if limits have been reached and if so return new code details
    ret = None
    while ncode:
        if not age_l and not or_s:
            break
        if not and_s and age_l and age > age_l:
            chg = True
        elif and_s and age_l and age >= age_l and mem >= and_s:
            chg = True
        elif or_s and mem >= or_s:
            chg = True
        else:
            chg = False
        if chg:
            ret = (ocode, start, end, last, freq, desc, glac, ncode)
        chk = sql.getRec(tables="memctc", where=[("mcc_cono", "=", cono),
            ("mcc_type", "=", "B"), ("mcc_code", "=", ncode)], limit=1)
        ncode = chk[sql.memctc_col.index("mcc_ncode")]
        age_l = chk[sql.memctc_col.index("mcc_age_l")]
        and_s = rec[sql.memctc_col.index("mcc_and_s")]
        or_s = rec[sql.memctc_col.index("mcc_or_s")]
    return ret

def doWriteExport(**args):
    """
    Write an Export file using the following arguments:
        xtype  - C for csv and X for xls or a tuple of both
        name   - The output file name without an extension
        heads  - A list of the Report headings
        colsh  - A list of lists of Column headings comprising:
                    Text
                    Start Col (Default is current col)
                    End Col   (Default is current col)
        forms  - A list of column format details comprising:
                    Format per CCD
                    Size   per CCD
                    Clickable (True or False) - Default is False (ScrollGrid)
                    Negative colour (ScrollGrid)
                    Ignore for ULINE[D/S] and TOTAL - Default is False
        datas  - The column data comprising:
                    Type of data:
                        "PAGE"   - A new page/sheet
                        "HEAD"   - A heading in column 0 only
                        "BODY"   - A body line of columns of data
                        "BLANK"  - A blank line - skip
                        "ULINED" - A double underline - skip
                        "ULINES" - A sigle underline - skip
                        "TOTAL"  - A total line of columns of data
                    List of values
        rcdic  - The tartanrc dictionary
        view   - Whether to view the report
    """
    import os, sys

    def viewFile(name):
        if not args["rcdic"] or not args["rcdic"]["exp"]:
            if sys.platform == "win32":
                os.startfile(name)
            elif os.path.isfile("/usr/bin/ooffice"):
                os.spawnv(os.P_NOWAIT, "/usr/bin/ooffice",
                    ("ooffice", name))
            elif os.path.isfile("/usr/bin/libreoffice"):
                os.spawnv(os.P_NOWAIT, "/usr/bin/libreoffice",
                    ("libreoffice", name))
            return
        try:
            exe, cmd = parsePrg(args["rcdic"]["exp"])
            cmd.append(name)
            os.spawnv(os.P_NOWAIT, exe, tuple(cmd))
        except:
            return "%s Not Found or Not Accessible" % exe

    if not "view" in args:
        view = True
    else:
        view = args["view"]
    if type(args["xtype"]) == str:
        args["xtype"] = [args["xtype"]]
    types = []
    for t in args["xtype"]:
        types.append(t.upper())

    if "C" in types:
        flenam = open(args["name"] + ".csv", "w")
        head = ""
        for valc in args["colsh"][-1]:
            if type(valc) in (list, tuple):
                text = valc[0]
            else:
                text = valc
            if head:
                head = '%s,"%s"' % (head, text)
            else:
                head = '"%s"' % text
        if head:
            flenam.write(head + "\n")
        for valx in args["datas"]:
            if valx[0] != "BODY":
                continue
            if not valx[1] or len(valx[1]) != len(args["forms"]):
                # Blank line or invalid line length
                continue
            line = ""
            for colc, valc in enumerate(valx[1]):
                if args["forms"][colc][0][0].lower() == "d" or \
                   args["forms"][colc][0][1].lower() in ("d", "i", "l"):
                    if line:
                        line = "%s,%s" % (line, valc)
                    else:
                        line = "%s" % valc
                elif line:
                    line = '%s,"%s"' % (line, valc)
                else:
                    line = '"%s"' % valc
            flenam.write(line + "\n")
        # Save the csv file
        flenam.close()
        if view and not "X" in types:
            # View the csv file
            viewFile(args["name"] + ".csv")

    styles = {}
    def getStyle(fmt, num=None):
        if not num:
            idx = fmt
        else:
            idx = fmt + num
        if not idx in styles:
            if num:
                styles[idx] = ezxf(fmt, num_format_str=num)
            else:
                styles[idx] = ezxf(fmt)
        return styles[idx]

    def createSheet(page, heads=args["heads"], colsh=args["colsh"]):
        sheet = book.add_sheet("Page %s" % page)
        # Main headings
        rowx = 0
        blank = False
        hxf = getStyle(fmt + ", bold on")
        for valx in heads:
            if valx and blank:
                blank = False
                sheet.write(rowx, 0, "", hxf)
                rowx += 1
            sheet.write(rowx, 0, valx, hxf)
            if valx:
                blank = True
            rowx += 1
        rowx += 1
        # Column headings
        for colx, valx in enumerate(colsh):
            for cx, vx in enumerate(valx):
                if type(vx) in (list, tuple) and len(vx) == 3:
                    hxf = getStyle(fmt + ", bold on; align: horz centre")
                elif args["forms"][cx][0] in alpha:
                    hxf = getStyle(fmt + ", bold on; align: horz left")
                else:
                    hxf = getStyle(fmt + ", bold on; align: horz right")
                if type(vx) in (list, tuple):
                    if len(vx) == 3:
                        sheet.write_merge(rowx, rowx, vx[1], vx[2], vx[0], hxf)
                    else:
                        sheet.write(rowx, cx, vx[0], hxf)
                else:
                    sheet.write(rowx, cx, vx, hxf)
                sheet.col(cx).width = int(args["forms"][cx][1]) * 400
            if colx != len(args["colsh"]) - 1:
                rowx += 1
        # Freeze the headings
        sheet.set_panes_frozen(True) # frozen headings instead of split panes
        sheet.set_horz_split_pos(rowx + 1) # in general, freeze after last row
        sheet.set_remove_splits(True) # if user unfreezes, don't leave a split
        return sheet, rowx

    if "X" in types:
        import datetime
        import xlwt
        ezxf = xlwt.easyxf

        # Variables and Formats
        fmt = "font: name courier new"
        alpha = ("HA", "LA", "La", "NA", "Na", "TX", "UA", "Ua")
        xf_map = {
            "D1": "yyyy-mm-dd",
            "d1": "yyyy-mm-dd",
            "D2": "yyyy-mm",
            "d2": "yyyy-mm",
            "UI": "########0",
            "UL": "#0",
            "SI": "#0",
            "SL": "#0",
            "US": "#0",
            "SD": "#0.00",
            "UD": "#0.00"}
        for an in alpha:
            xf_map[an] = "@"
        # Create the Workbook and Worksheet
        style = xlwt.XFStyle()
        font = xlwt.Formatting.Font()
        font.name = "Courier New"
        style.font = font
        book = xlwt.Workbook(encoding="utf8")
        page = 1
        if args["datas"][0][0] != "PAGE":
            sheet, rowx = createSheet(page)
            page += 1
        xfstyles = []
        # Generate the body
        for num, row in enumerate(args["datas"]):
            if row[0] == "PAGE":
                sheet, rowx = createSheet(page, row[1][0], row[1][1])
                page += 1
                continue
            if row[0] in ("ULINES", "ULINED"):
                continue
            rowx += 1
            if row[0] == "BLANK":
                sheet.write(rowx, 0, "", getStyle(fmt))
                continue
            chk = num + 1
            if not chk == len(args["datas"]):
                if args["datas"][chk][0] == "ULINES":
                    unl = "s"
                elif args["datas"][chk][0] == "ULINED":
                    unl = "d"
                else:
                    unl = False
            else:
                unl = False
            for colx, valx in enumerate(row[1]):
                if row[0] in ("HEAD", "TOTAL"):
                    hxf = fmt + ", bold on"
                else:
                    hxf = fmt
                if type(valx) == str:
                    sheet.write(rowx, colx, valx,
                        getStyle(hxf + "; align: horz left"))
                    continue
                if len(args["forms"][colx]) == 5:
                    if row[0] == "TOTAL" or not valx:
                        sheet.write(rowx, colx, "",
                            getStyle(hxf + "; align: horz left"))
                        continue
                elif valx and args["forms"][colx][0].lower() == "d1":
                    # D1 Date conversion
                    valx = datetime.datetime(valx/10000,(valx/100)%100,valx%100)
                elif valx and args["forms"][colx][0].lower() == "d2":
                    # D2 Date conversion
                    valx = datetime.datetime(valx/100, valx%100, 1)
                elif args["forms"][colx][0] not in alpha and unl:
                    if unl == "s":
                        hxf = hxf + "; borders: bottom thin"
                    else:
                        hxf = hxf + "; borders: bottom double"
                if args["forms"][colx][0][0] == "d" and not valx:
                    sheet.write(rowx, colx, valx,
                        getStyle(hxf, num=xf_map["UI"]))
                else:
                    sheet.write(rowx, colx, valx,
                        getStyle(hxf, num=xf_map[args["forms"][colx][0]]))
        # Save the spreadsheet
        book.save(args["name"] + ".xls")
        # View the spreadsheet
        if view:
            viewFile(args["name"] + ".xls")

def ftpDownload(server, srce, name=None, word=None, dest=None, close=True):
    import ftplib

    if not name:
        name = "anonymous"
    if not word:
        password = "info@tartan.co.za"
    if not dest:
        dest = srce
    if close:
        ofle = open(dest, "wb")
    else:
        ofle = dest
    try:
        ftp = ftplib.FTP(server, name, word)
        ftp.retrbinary("RETR " + srce, ofle.write)
        if close:
            ofle.close()
        ftp.close()
    except:
        return "Download Error"

def doDrawTable(canvas, cw, ld, rr, ppad=1, spad=1, font=True):
    """
    Draw a pdf table in reportlab

        canvas  -   A reportlab canvas
        cw      -   The width of a single character i.e. canvas.stringWidth
        ld      -   The line depth e.g. inch / 6
        rr      -   The layout options of the table as follows:
                        "margins": ((left, right), (top, bottom))
                        "repeat":  (across, below)  # repeat the table
                        "rows":    (
                                 (x, y, w, h, fill, text, qty), )
        ppad    -   Padding
        spad    -   Padding
        font    -   Whether or not to set a font.
    """
    table = []
    last = 0
    cols = []
    rows = []
    rest = []
    for row in rr["rows"]:
        if type(row[2]) in (list, tuple):
            # Mutiple columns
            if len(row) == 4:
                qty = row[3]
            else:
                qty = 1
            cv = row[0] # x
            tc = [[cv]] # total cols
            rv = row[1] # y
            tr = [[rv]] * qty # total rows
            rt = []
            for q in range(qty):
                cv = row[0] # x
                tc = [[cv]] # total cols
                rt = []
                for n, r in enumerate(row[2]):
                    if len(row[2]) != n+1:
                        cv += r[0]  # add col width
                        tc.append([cv])
                    if len(r) == 4 and type(r[3]) in (list, tuple):
                        ap = list(r[:3])
                        ap.append(r[3][q])
                        rt.append(ap)
                    else:
                        rt.append(list(r))
                cols.extend(tc)
                rest.extend(rt)
            rv = row[1] # y
            tr = [] # total rows
            for q in range(qty):
                tr.append([rv]) # total rows
                for n, r in enumerate(row[2]):
                    if len(row[2]) != n+1:
                        tr.append([rv])
                rv += r[1]  # add row height
            rows.extend(tr)
        else:
            cols.append([row[0]])
            rows.append([row[1]])
            rest.append(list(row[2:]))
    for x in range(1, rr["repeat"][0]):
        for n, c in enumerate(cols):
            cols[n] = c + [c[-1] + rr["margins"][0][1] + 1]
    d = (rr["margins"][1][1] - rr["margins"][1][0])
    for n, r in enumerate(rows):
        nxt = r
        for x in range(1, rr["repeat"][1]):
            last = (nxt[-1] + d) + spad
            nxt = nxt + [last]
        rows[n] = nxt
    for n, r in enumerate(rows):
        table.append([cols[n]] + [rows[n]] + rest[n])
    for r in table:
        if len(r) > 4 and r[4]:
            canvas.setFillGray(r[4])
            fill = 1
        else:
            fill=0
        for y in r[0]:
            for z in r[1]:
                canvas.rect(y*cw, z*ld, r[2]*cw, r[3]*ld, fill=fill)
                if len(r) == 6:
                    canvas.setFillGray(0)
                    if font:
                        if ppad == 1:
                            canvas.setFont("Helvetica-Bold", 10)
                        else:
                            canvas.setFont("Helvetica-Bold", 12)
                    canvas.drawString((y+.5)*cw, (z+ppad)*ld, r[5])
                    canvas.setFillGray(r[4])
        canvas.setFillGray(0)
    return last, table

def getDeposit(mf, conum, date, owner, code, acno):
    """
    Calculate the deposit plus interest on rented premises:

    conum - The company number
    owner - The owner's code
    code  - the premise's code
    acno  - The tenant's account number
    """
    from TartanClasses import ASD, SChoice, Sql

    sql = Sql(mf.dbm, ["rcaint", "rcatnt"], prog=__name__)
    tnt = sql.getRec(tables="rcatnt", where=[("rtu_cono", "=", conum),
        ("rtu_owner", "=", owner), ("rtu_code", "=", code), ("rtu_acno",
        "=", acno), ("rtu_trdt", "<=", date), ("rtu_mtyp", "=", 2)],
        order="rtu_trdt")
    if not tnt:
        return
    coi = sql.rcaint_col
    cot = sql.rcatnt_col
    dep = 0
    all = []
    for n, t in enumerate(tnt):
        start = t[cot.index("rtu_trdt")]
        dep = float(ASD(dep) + ASD(t[cot.index("rtu_tramt")]))
        rte = sql.getRec(tables="rcaint", where=[("rci_cono", "=", conum),
            ("rci_date", "<=", start)], order="rci_date")
        if not rte:
            return
        all.append([start, t[cot.index("rtu_tramt")], rte[-1]])
        if (n + 1) == len(tnt):
            end = date
        else:
            end = tnt[n + 1][cot.index("rtu_trdt")]
        rat = sql.getRec(tables="rcaint", where=[("rci_cono", "=", conum),
            ("rci_date", ">", start), ("rci_date", "<", end)], order="rci_date")
        for r in rat:
            all.append([r[coi.index("rci_date")], 0, r])
    if not all:
        return
    if all[-1][2][0] < date:
        all.append([date, 0, []])
    begin = True
    bal = 0
    dep = 0
    dat = []
    for n, a in enumerate(all):
        cap = float(ASD(0) - ASD(a[1]))
        dep =float(ASD(dep) + ASD(cap))
        if begin:
            begin = False
            start = a[0]
            bal = cap
            dat.append([start, cap, 0, 0, 0, bal, 0])
            continue
        cdate = all[n][0]
        bankr = all[n-1][2][3]
        commr = all[n-1][2][4]
        invtr = float(ASD(bankr) - ASD(commr))
        days = dateDiff(start, cdate, ptype="days")
        banki = round(bal * bankr * days / 365.0 / 100.0, 2)
        commi = round(bal * commr * days / 365.0 / 100.0, 2)
        invti = round(bal * invtr * days / 365.0 / 100.0, 2)
        bal = float(ASD(bal) + ASD(invti) + ASD(cap))
        dat.append([cdate, cap, days, invtr, invti, bal, commi])
        if not dep:
            break
        start = cdate
    return dat

def genAccNum(name, seq, size=7):
    valid = map(chr, range(65, 91))
    valid.extend(map(chr, range(97, 123)))
    valid.append("&")
    data = ""
    for c in name:
        if c in valid:
            data = data + c
    if len(data) < (size - 2):
        data = data + ("0" * ((size - 2) - len(data)))
    acno = ""
    for c in range(0, (size - 2)):
        acno = (acno + data[c]).upper()
    acno = acno.strip()
    text = "%s%0" + str((size - len(acno))) + "d"
    return text % (acno, seq)

def getNextCode(sql, table, column, where=None, last=999):
    nxt = sql.getRec(tables=table, cols=[column], where=where, order=column)
    code = 1
    for n in range(1, last):
        if not [n] in nxt:
            code = n
            break
    return code

def flipImage(image):
    import os
    try:
        if not os.path.exists(image):
            raise Exception
        import tempfile
        from PIL import Image
        sufx = image.split(".")[-1]
        tmpfle = tempfile.NamedTemporaryFile(suffix=".%s" % sufx, delete=False)
        tmpfle.close()
        flenam = tmpfle.name
        image = Image.open(image)
        image.transpose(Image.FLIP_TOP_BOTTOM).save(flenam)
        return flenam
    except:
        return
# END
